@@ -1,8 +1,8 @@
 /* attr.c - routines for dealing with attributes */
-/* $OpenLDAP: pkg/ldap/servers/slapd/attr.c,v 1.100.2.4 2005/04/29 21:29:04 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/attr.c,v 1.112.2.4 2007/08/31 23:13:58 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,14 +40,129 @@
 
 #include "slap.h"
 
+/*
+ * Allocate in chunks, minimum of 1000 at a time.
+ */
+#define	CHUNK_SIZE	1000
+typedef struct slap_list {
+	struct slap_list *next;
+} slap_list;
+static slap_list *attr_chunks;
+static Attribute *attr_list;
+static ldap_pvt_thread_mutex_t attr_mutex;
+
+int
+attr_prealloc( int num )
+{
+	Attribute *a;
+	slap_list *s;
+
+	if (!num) return 0;
+
+	s = ch_calloc( 1, sizeof(slap_list) + num * sizeof(Attribute));
+	s->next = attr_chunks;
+	attr_chunks = s;
+
+	a = (Attribute *)(s+1);
+	for ( ;num>1; num--) {
+		a->a_next = a+1;
+		a++;
+	}
+	a->a_next = attr_list;
+	attr_list = (Attribute *)(s+1);
+
+	return 0;
+}
+
+Attribute *
+attr_alloc( AttributeDescription *ad )
+{
+	Attribute *a;
+
+	ldap_pvt_thread_mutex_lock( &attr_mutex );
+	if ( !attr_list )
+		attr_prealloc( CHUNK_SIZE );
+	a = attr_list;
+	attr_list = a->a_next;
+	a->a_next = NULL;
+	ldap_pvt_thread_mutex_unlock( &attr_mutex );
+	
+	a->a_desc = ad;
+
+	return a;
+}
+
+/* Return a list of num attrs */
+Attribute *
+attrs_alloc( int num )
+{
+	Attribute *head = NULL;
+	Attribute **a;
+
+	ldap_pvt_thread_mutex_lock( &attr_mutex );
+	for ( a = &attr_list; *a && num > 0; a = &(*a)->a_next ) {
+		if ( !head )
+			head = *a;
+		num--;
+	}
+	attr_list = *a;
+	if ( num > 0 ) {
+		attr_prealloc( num > CHUNK_SIZE ? num : CHUNK_SIZE );
+		*a = attr_list;
+		for ( ; *a && num > 0; a = &(*a)->a_next ) {
+			if ( !head )
+				head = *a;
+			num--;
+		}
+		attr_list = *a;
+	}
+	*a = NULL;
+	ldap_pvt_thread_mutex_unlock( &attr_mutex );
+
+	return head;
+}
+
+
+void
+attr_clean( Attribute *a )
+{
+	if ( a->a_nvals && a->a_nvals != a->a_vals &&
+		!( a->a_flags & SLAP_ATTR_DONT_FREE_VALS )) {
+		if ( a->a_flags & SLAP_ATTR_DONT_FREE_DATA ) {
+			free( a->a_nvals );
+		} else {
+			ber_bvarray_free( a->a_nvals );
+		}
+	}
+	/* a_vals may be equal to slap_dummy_bv, a static empty berval;
+	 * this is used as a placeholder for attributes that do not carry
+	 * values, e.g. when proxying search entries with the "attrsonly"
+	 * bit set. */
+	if ( a->a_vals != &slap_dummy_bv &&
+		!( a->a_flags & SLAP_ATTR_DONT_FREE_VALS )) {
+		if ( a->a_flags & SLAP_ATTR_DONT_FREE_DATA ) {
+			free( a->a_vals );
+		} else {
+			ber_bvarray_free( a->a_vals );
+		}
+	}
+	a->a_desc = NULL;
+	a->a_vals = NULL;
+	a->a_nvals = NULL;
+#ifdef LDAP_COMP_MATCH
+	a->a_comp_data = NULL;
+#endif
+	a->a_flags = 0;
+}
+
 void
 attr_free( Attribute *a )
 {
-	if ( a->a_nvals && a->a_nvals != a->a_vals ) {
-		ber_bvarray_free( a->a_nvals );
-	}
-	ber_bvarray_free( a->a_vals );
-	free( a );
+	attr_clean( a );
+	ldap_pvt_thread_mutex_lock( &attr_mutex );
+	a->a_next = attr_list;
+	attr_list = a;
+	ldap_pvt_thread_mutex_unlock( &attr_mutex );
 }
 
 #ifdef LDAP_COMP_MATCH
@@ -70,25 +185,33 @@ comp_tree_free( Attribute *a )
 void
 attrs_free( Attribute *a )
 {
-	Attribute *next;
+	if ( a ) {
+		Attribute *b = (Attribute *)0xBAD, *tail, *next;
 
-	for( ; a != NULL ; a = next ) {
-		next = a->a_next;
-		attr_free( a );
+		/* save tail */
+		tail = a;
+		do {
+			next = a->a_next;
+			attr_clean( a );
+			a->a_next = b;
+			b = a;
+			a = next;
+		} while ( next );
+
+		ldap_pvt_thread_mutex_lock( &attr_mutex );
+		/* replace NULL with current attr list and let attr list
+		 * start from last attribute returned to list */
+		tail->a_next = attr_list;
+		attr_list = b;
+		ldap_pvt_thread_mutex_unlock( &attr_mutex );
 	}
 }
 
-Attribute *
-attr_dup( Attribute *a )
+static void
+attr_dup2( Attribute *tmp, Attribute *a )
 {
-	Attribute *tmp;
-
-	if ( a == NULL) return NULL;
-
-	tmp = ch_malloc( sizeof(Attribute) );
-
 	if ( a->a_vals != NULL ) {
-		int i;
+		int	i;
 
 		for ( i = 0; !BER_BVISNULL( &a->a_vals[i] ); i++ ) {
 			/* EMPTY */ ;
@@ -103,57 +226,64 @@ attr_dup( Attribute *a )
 		BER_BVZERO( &tmp->a_vals[i] );
 
 		/* a_nvals must be non null; it may be equal to a_vals */
-		assert( a->a_nvals );
+		assert( a->a_nvals != NULL );
 
 		if ( a->a_nvals != a->a_vals ) {
+			int	j;
+
 			tmp->a_nvals = ch_malloc( (i + 1) * sizeof(struct berval) );
-			for ( i = 0; !BER_BVISNULL( &a->a_nvals[i] ); i++ ) {
-				ber_dupbv( &tmp->a_nvals[i], &a->a_nvals[i] );
-				if ( BER_BVISNULL( &tmp->a_nvals[i] ) ) break;
+			for ( j = 0; !BER_BVISNULL( &a->a_nvals[j] ); j++ ) {
+				assert( j < i );
+				ber_dupbv( &tmp->a_nvals[j], &a->a_nvals[j] );
+				if ( BER_BVISNULL( &tmp->a_nvals[j] ) ) break;
 				/* FIXME: error? */
 			}
-			BER_BVZERO( &tmp->a_nvals[i] );
+			assert( j == i );
+			BER_BVZERO( &tmp->a_nvals[j] );
 
 		} else {
 			tmp->a_nvals = tmp->a_vals;
 		}
-
-	} else {
-		tmp->a_vals = NULL;
-		tmp->a_nvals = NULL;
 	}
+}
 
-	tmp->a_desc = a->a_desc;
-	tmp->a_next = NULL;
-	tmp->a_flags = 0;
-#ifdef LDAP_COMP_MATCH
-	tmp->a_comp_data = NULL;
-#endif
+Attribute *
+attr_dup( Attribute *a )
+{
+	Attribute *tmp;
 
+	if ( a == NULL) return NULL;
+
+	tmp = attr_alloc( a->a_desc );
+	attr_dup2( tmp, a );
 	return tmp;
 }
 
 Attribute *
 attrs_dup( Attribute *a )
 {
-	Attribute *tmp, **next;
+	int i;
+	Attribute *tmp, *anew;
 
 	if( a == NULL ) return NULL;
 
-	tmp = NULL;
-	next = &tmp;
-
-	for( ; a != NULL ; a = a->a_next ) {
-		*next = attr_dup( a );
-		next = &((*next)->a_next);
+	/* count them */
+	for( tmp=a,i=0; tmp; tmp=tmp->a_next ) {
+		i++;
 	}
-	*next = NULL;
 
-	return tmp;
+	anew = attrs_alloc( i );
+
+	for( tmp=anew; a; a=a->a_next ) {
+		tmp->a_desc = a->a_desc;
+		attr_dup2( tmp, a );
+		tmp=tmp->a_next;
+	}
+
+	return anew;
 }
 
 
-
 /*
  * attr_merge - merge the given type and value with the list of
  * attributes in attrs.
@@ -183,22 +313,16 @@ attr_merge(
 	}
 
 	if ( *a == NULL ) {
-		*a = (Attribute *) ch_malloc( sizeof(Attribute) );
-		(*a)->a_desc = desc;
-		(*a)->a_vals = NULL;
-		(*a)->a_nvals = NULL;
-		(*a)->a_next = NULL;
-		(*a)->a_flags = 0;
-#ifdef LDAP_COMP_MATCH
-		(*a)->a_comp_data = NULL;
-#endif
+		*a = attr_alloc( desc );
 	} else {
 		/*
 		 * FIXME: if the attribute already exists, the presence
 		 * of nvals and the value of (*a)->a_nvals must be consistent
 		 */
 		assert( ( nvals == NULL && (*a)->a_nvals == (*a)->a_vals )
-				|| ( nvals != NULL && (*a)->a_nvals != (*a)->a_vals ) );
+				|| ( nvals != NULL && (
+					( (*a)->a_vals == NULL && (*a)->a_nvals == NULL )
+					|| ( (*a)->a_nvals != (*a)->a_vals ) ) ) );
 	}
 
 	rc = value_add( &(*a)->a_vals, vals );
@@ -215,15 +339,22 @@ attr_merge(
 	return rc;
 }
 
+/*
+ * if a normalization function is defined for the equality matchingRule
+ * of desc, the value is normalized and stored in nval; otherwise nval 
+ * is NULL
+ */
 int
-attr_merge_normalize(
-	Entry		*e,
-	AttributeDescription *desc,
-	BerVarray	vals,
-	void	 *memctx )
+attr_normalize(
+	AttributeDescription	*desc,
+	BerVarray		vals,
+	BerVarray		*nvalsp,
+	void	 		*memctx )
 {
+	int		rc = LDAP_SUCCESS;
 	BerVarray	nvals = NULL;
-	int		rc;
+
+	*nvalsp = NULL;
 
 	if ( desc->ad_type->sat_equality &&
 		desc->ad_type->sat_equality->smr_normalize )
@@ -234,7 +365,7 @@ attr_merge_normalize(
 
 		nvals = slap_sl_calloc( sizeof(struct berval), i + 1, memctx );
 		for ( i = 0; !BER_BVISNULL( &vals[i] ); i++ ) {
-			rc = (*desc->ad_type->sat_equality->smr_normalize)(
+			rc = desc->ad_type->sat_equality->smr_normalize(
 					SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
 					desc->ad_type->sat_syntax,
 					desc->ad_type->sat_equality,
@@ -242,18 +373,38 @@ attr_merge_normalize(
 
 			if ( rc != LDAP_SUCCESS ) {
 				BER_BVZERO( &nvals[i + 1] );
-				goto error_return;
+				break;
 			}
 		}
 		BER_BVZERO( &nvals[i] );
+		*nvalsp = nvals;
 	}
 
-	rc = attr_merge( e, desc, vals, nvals );
-
-error_return:;
-	if ( nvals != NULL ) {
+	if ( rc != LDAP_SUCCESS && nvals != NULL ) {
 		ber_bvarray_free_x( nvals, memctx );
 	}
+
+	return rc;
+}
+
+int
+attr_merge_normalize(
+	Entry			*e,
+	AttributeDescription	*desc,
+	BerVarray		vals,
+	void	 		*memctx )
+{
+	BerVarray	nvals = NULL;
+	int		rc;
+
+	rc = attr_normalize( desc, vals, &nvals, memctx );
+	if ( rc == LDAP_SUCCESS ) {
+		rc = attr_merge( e, desc, vals, nvals );
+		if ( nvals != NULL ) {
+			ber_bvarray_free_x( nvals, memctx );
+		}
+	}
+
 	return rc;
 }
 
@@ -274,15 +425,7 @@ attr_merge_one(
 	}
 
 	if ( *a == NULL ) {
-		*a = (Attribute *) ch_malloc( sizeof(Attribute) );
-		(*a)->a_desc = desc;
-		(*a)->a_vals = NULL;
-		(*a)->a_nvals = NULL;
-		(*a)->a_next = NULL;
-		(*a)->a_flags = 0;
-#ifdef LDAP_COMP_MATCH
-		(*a)->a_comp_data = NULL;
-#endif
+		*a = attr_alloc( desc );
 	}
 
 	rc = value_add_one( &(*a)->a_vals, val );
@@ -298,6 +441,39 @@ attr_merge_one(
 	return rc;
 }
 
+/*
+ * if a normalization function is defined for the equality matchingRule
+ * of desc, the value is normalized and stored in nval; otherwise nval 
+ * is NULL
+ */
+int
+attr_normalize_one(
+	AttributeDescription *desc,
+	struct berval	*val,
+	struct berval	*nval,
+	void		*memctx )
+{
+	int		rc = LDAP_SUCCESS;
+
+	BER_BVZERO( nval );
+
+	if ( desc->ad_type->sat_equality &&
+		desc->ad_type->sat_equality->smr_normalize )
+	{
+		rc = desc->ad_type->sat_equality->smr_normalize(
+				SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
+				desc->ad_type->sat_syntax,
+				desc->ad_type->sat_equality,
+				val, nval, memctx );
+
+		if ( rc != LDAP_SUCCESS ) {
+			return rc;
+		}
+	}
+
+	return rc;
+}
+
 int
 attr_merge_normalize_one(
 	Entry		*e,
@@ -305,22 +481,12 @@ attr_merge_normalize_one(
 	struct berval	*val,
 	void		*memctx )
 {
-	struct berval	nval;
+	struct berval	nval = BER_BVNULL;
 	struct berval	*nvalp = NULL;
 	int		rc;
 
-	if ( desc->ad_type->sat_equality &&
-		desc->ad_type->sat_equality->smr_normalize )
-	{
-		rc = (*desc->ad_type->sat_equality->smr_normalize)(
-				SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
-				desc->ad_type->sat_syntax,
-				desc->ad_type->sat_equality,
-				val, &nval, memctx );
-
-		if ( rc != LDAP_SUCCESS ) {
-			return rc;
-		}
+	rc = attr_normalize_one( desc, val, &nval, memctx );
+	if ( rc == LDAP_SUCCESS && !BER_BVISNULL( &nval ) ) {
 		nvalp = &nval;
 	}
 
@@ -395,3 +561,22 @@ attr_delete(
 	return LDAP_NO_SUCH_ATTRIBUTE;
 }
 
+int
+attr_init( void )
+{
+	ldap_pvt_thread_mutex_init( &attr_mutex );
+	return 0;
+}
+
+int
+attr_destroy( void )
+{
+	slap_list *a;
+
+	for ( a=attr_chunks; a; a=attr_chunks ) {
+		attr_chunks = a->next;
+		free( a );
+	}
+	ldap_pvt_thread_mutex_destroy( &attr_mutex );
+	return 0;
+}
