@@ -1,8 +1,8 @@
 /* dn2entry.c - routines to deal with the dn2id / id2entry glue */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/dn2entry.c,v 1.19.2.3 2003/12/01 00:58:38 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/dn2entry.c,v 1.28.2.3 2007/08/31 23:14:01 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2003 The OpenLDAP Foundation.
+ * Copyright 2000-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,18 +34,14 @@ bdb_dn2entry(
 	struct berval *dn,
 	EntryInfo **e,
 	int matched,
-	u_int32_t locker,
+	BDB_LOCKER locker,
 	DB_LOCK *lock )
 {
 	EntryInfo *ei = NULL;
-	int rc;
+	int rc, rc2;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( CACHE, ARGS, "bdb_dn2entry(\"%s\")\n", dn->bv_val, 0, 0 );
-#else
 	Debug(LDAP_DEBUG_TRACE, "bdb_dn2entry(\"%s\")\n",
 		dn->bv_val, 0, 0 );
-#endif
 
 	*e = NULL;
 
@@ -56,11 +52,15 @@ bdb_dn2entry(
 			 * or not.
 			 */
 			*e = ei;
-			if ( ei && ei->bei_id )
-				bdb_cache_find_id( op, tid, ei->bei_id,
+			if ( ei && ei->bei_id ) {
+				rc2 = bdb_cache_find_id( op, tid, ei->bei_id,
 					&ei, 1, locker, lock );
-			else if ( ei )
+				if ( rc2 ) rc = rc2;
+			} else if ( ei ) {
 				bdb_cache_entryinfo_unlock( ei );
+				memset( lock, 0, sizeof( *lock ));
+				lock->mode = DB_LOCK_NG;
+			}
 		} else if ( ei ) {
 			bdb_cache_entryinfo_unlock( ei );
 		}
@@ -71,9 +71,12 @@ bdb_dn2entry(
 			*e = ei;
 		} else if ( matched && rc == DB_NOTFOUND ) {
 			/* always return EntryInfo */
-			ei = ei->bei_parent;
-			bdb_cache_find_id( op, tid, ei->bei_id, &ei, 1,
-				locker, lock );
+			if ( ei->bei_parent ) {
+				ei = ei->bei_parent;
+				rc2 = bdb_cache_find_id( op, tid, ei->bei_id, &ei, 1,
+					locker, lock );
+				if ( rc2 ) rc = rc2;
+			}
 			*e = ei;
 		}
 	}
