@@ -1,8 +1,8 @@
 /* init.c - initialize bdb backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.177.2.10 2005/05/10 16:07:03 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.247.2.4 2007/08/31 23:14:01 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2005 The OpenLDAP Foundation.
+ * Copyright 2000-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,6 +26,7 @@
 #include <lutil.h>
 #include <ldap_rq.h>
 #include "alock.h"
+#include "config.h"
 
 static const struct bdbi_database {
 	char *file;
@@ -41,10 +42,15 @@ static const struct bdbi_database {
 typedef void * db_malloc(size_t);
 typedef void * db_realloc(void *, size_t);
 
+#define bdb_db_init	BDB_SYMBOL(db_init)
+#define bdb_db_open BDB_SYMBOL(db_open)
+#define bdb_db_close BDB_SYMBOL(db_close)
+
 static int
-bdb_db_init( BackendDB *be )
+bdb_db_init( BackendDB *be, ConfigReply *cr )
 {
 	struct bdb_info	*bdb;
+	int rc;
 
 	Debug( LDAP_DEBUG_TRACE,
 		LDAP_XSTRING(bdb_db_init) ": Initializing " BDB_UCTYPE " database\n",
@@ -59,6 +65,7 @@ bdb_db_init( BackendDB *be )
 	bdb->bi_dbenv_mode = SLAPD_DEFAULT_DB_MODE;
 
 	bdb->bi_cache.c_maxsize = DEFAULT_CACHE_SIZE;
+	bdb->bi_cache.c_minfree = 1;
 
 	bdb->bi_lock_detect = DB_LOCK_DEFAULT;
 	bdb->bi_search_stack_depth = DEFAULT_SEARCH_STACK_DEPTH;
@@ -66,205 +73,59 @@ bdb_db_init( BackendDB *be )
 
 	ldap_pvt_thread_mutex_init( &bdb->bi_database_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_lastid_mutex );
-	ldap_pvt_thread_mutex_init( &bdb->bi_cache.lru_mutex );
+#ifdef BDB_HIER
+	ldap_pvt_thread_mutex_init( &bdb->bi_modrdns_mutex );
+#endif
+	ldap_pvt_thread_mutex_init( &bdb->bi_cache.c_lru_mutex );
+	ldap_pvt_thread_mutex_init( &bdb->bi_cache.c_count_mutex );
+	ldap_pvt_thread_mutex_init( &bdb->bi_cache.c_eifree_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_cache.c_dntree.bei_kids_mutex );
 	ldap_pvt_thread_rdwr_init ( &bdb->bi_cache.c_rwlock );
+	ldap_pvt_thread_rdwr_init( &bdb->bi_idl_tree_rwlock );
+	ldap_pvt_thread_mutex_init( &bdb->bi_idl_tree_lrulock );
 
 	be->be_private = bdb;
 	be->be_cf_ocs = be->bd_info->bi_cf_ocs;
 
-	return 0;
+	rc = bdb_monitor_db_init( be );
+
+	return rc;
 }
 
-/*
- * Unconditionally perform a database recovery. Only works on
- * databases that were previously opened with transactions and
- * logs enabled.
- */
 static int
-bdb_do_recovery( BackendDB *be )
-{
-	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
-	DB_ENV  *re_dbenv;
-	u_int32_t flags;
-	int		rc;
-	char	path[MAXPATHLEN], *ptr;
-
-	/* Create and init the recovery environment */
-	rc = db_env_create( &re_dbenv, 0 );
-	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-			"bdb_do_recovery: db_env_create failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
-	}
-	re_dbenv->set_errpfx( re_dbenv, be->be_suffix[0].bv_val );
-	re_dbenv->set_errcall( re_dbenv, bdb_errcall );
-	(void)re_dbenv->set_verbose(re_dbenv, DB_VERB_RECOVERY, 1);
-#if DB_VERSION_FULL < 0x04030000
-	(void)re_dbenv->set_verbose(re_dbenv, DB_VERB_CHKPOINT, 1);
-#else
-	re_dbenv->set_msgcall( re_dbenv, bdb_msgcall );
-#endif
-
-	flags = DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL |
-		DB_INIT_TXN | DB_USE_ENVIRON | DB_RECOVER;
-
-	/* If a key was set, use shared memory for the BDB environment */
-	if ( bdb->bi_shm_key ) {
-		re_dbenv->set_shm_key( re_dbenv, bdb->bi_shm_key );
-		flags |= DB_SYSTEM_MEM;
-	}
-
-	/* Open the environment, which will also perform the recovery */
-#ifdef HAVE_EBCDIC
-	strcpy( path, bdb->bi_dbenv_home );
-	__atoe( path );
-	rc = re_dbenv->open( re_dbenv,
-		path,
-		flags,
-		bdb->bi_dbenv_mode );
-#else
-	rc = re_dbenv->open( re_dbenv,
-		bdb->bi_dbenv_home,
-		flags,
-		bdb->bi_dbenv_mode );
-#endif
-	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-			"bdb_do_recovery: dbenv_open failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
-	}
-	(void) re_dbenv->close( re_dbenv, 0 );
-
-	/* By convention we reset the mtime for id2entry.bdb to the current time */
-	ptr = lutil_strcopy( path, bdb->bi_dbenv_home);
-	*ptr++ = LDAP_DIRSEP[0];
-	strcpy( ptr, bdbi_databases[0].file);
-	(void) utime( path, NULL);
-
-	return 0;
-}
-
-/*
- * Database recovery logic:
- * This function is called whenever the database appears to have been
- * shut down uncleanly, as determined by the alock functions. 
- * Because of the -q function in slapadd, there is also the possibility
- * that the shutdown happened when transactions weren't being used and
- * the database is likely to be corrupt. The function checks for this
- * condition by examining the environment to make sure it had previously
- * been opened with transactions enabled. If this is the case, the
- * database is recovered as usual. If transactions were not enabled,
- * then this function will return a fail.
- */
-static int
-bdb_db_recover( BackendDB *be )
-{
-	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
-	DB_ENV  *re_dbenv;
-	u_int32_t flags;
-	int		rc;
-#ifdef HAVE_EBCDIC
-	char	path[MAXPATHLEN];
-#endif
-
-	/* Create the recovery environment, then open it.
-	 * We use the DB_JOIN in combination with a flags value of
-	 * zero so we join an existing environment and can read the
-	 * value of the flags that were used the last time the 
-	 * environment was opened. DB_CREATE is added because the
-	 * open would fail if the only thing that had been done
-	 * was an open with transactions and logs disabled.
-	 */
-	rc = db_env_create( &re_dbenv, 0 );
-	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_recover: db_env_create failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
-	}
-	re_dbenv->set_errpfx( re_dbenv, be->be_suffix[0].bv_val );
-	re_dbenv->set_errcall( re_dbenv, bdb_errcall );
-
-	Debug( LDAP_DEBUG_TRACE,
-		"bdb_db_recover: dbenv_open(%s)\n",
-		bdb->bi_dbenv_home, 0, 0);
-
-#ifdef HAVE_EBCDIC
-	strcpy( path, bdb->bi_dbenv_home );
-	__atoe( path );
-	rc = re_dbenv->open( re_dbenv,
-		path,
-		DB_JOINENV,
-		bdb->bi_dbenv_mode );
-#else
-	rc = re_dbenv->open( re_dbenv,
-		bdb->bi_dbenv_home,
-		DB_JOINENV,
-		bdb->bi_dbenv_mode );
-#endif
-
-	if( rc == ENOENT ) {
-		goto re_exit;
-	}
-	else if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_recover: dbenv_open failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
-	}
-
-	/*
-	 * Check the flags that had been used in the previous open.
-	 * The environment needed to have had both
-	 * DB_INIT_LOG and DB_INIT_TXN set for us to be willing to
-	 * recover the database. Otherwise the an app failed while running
-	 * without transactions and logs enabled and the dn2id and id2entry
-	 * mapping is likely to be corrupt.
-	 */
-	rc = re_dbenv->get_open_flags( re_dbenv, &flags );
-	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_recover: get_open_flags failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
-	}
-
-	(void) re_dbenv->close( re_dbenv, 0 );
-
-	if( (flags & DB_INIT_LOG) && (flags & DB_INIT_TXN) ) {
-		return bdb_do_recovery( be );
-	}
-
-re_exit:
-	Debug( LDAP_DEBUG_ANY,
-		"bdb_db_recover: Database cannot be recovered. "\
-		"Restore from backup!\n", 0, 0, 0);
-	return -1;
-
-}
-
+bdb_db_close( BackendDB *be, ConfigReply *cr );
 
 static int
-bdb_db_open( BackendDB *be )
+bdb_db_open( BackendDB *be, ConfigReply *cr )
 {
 	int rc, i;
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
 	struct stat stat1, stat2;
 	u_int32_t flags;
 	char path[MAXPATHLEN];
-	char *ptr;
+	char *dbhome;
+	Entry *e = NULL;
+	int do_recover = 0, do_alock_recover = 0;
+	int alockt, quick = 0;
+
+	if ( be->be_suffix == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			LDAP_XSTRING(bdb_db_open) ": need suffix\n",
+			1, 0, 0 );
+		return -1;
+	}
 
 	Debug( LDAP_DEBUG_ARGS,
-		"bdb_db_open: %s\n",
+		LDAP_XSTRING(bdb_db_open) ": %s\n",
 		be->be_suffix[0].bv_val, 0, 0 );
 
 #ifndef BDB_MULTIPLE_SUFFIXES
 	if ( be->be_suffix[1].bv_val ) {
-	Debug( LDAP_DEBUG_ANY,
-		"bdb_db_open: only one suffix allowed\n", 0, 0, 0 );
+		if (cr) {
+			snprintf(cr->msg, sizeof(cr->msg), "only one suffix allowed");
+			Debug( LDAP_DEBUG_ANY,
+				LDAP_XSTRING(bdb_db_open) ": %s\n", cr->msg, 0, 0 );
+		}
 		return -1;
 	}
 #endif
@@ -273,162 +134,223 @@ bdb_db_open( BackendDB *be )
 	rc = stat( bdb->bi_dbenv_home, &stat1 );
 	if( rc !=0 ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: Cannot access database directory %s (%d)\n",
+			LDAP_XSTRING(bdb_db_open) ": Cannot access database directory %s (%d)\n",
 			bdb->bi_dbenv_home, errno, 0 );
-			return -1;
+		return -1;
 	}
-	
+
 	/* Perform database use arbitration/recovery logic */
+	alockt = (slapMode & SLAP_TOOL_READONLY) ? ALOCK_LOCKED : ALOCK_UNIQUE;
+	if ( slapMode & SLAP_TOOL_QUICK ) {
+		alockt |= ALOCK_NOSAVE;
+		quick = 1;
+	}
+
 	rc = alock_open( &bdb->bi_alock_info, 
 				"slapd", 
-				bdb->bi_dbenv_home,
-				slapMode & SLAP_TOOL_READONLY ?
-				ALOCK_LOCKED : ALOCK_UNIQUE );
+				bdb->bi_dbenv_home, alockt );
+
+	/* alockt is TRUE if the existing environment was created in Quick mode */
+	alockt = (rc & ALOCK_NOSAVE) ? 1 : 0;
+	rc &= ~ALOCK_NOSAVE;
 
 	if( rc == ALOCK_RECOVER ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: unclean shutdown detected;"
+			LDAP_XSTRING(bdb_db_open) ": unclean shutdown detected;"
 			" attempting recovery.\n", 
 			0, 0, 0 );
-		if( bdb_db_recover( be ) != 0 ) {
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: DB recovery failed.\n",
-				0, 0, 0 );
-			return -1;
-		}
-		if( alock_recover (&bdb->bi_alock_info) != 0 ) {
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: alock_recover failed\n",
-				0, 0, 0 );
-			return -1;
-		}
-
+		do_alock_recover = 1;
+		do_recover = DB_RECOVER;
 	} else if( rc == ALOCK_BUSY ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: database already in use\n", 
+			LDAP_XSTRING(bdb_db_open) ": database already in use\n", 
 			0, 0, 0 );
 		return -1;
 	} else if( rc != ALOCK_CLEAN ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: alock package is unstable\n", 
+			LDAP_XSTRING(bdb_db_open) ": alock package is unstable\n", 
 			0, 0, 0 );
 		return -1;
 	}
-	
+
 	/*
 	 * The DB_CONFIG file may have changed. If so, recover the
 	 * database so that new settings are put into effect. Also
 	 * note the possible absence of DB_CONFIG in the log.
 	 */
 	if( stat( bdb->bi_db_config_path, &stat1 ) == 0 ) {
-		ptr = lutil_strcopy(path, bdb->bi_dbenv_home);
-		*ptr++ = LDAP_DIRSEP[0];
-		strcpy( ptr, bdbi_databases[0].file);
-		if( stat( path, &stat2 ) == 0 ) {
-			if( stat2.st_mtime <= stat1.st_mtime ) {
-				Debug( LDAP_DEBUG_ANY,
-					"bdb_db_open: DB_CONFIG for suffix %s has changed.\n"
-					"Performing database recovery to activate new settings.\n",
-					be->be_suffix[0].bv_val, 0, 0 );
-				if( bdb_do_recovery( be ) != 0) {
+		if ( !do_recover ) {
+			char *ptr = lutil_strcopy(path, bdb->bi_dbenv_home);
+			*ptr++ = LDAP_DIRSEP[0];
+			strcpy( ptr, "__db.001" );
+			if( stat( path, &stat2 ) == 0 ) {
+				if( stat2.st_mtime < stat1.st_mtime ) {
 					Debug( LDAP_DEBUG_ANY,
-						"bdb_db_open: db recovery failed.\n",
-						0, 0, 0 );
-					return -1;
+						LDAP_XSTRING(bdb_db_open) ": DB_CONFIG for suffix %s has changed.\n"
+						"Performing database recovery to activate new settings.\n",
+						be->be_suffix[0].bv_val, 0, 0 );
+					do_recover = DB_RECOVER;
 				}
 			}
-					
 		}
 	}
 	else {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: Warning - No DB_CONFIG file found "
+			LDAP_XSTRING(bdb_db_open) ": Warning - No DB_CONFIG file found "
 			"in directory %s: (%d)\n"
 			"Expect poor performance for suffix %s.\n",
 			bdb->bi_dbenv_home, errno, be->be_suffix[0].bv_val );
 	}
-		
-	flags = DB_INIT_MPOOL | DB_THREAD | DB_CREATE;
-	if ( !( slapMode & SLAP_TOOL_QUICK ))
-		flags |= DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN;
+
+	/* Always let slapcat run, regardless of environment state.
+	 * This can be used to cause a cache flush after an unclean
+	 * shutdown.
+	 */
+	if ( do_recover && ( slapMode & SLAP_TOOL_READONLY )) {
+		Debug( LDAP_DEBUG_ANY,
+			LDAP_XSTRING(bdb_db_open) ": Recovery skipped in read-only mode. "
+			"Run manual recovery if errors are encountered.\n",
+			0, 0, 0 );
+		do_recover = 0;
+		quick = alockt;
+	}
+
+	/* An existing environment in Quick mode has nothing to recover. */
+	if ( alockt && do_recover ) {
+		Debug( LDAP_DEBUG_ANY,
+			LDAP_XSTRING(bdb_db_open) ": cannot recover, database must be reinitialized.\n", 
+			0, 0, 0 );
+		rc = -1;
+		goto fail;
+	}
 
 	rc = db_env_create( &bdb->bi_dbenv, 0 );
 	if( rc != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: db_env_create failed: %s (%d)\n",
+			LDAP_XSTRING(bdb_db_open) ": db_env_create failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-		return rc;
+		goto fail;
 	}
 
-	/* If a key was set, use shared memory for the BDB environment */
-	if ( bdb->bi_shm_key ) {
-		bdb->bi_dbenv->set_shm_key( bdb->bi_dbenv, bdb->bi_shm_key );
-		flags |= DB_SYSTEM_MEM;
+#ifdef HAVE_EBCDIC
+	strcpy( path, bdb->bi_dbenv_home );
+	__atoe( path );
+	dbhome = path;
+#else
+	dbhome = bdb->bi_dbenv_home;
+#endif
+
+	/* If existing environment is clean but doesn't support
+	 * currently requested modes, remove it.
+	 */
+	if ( !do_recover && ( alockt ^ quick )) {
+shm_retry:
+		rc = bdb->bi_dbenv->remove( bdb->bi_dbenv, dbhome, DB_FORCE );
+		if ( rc ) {
+			Debug( LDAP_DEBUG_ANY,
+				LDAP_XSTRING(bdb_db_open) ": dbenv remove failed: %s (%d)\n",
+				db_strerror(rc), rc, 0 );
+			bdb->bi_dbenv = NULL;
+			goto fail;
+		}
+		rc = db_env_create( &bdb->bi_dbenv, 0 );
+		if( rc != 0 ) {
+			Debug( LDAP_DEBUG_ANY,
+				LDAP_XSTRING(bdb_db_open) ": db_env_create failed: %s (%d)\n",
+				db_strerror(rc), rc, 0 );
+			goto fail;
+		}
 	}
 
 	bdb->bi_dbenv->set_errpfx( bdb->bi_dbenv, be->be_suffix[0].bv_val );
 	bdb->bi_dbenv->set_errcall( bdb->bi_dbenv, bdb_errcall );
+
 	bdb->bi_dbenv->set_lk_detect( bdb->bi_dbenv, bdb->bi_lock_detect );
 
 	/* One long-lived TXN per thread, two TXNs per write op */
 	bdb->bi_dbenv->set_tx_max( bdb->bi_dbenv, connection_pool_max * 3 );
 
-#ifdef SLAP_ZONE_ALLOC
-	if ( bdb->bi_cache.c_maxsize ) {
-		bdb->bi_cache.c_zctx = slap_zn_mem_create(
-								SLAP_ZONE_INITSIZE,
-								SLAP_ZONE_MAXSIZE,
-								SLAP_ZONE_DELTA,
-								SLAP_ZONE_SIZE);
-	}
-#endif
-
-	if ( bdb->bi_idl_cache_max_size ) {
-		bdb->bi_idl_tree = NULL;
-		ldap_pvt_thread_rdwr_init( &bdb->bi_idl_tree_rwlock );
-		ldap_pvt_thread_mutex_init( &bdb->bi_idl_tree_lrulock );
-		bdb->bi_idl_cache_size = 0;
-	}
-
 	if( bdb->bi_dbenv_xflags != 0 ) {
 		rc = bdb->bi_dbenv->set_flags( bdb->bi_dbenv,
 			bdb->bi_dbenv_xflags, 1);
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: dbenv_set_flags failed: %s (%d)\n",
+				LDAP_XSTRING(bdb_db_open) ": dbenv_set_flags failed: %s (%d)\n",
 				db_strerror(rc), rc, 0 );
-			return rc;
+			goto fail;
 		}
 	}
 
+#define	BDB_TXN_FLAGS	(DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN)
+
 	Debug( LDAP_DEBUG_TRACE,
-		"bdb_db_open: dbenv_open(%s)\n",
+		LDAP_XSTRING(bdb_db_open) ": dbenv_open(%s)\n",
 		bdb->bi_dbenv_home, 0, 0);
 
-#ifdef HAVE_EBCDIC
-	strcpy( path, bdb->bi_dbenv_home );
-	__atoe( path );
-	rc = bdb->bi_dbenv->open( bdb->bi_dbenv,
-		path,
-		flags,
-		bdb->bi_dbenv_mode );
-#else
-	rc = bdb->bi_dbenv->open( bdb->bi_dbenv,
-		bdb->bi_dbenv_home,
-		flags,
-		bdb->bi_dbenv_mode );
-#endif
-	if( rc != 0 ) {
+	flags = DB_INIT_MPOOL | DB_CREATE | DB_THREAD;
+
+	if ( !quick )
+		flags |= BDB_TXN_FLAGS;
+
+	/* If a key was set, use shared memory for the BDB environment */
+	if ( bdb->bi_shm_key ) {
+		bdb->bi_dbenv->set_shm_key( bdb->bi_dbenv, bdb->bi_shm_key );
+		flags |= DB_SYSTEM_MEM;
+	}
+	rc = (bdb->bi_dbenv->open)( bdb->bi_dbenv, dbhome,
+			flags | do_recover, bdb->bi_dbenv_mode );
+
+	if ( rc ) {
+		/* Regular open failed, probably a missing shm environment.
+		 * Start over, do a recovery.
+		 */
+		if ( !do_recover && bdb->bi_shm_key ) {
+			bdb->bi_dbenv->close( bdb->bi_dbenv, 0 );
+			rc = db_env_create( &bdb->bi_dbenv, 0 );
+			if( rc == 0 ) {
+				Debug( LDAP_DEBUG_ANY, LDAP_XSTRING(bdb_db_open)
+					": Shared memory env open failed, assuming stale env\n",
+					0, 0, 0 );
+				goto shm_retry;
+			}
+		}
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: dbenv_open failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
+			LDAP_XSTRING(bdb_db_open) ": Database cannot be %s, err %d. "
+			"Restore from backup!\n",
+				do_recover ? "recovered" : "opened", rc, 0);
+		goto fail;
+	}
+
+	if ( do_alock_recover && alock_recover (&bdb->bi_alock_info) != 0 ) {
+		Debug( LDAP_DEBUG_ANY,
+			LDAP_XSTRING(bdb_db_open) ": alock_recover failed\n",
+			0, 0, 0 );
+		rc = -1;
+		goto fail;
+	}
+
+#ifdef SLAP_ZONE_ALLOC
+	if ( bdb->bi_cache.c_maxsize ) {
+		bdb->bi_cache.c_zctx = slap_zn_mem_create(
+			SLAP_ZONE_INITSIZE, SLAP_ZONE_MAXSIZE,
+			SLAP_ZONE_DELTA, SLAP_ZONE_SIZE);
+	}
+#endif
+
+	/* Default dncache to 2x entrycache */
+	if ( bdb->bi_cache.c_maxsize && !bdb->bi_cache.c_eimax ) {
+		bdb->bi_cache.c_eimax = bdb->bi_cache.c_maxsize * 2;
+	}
+
+	if ( bdb->bi_idl_cache_max_size ) {
+		bdb->bi_idl_tree = NULL;
+		bdb->bi_idl_cache_size = 0;
 	}
 
 	flags = DB_THREAD | bdb->bi_db_opflags;
 
 #ifdef DB_AUTO_COMMIT
-	if ( !( slapMode & SLAP_TOOL_QUICK ))
+	if ( !quick )
 		flags |= DB_AUTO_COMMIT;
 #endif
 
@@ -444,12 +366,16 @@ bdb_db_open( BackendDB *be )
 		rc = db_create( &db->bdi_db, bdb->bi_dbenv, 0 );
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: db_create(%s) failed: %s (%d)\n",
+				LDAP_XSTRING(bdb_db_open) ": db_create(%s) failed: %s (%d)\n",
 				bdb->bi_dbenv_home, db_strerror(rc), rc );
-			return rc;
+			goto fail;
 		}
 
 		if( i == BDB_ID2ENTRY ) {
+			if ( slapMode & SLAP_TOOL_MODE )
+				db->bdi_db->mpf->set_priority( db->bdi_db->mpf,
+					DB_PRIORITY_VERY_LOW );
+
 			rc = db->bdi_db->set_pagesize( db->bdi_db,
 				BDB_ID2ENTRY_PAGESIZE );
 			if ( slapMode & SLAP_TOOL_READMAIN ) {
@@ -501,9 +427,10 @@ bdb_db_open( BackendDB *be )
 			snprintf( buf, sizeof(buf), "%s/%s", 
 				bdb->bi_dbenv_home, bdbi_databases[i].file );
 			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: db_open(%s) failed: %s (%d)\n",
+				LDAP_XSTRING(bdb_db_open) ": db_open(%s) failed: %s (%d)\n",
 				buf, db_strerror(rc), rc );
-			return rc;
+			db->bdi_db->close( db->bdi_db, 0 );
+			goto fail;
 		}
 
 		flags &= ~(DB_CREATE | DB_RDONLY);
@@ -518,28 +445,73 @@ bdb_db_open( BackendDB *be )
 	rc = bdb_last_id( be, NULL );
 	if( rc != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: last_id(%s) failed: %s (%d)\n",
+			LDAP_XSTRING(bdb_db_open) ": last_id(%s) failed: %s (%d)\n",
 			bdb->bi_dbenv_home, db_strerror(rc), rc );
-		return rc;
+		goto fail;
 	}
 
-	if ( !( slapMode & SLAP_TOOL_QUICK )) {
+	if ( !quick ) {
+#if DB_VERSION_FULL >= 0x04060012
+		u_int32_t lid;
+		XLOCK_ID(bdb->bi_dbenv, &lid);
+		__lock_getlocker(bdb->bi_dbenv->lk_handle, lid, 0, &bdb->bi_cache.c_locker);
+#else
 		XLOCK_ID(bdb->bi_dbenv, &bdb->bi_cache.c_locker);
+#endif
+	}
+
+	entry_prealloc( bdb->bi_cache.c_maxsize );
+	attr_prealloc( bdb->bi_cache.c_maxsize * 20 );
+
+	/* setup for empty-DN contexts */
+	if ( BER_BVISEMPTY( &be->be_nsuffix[0] )) {
+		rc = bdb_id2entry( be, NULL, 0, 0, &e );
+	}
+	if ( !e ) {
+		e = entry_alloc();
+		e->e_id = 0;
+		ber_dupbv( &e->e_name, (struct berval *)&slap_empty_bv );
+		ber_dupbv( &e->e_nname, (struct berval *)&slap_empty_bv );
+	}
+	e->e_ocflags = SLAP_OC_GLUE|SLAP_OC__END;
+	e->e_private = &bdb->bi_cache.c_dntree;
+	bdb->bi_cache.c_dntree.bei_e = e;
+
+	/* monitor setup */
+	rc = bdb_monitor_db_open( be );
+	if ( rc != 0 ) {
+		goto fail;
 	}
 
 	bdb->bi_flags |= BDB_IS_OPEN;
 
 	return 0;
+
+fail:
+	bdb_db_close( be, NULL );
+	return rc;
 }
 
 static int
-bdb_db_close( BackendDB *be )
+bdb_db_close( BackendDB *be, ConfigReply *cr )
 {
 	int rc;
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
 	struct bdb_db_info *db;
 	bdb_idl_cache_entry_t *entry, *next_entry;
 
+	/* monitor handling */
+	(void)bdb_monitor_db_close( be );
+
+	{
+		Entry *e = bdb->bi_cache.c_dntree.bei_e;
+		if ( e ) {
+			bdb->bi_cache.c_dntree.bei_e = NULL;
+			e->e_private = NULL;
+			bdb_entry_return( e );
+		}
+	}
+
 	bdb->bi_flags &= ~BDB_IS_OPEN;
 
 	ber_bvarray_free( bdb->bi_db_config );
@@ -558,30 +530,35 @@ bdb_db_close( BackendDB *be )
 
 	bdb_cache_release_all (&bdb->bi_cache);
 
-	if ( bdb->bi_idl_cache_max_size ) {
-		ldap_pvt_thread_rdwr_wlock ( &bdb->bi_idl_tree_rwlock );
+	if ( bdb->bi_idl_cache_size ) {
 		avl_free( bdb->bi_idl_tree, NULL );
 		bdb->bi_idl_tree = NULL;
 		entry = bdb->bi_idl_lru_head;
-		while ( entry != NULL ) {
+		do {
 			next_entry = entry->idl_lru_next;
 			if ( entry->idl )
 				free( entry->idl );
 			free( entry->kstr.bv_val );
 			free( entry );
 			entry = next_entry;
-		}
+		} while ( entry != bdb->bi_idl_lru_head );
 		bdb->bi_idl_lru_head = bdb->bi_idl_lru_tail = NULL;
-		ldap_pvt_thread_rdwr_wunlock ( &bdb->bi_idl_tree_rwlock );
-	}
-
-	if ( !( slapMode & SLAP_TOOL_QUICK ) && bdb->bi_dbenv ) {
-		XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker);
-		bdb->bi_cache.c_locker = 0;
 	}
 
 	/* close db environment */
 	if( bdb->bi_dbenv ) {
+		/* Free cache locker if we enabled locking */
+		if ( !( slapMode & SLAP_TOOL_QUICK )) {
+#if DB_VERSION_FULL >= 0x04060012
+			XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker->id);
+#else
+			XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker);
+#endif
+			bdb->bi_cache.c_locker = 0;
+		}
+#ifdef BDB_REUSE_LOCKERS
+		bdb_locker_flush( bdb->bi_dbenv );
+#endif
 		/* force a checkpoint, but not if we were ReadOnly,
 		 * and not in Quick mode since there are no transactions there.
 		 */
@@ -615,25 +592,30 @@ bdb_db_close( BackendDB *be )
 }
 
 static int
-bdb_db_destroy( BackendDB *be )
+bdb_db_destroy( BackendDB *be, ConfigReply *cr )
 {
-	int rc;
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
 
+	/* monitor handling */
+	(void)bdb_monitor_db_destroy( be );
+
 	if( bdb->bi_dbenv_home ) ch_free( bdb->bi_dbenv_home );
 	if( bdb->bi_db_config_path ) ch_free( bdb->bi_db_config_path );
 
-	bdb_attr_index_destroy( bdb->bi_attrs );
+	bdb_attr_index_destroy( bdb );
 
 	ldap_pvt_thread_rdwr_destroy ( &bdb->bi_cache.c_rwlock );
-	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.lru_mutex );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_lru_mutex );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_count_mutex );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_eifree_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_dntree.bei_kids_mutex );
+#ifdef BDB_HIER
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_modrdns_mutex );
+#endif
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_lastid_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_database_mutex );
-	if ( bdb->bi_idl_cache_max_size ) {
-		ldap_pvt_thread_rdwr_destroy( &bdb->bi_idl_tree_rwlock );
-		ldap_pvt_thread_mutex_destroy( &bdb->bi_idl_tree_lrulock );
-	}
+	ldap_pvt_thread_rdwr_destroy( &bdb->bi_idl_tree_rwlock );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_idl_tree_lrulock );
 
 	ch_free( bdb );
 	be->be_private = NULL;
@@ -652,11 +634,12 @@ bdb_back_initialize(
 		LDAP_CONTROL_MANAGEDSAIT,
 		LDAP_CONTROL_NOOP,
 		LDAP_CONTROL_PAGEDRESULTS,
-#ifdef LDAP_CONTROL_SUBENTRIES
+		LDAP_CONTROL_PRE_READ,
+		LDAP_CONTROL_POST_READ,
 		LDAP_CONTROL_SUBENTRIES,
-#endif
-#ifdef LDAP_CONTROL_X_PERMISSIVE_MODIFY
 		LDAP_CONTROL_X_PERMISSIVE_MODIFY,
+#ifdef LDAP_X_TXN
+		LDAP_CONTROL_X_TXN_SPEC,
 #endif
 		NULL
 	};
@@ -668,9 +651,7 @@ bdb_back_initialize(
 
 	bi->bi_flags |=
 		SLAP_BFLAG_INCREMENT |
-#ifdef BDB_SUBENTRIES
 		SLAP_BFLAG_SUBENTRIES |
-#endif
 		SLAP_BFLAG_ALIASES |
 		SLAP_BFLAG_REFERRALS;
 
@@ -757,7 +738,6 @@ bdb_back_initialize(
 	bi->bi_tool_entry_reindex = bdb_tool_entry_reindex;
 	bi->bi_tool_sync = 0;
 	bi->bi_tool_dn2id_get = bdb_tool_dn2id_get;
-	bi->bi_tool_id2entry_get = bdb_tool_id2entry_get;
 	bi->bi_tool_entry_modify = bdb_tool_entry_modify;
 
 	bi->bi_connection_init = 0;
