@@ -1,8 +1,8 @@
 /* map.c - ldap backend mapping routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/map.c,v 1.1.2.3 2005/01/24 21:28:53 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/map.c,v 1.15.2.4 2007/08/31 23:14:03 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -87,22 +87,22 @@ ldap_back_map_init ( struct ldapmap *lm, struct ldapmapping **m )
 {
 	struct ldapmapping *mapping;
 
-	assert( m );
+	assert( m != NULL );
 
 	*m = NULL;
-	
+
 	mapping = (struct ldapmapping *)ch_calloc( 2, 
 			sizeof( struct ldapmapping ) );
 	if ( mapping == NULL ) {
 		return;
 	}
 
-	ber_str2bv( "objectclass", sizeof("objectclass")-1, 1, &mapping->src);
-	ber_dupbv( &mapping->dst, &mapping->src );
-	mapping[1].src = mapping->src;
-	mapping[1].dst = mapping->dst;
+	ber_str2bv( "objectclass", STRLENOF("objectclass"), 1, &mapping[0].src);
+	ber_dupbv( &mapping[0].dst, &mapping[0].src );
+	mapping[1].src = mapping[0].src;
+	mapping[1].dst = mapping[0].dst;
 
-	avl_insert( &lm->map, (caddr_t)mapping, 
+	avl_insert( &lm->map, (caddr_t)&mapping[0], 
 			mapping_cmp, mapping_dup );
 	avl_insert( &lm->remap, (caddr_t)&mapping[1], 
 			mapping_cmp, mapping_dup );
@@ -116,10 +116,11 @@ ldap_back_mapping ( struct ldapmap *map, struct berval *s, struct ldapmapping **
 	Avlnode *tree;
 	struct ldapmapping fmapping;
 
-	assert( m );
+	assert( m != NULL );
 
 	if ( remap == BACKLDAP_REMAP ) {
 		tree = map->remap;
+
 	} else {
 		tree = map->map;
 	}
@@ -139,6 +140,13 @@ ldap_back_map ( struct ldapmap *map, struct berval *s, struct berval *bv,
 {
 	struct ldapmapping *mapping;
 
+	/* map->map may be NULL when mapping is configured,
+	 * but map->remap can't */
+	if ( map->remap == NULL ) {
+		*bv = *s;
+		return;
+	}
+
 	BER_BVZERO( bv );
 	( void )ldap_back_mapping( map, s, &mapping, remap );
 	if ( mapping != NULL ) {
@@ -206,15 +214,22 @@ map_attr_value(
 	struct berval		vtmp;
 	int			freeval = 0;
 
-	ldap_back_map( &dc->rwmap->rwm_at, &ad->ad_cname, mapped_attr, remap );
+	ldap_back_map( &dc->target->mt_rwmap.rwm_at, &ad->ad_cname, mapped_attr, remap );
 	if ( BER_BVISNULL( mapped_attr ) || BER_BVISEMPTY( mapped_attr ) ) {
+#if 0
 		/*
 		 * FIXME: are we sure we need to search oc_map if at_map fails?
 		 */
-		ldap_back_map( &dc->rwmap->rwm_oc, &ad->ad_cname, mapped_attr, remap );
+		ldap_back_map( &dc->target->mt_rwmap.rwm_oc, &ad->ad_cname, mapped_attr, remap );
 		if ( BER_BVISNULL( mapped_attr ) || BER_BVISEMPTY( mapped_attr ) ) {
 			*mapped_attr = ad->ad_cname;
 		}
+#endif
+		if ( dc->target->mt_rwmap.rwm_at.drop_missing ) {
+			return -1;
+		}
+
+		*mapped_attr = ad->ad_cname;
 	}
 
 	if ( value == NULL ) {
@@ -244,7 +259,7 @@ map_attr_value(
 		}
 
 	} else if ( ad == slap_schema.si_ad_objectClass || ad == slap_schema.si_ad_structuralObjectClass ) {
-		ldap_back_map( &dc->rwmap->rwm_oc, value, &vtmp, remap );
+		ldap_back_map( &dc->target->mt_rwmap.rwm_oc, value, &vtmp, remap );
 		if ( BER_BVISNULL( &vtmp ) || BER_BVISEMPTY( &vtmp ) ) {
 			vtmp = *value;
 		}
@@ -271,13 +286,31 @@ ldap_back_int_filter_map_rewrite(
 {
 	int		i;
 	Filter		*p;
-	struct berval	atmp;
-	struct berval	vtmp;
+	struct berval	atmp,
+			vtmp,
+			*tmp;
+	static struct berval
+			/* better than nothing... */
+			ber_bvfalse = BER_BVC( "(!(objectClass=*))" ),
+			ber_bvtf_false = BER_BVC( "(|)" ),
+			/* better than nothing... */
+			ber_bvtrue = BER_BVC( "(objectClass=*)" ),
+			ber_bvtf_true = BER_BVC( "(&)" ),
+#if 0
+			/* no longer needed; preserved for completeness */
+			ber_bvundefined = BER_BVC( "(?=undefined)" ),
+#endif
+			ber_bverror = BER_BVC( "(?=error)" ),
+			ber_bvunknown = BER_BVC( "(?=unknown)" ),
+			ber_bvnone = BER_BVC( "(?=none)" );
 	ber_len_t	len;
 
+	assert( fstr != NULL );
+	BER_BVZERO( fstr );
+
 	if ( f == NULL ) {
-		ber_str2bv( "No filter!", sizeof("No filter!")-1, 1, fstr );
-		return -1;
+		ber_dupbv( fstr, &ber_bvnone );
+		return LDAP_OTHER;
 	}
 
 	switch ( f->f_choice ) {
@@ -285,7 +318,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_av_desc, &atmp,
 					&f->f_av_value, &vtmp, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		fstr->bv_len = atmp.bv_len + vtmp.bv_len
@@ -293,7 +326,7 @@ ldap_back_int_filter_map_rewrite(
 		fstr->bv_val = malloc( fstr->bv_len + 1 );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=%s)",
-			atmp.bv_val, vtmp.bv_val );
+			atmp.bv_val, vtmp.bv_len ? vtmp.bv_val : "" );
 
 		ber_memfree( vtmp.bv_val );
 		break;
@@ -302,7 +335,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_av_desc, &atmp,
 					&f->f_av_value, &vtmp, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		fstr->bv_len = atmp.bv_len + vtmp.bv_len
@@ -310,7 +343,7 @@ ldap_back_int_filter_map_rewrite(
 		fstr->bv_val = malloc( fstr->bv_len + 1 );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s>=%s)",
-			atmp.bv_val, vtmp.bv_val );
+			atmp.bv_val, vtmp.bv_len ? vtmp.bv_val : "" );
 
 		ber_memfree( vtmp.bv_val );
 		break;
@@ -319,7 +352,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_av_desc, &atmp,
 					&f->f_av_value, &vtmp, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		fstr->bv_len = atmp.bv_len + vtmp.bv_len
@@ -327,7 +360,7 @@ ldap_back_int_filter_map_rewrite(
 		fstr->bv_val = malloc( fstr->bv_len + 1 );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s<=%s)",
-			atmp.bv_val, vtmp.bv_val );
+			atmp.bv_val, vtmp.bv_len ? vtmp.bv_val : "" );
 
 		ber_memfree( vtmp.bv_val );
 		break;
@@ -336,7 +369,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_av_desc, &atmp,
 					&f->f_av_value, &vtmp, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		fstr->bv_len = atmp.bv_len + vtmp.bv_len
@@ -344,7 +377,7 @@ ldap_back_int_filter_map_rewrite(
 		fstr->bv_val = malloc( fstr->bv_len + 1 );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s~=%s)",
-			atmp.bv_val, vtmp.bv_val );
+			atmp.bv_val, vtmp.bv_len ? vtmp.bv_val : "" );
 
 		ber_memfree( vtmp.bv_val );
 		break;
@@ -353,7 +386,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_sub_desc, &atmp,
 					NULL, NULL, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		/* cannot be a DN ... */
@@ -374,7 +407,7 @@ ldap_back_int_filter_map_rewrite(
 
 			snprintf( &fstr->bv_val[len - 2], vtmp.bv_len + 3,
 				/* "(attr=" */ "%s*)",
-				vtmp.bv_val );
+				vtmp.bv_len ? vtmp.bv_val : "" );
 
 			ber_memfree( vtmp.bv_val );
 		}
@@ -389,7 +422,7 @@ ldap_back_int_filter_map_rewrite(
 
 				snprintf( &fstr->bv_val[len - 1], vtmp.bv_len + 3,
 					/* "(attr=[init]*[any*]" */ "%s*)",
-					vtmp.bv_val );
+					vtmp.bv_len ? vtmp.bv_val : "" );
 				ber_memfree( vtmp.bv_val );
 			}
 		}
@@ -404,7 +437,7 @@ ldap_back_int_filter_map_rewrite(
 
 			snprintf( &fstr->bv_val[len - 1], vtmp.bv_len + 3,
 				/* "(attr=[init*][any*]" */ "%s)",
-				vtmp.bv_val );
+				vtmp.bv_len ? vtmp.bv_val : "" );
 
 			ber_memfree( vtmp.bv_val );
 		}
@@ -415,7 +448,7 @@ ldap_back_int_filter_map_rewrite(
 		if ( map_attr_value( dc, f->f_desc, &atmp,
 					NULL, NULL, remap ) )
 		{
-			return -1;
+			goto computed;
 		}
 
 		fstr->bv_len = atmp.bv_len + ( STRLENOF( "(=*)" ) );
@@ -436,30 +469,32 @@ ldap_back_int_filter_map_rewrite(
 			f->f_choice == LDAP_FILTER_OR ? '|' : '!' );
 
 		for ( p = f->f_list; p != NULL; p = p->f_next ) {
+			int	rc;
+
 			len = fstr->bv_len;
 
-			if ( ldap_back_int_filter_map_rewrite( dc, p, &vtmp, remap ) )
-			{
-				return -1;
+			rc = ldap_back_int_filter_map_rewrite( dc, p, &vtmp, remap );
+			if ( rc != LDAP_SUCCESS ) {
+				return rc;
 			}
 			
 			fstr->bv_len += vtmp.bv_len;
 			fstr->bv_val = ch_realloc( fstr->bv_val, fstr->bv_len + 1 );
 
 			snprintf( &fstr->bv_val[len-1], vtmp.bv_len + 2, 
-				/*"("*/ "%s)", vtmp.bv_val );
+				/*"("*/ "%s)", vtmp.bv_len ? vtmp.bv_val : "" );
 
 			ch_free( vtmp.bv_val );
 		}
 
 		break;
 
-	case LDAP_FILTER_EXT: {
+	case LDAP_FILTER_EXT:
 		if ( f->f_mr_desc ) {
 			if ( map_attr_value( dc, f->f_mr_desc, &atmp,
 						&f->f_mr_value, &vtmp, remap ) )
 			{
-				return -1;
+				goto computed;
 			}
 
 		} else {
@@ -479,29 +514,42 @@ ldap_back_int_filter_map_rewrite(
 			f->f_mr_dnattrs ? ":dn" : "",
 			!BER_BVISEMPTY( &f->f_mr_rule_text ) ? ":" : "",
 			!BER_BVISEMPTY( &f->f_mr_rule_text ) ? f->f_mr_rule_text.bv_val : "",
-			vtmp.bv_val );
+			vtmp.bv_len ? vtmp.bv_val : "" );
 		ber_memfree( vtmp.bv_val );
-		} break;
+		break;
 
 	case SLAPD_FILTER_COMPUTED:
 		switch ( f->f_result ) {
 		case LDAP_COMPARE_FALSE:
-			ber_str2bv( "(?=false)", STRLENOF( "(?=false)" ), 1, fstr );
-			break;
-		case LDAP_COMPARE_TRUE:
-			ber_str2bv( "(?=true)", STRLENOF( "(?=true)" ), 1, fstr );
-			break;
+		/* FIXME: treat UNDEFINED as FALSE */
 		case SLAPD_COMPARE_UNDEFINED:
-			ber_str2bv( "(?=undefined)", STRLENOF( "(?=undefined)" ), 1, fstr );
+computed:;
+			if ( META_BACK_TGT_T_F( dc->target ) ) {
+				tmp = &ber_bvtf_false;
+				break;
+			}
+			tmp = &ber_bvfalse;
 			break;
+
+		case LDAP_COMPARE_TRUE:
+			if ( META_BACK_TGT_T_F( dc->target ) ) {
+				tmp = &ber_bvtf_true;
+				break;
+			}
+
+			tmp = &ber_bvtrue;
+			break;
+
 		default:
-			ber_str2bv( "(?=error)", STRLENOF( "(?=error)" ), 1, fstr );
+			tmp = &ber_bverror;
 			break;
 		}
+
+		ber_dupbv( fstr, tmp );
 		break;
 
 	default:
-		ber_str2bv( "(?=unknown)", STRLENOF( "(?=unknown)" ), 1, fstr );
+		ber_dupbv( fstr, &ber_bvunknown );
 		break;
 	}
 
@@ -518,6 +566,7 @@ ldap_back_filter_map_rewrite(
 	int		rc;
 	dncookie	fdc;
 	struct berval	ftmp;
+	static char	*dmy = "";
 
 	rc = ldap_back_int_filter_map_rewrite( dc, f, fstr, remap );
 
@@ -531,8 +580,8 @@ ldap_back_filter_map_rewrite(
 
 	fdc.ctx = "searchFilter";
 	
-	switch ( rewrite_session( fdc.rwmap->rwm_rw, fdc.ctx,
-				( !BER_BVISEMPTY( &ftmp ) ? ftmp.bv_val : "" ),
+	switch ( rewrite_session( fdc.target->mt_rwmap.rwm_rw, fdc.ctx,
+				( !BER_BVISEMPTY( &ftmp ) ? ftmp.bv_val : dmy ),
 				fdc.conn, &fstr->bv_val ) )
 	{
 	case REWRITE_REGEXEC_OK:
@@ -544,7 +593,8 @@ ldap_back_filter_map_rewrite(
 		}
 		Debug( LDAP_DEBUG_ARGS,
 			"[rw] %s: \"%s\" -> \"%s\"\n",
-			fdc.ctx, ftmp.bv_val, fstr->bv_val );		
+			fdc.ctx, BER_BVISNULL( &ftmp ) ? "" : ftmp.bv_val,
+			BER_BVISNULL( fstr ) ? "" : fstr->bv_val );		
 		rc = LDAP_SUCCESS;
 		break;
  		
@@ -564,6 +614,10 @@ ldap_back_filter_map_rewrite(
 		rc = LDAP_OTHER;
 		break;
 	}
+
+	if ( fstr->bv_val == dmy ) {
+		BER_BVZERO( fstr );
+	}
 #endif /* ENABLE_REWRITE */
 
 	return rc;
@@ -577,15 +631,16 @@ ldap_back_referral_result_rewrite(
 {
 	int		i, last;
 
-	assert( dc );
-	assert( a_vals );
+	assert( dc != NULL );
+	assert( a_vals != NULL );
 
 	for ( last = 0; !BER_BVISNULL( &a_vals[ last ] ); last++ )
 		;
 	last--;
 
 	for ( i = 0; !BER_BVISNULL( &a_vals[ i ] ); i++ ) {
-		struct berval	dn, olddn;
+		struct berval	dn,
+				olddn = BER_BVNULL;
 		int		rc;
 		LDAPURLDesc	*ludp;
 
@@ -630,6 +685,7 @@ ldap_back_referral_result_rewrite(
 
 				ludp->lud_dn = dn.bv_val;
 				newurl = ldap_url_desc2str( ludp );
+				free( dn.bv_val );
 				if ( newurl == NULL ) {
 					/* FIXME: leave attr untouched
 					 * even if ldap_url_desc2str failed...
