@@ -1,8 +1,8 @@
 /* database.c - deals with database subsystem */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/database.c,v 1.61.2.4 2005/05/06 16:42:52 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/database.c,v 1.80.2.5 2007/08/31 23:14:03 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2005 The OpenLDAP Foundation.
+ * Copyright 2001-2007 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -33,19 +33,6 @@
 static int monitor_back_add_plugin( monitor_info_t *mi, Backend *be, Entry *e );
 #endif /* defined(LDAP_SLAPI) */
 
-#if defined(SLAPD_BDB)
-#include "../back-bdb/back-bdb.h"
-#endif /* defined(SLAPD_BDB) */
-#if defined(SLAPD_HDB)
-#include "../back-hdb/back-bdb.h"
-#endif /* defined(SLAPD_HDB) */
-#if defined(SLAPD_LDAP) 
-#include "../back-ldap/back-ldap.h"
-#endif /* defined(SLAPD_LDAP) */
-#if 0 && defined(SLAPD_LDBM) 
-#include "../back-ldbm/back-ldbm.h"
-#endif /* defined(SLAPD_LDBM) */
-
 /* for PATH_MAX on some systems (e.g. Solaris) */
 #ifdef HAVE_LIMITS_H
 #include <limits.h>
@@ -54,6 +41,12 @@ static int monitor_back_add_plugin( monitor_info_t *mi, Backend *be, Entry *e );
 #define PATH_MAX	4095
 #endif /* ! PATH_MAX */
 
+static int
+monitor_subsys_database_modify(
+	Operation	*op,
+	SlapReply	*rs,
+	Entry		*e );
+
 static struct restricted_ops_t {
 	struct berval	op;
 	unsigned int	tag;
@@ -70,8 +63,8 @@ static struct restricted_ops_t {
 }, restricted_exops[] = {
 	{ BER_BVC( LDAP_EXOP_START_TLS ),	SLAP_RESTRICT_EXOP_START_TLS },
 	{ BER_BVC( LDAP_EXOP_MODIFY_PASSWD ),	SLAP_RESTRICT_EXOP_MODIFY_PASSWD },
-	{ BER_BVC( LDAP_EXOP_X_WHO_AM_I ),	SLAP_RESTRICT_EXOP_WHOAMI },
-	{ BER_BVC( LDAP_EXOP_X_CANCEL ),	SLAP_RESTRICT_EXOP_CANCEL },
+	{ BER_BVC( LDAP_EXOP_WHO_AM_I ),	SLAP_RESTRICT_EXOP_WHOAMI },
+	{ BER_BVC( LDAP_EXOP_CANCEL ),	SLAP_RESTRICT_EXOP_CANCEL },
 	{ BER_BVNULL,				0 }
 };
 
@@ -114,21 +107,363 @@ init_restrictedOperation( monitor_info_t *mi, Entry *e, slap_mask_t restrictops
 	return LDAP_SUCCESS;
 }
 
-int
-monitor_subsys_database_init(
+static int
+monitor_subsys_database_init_one(
+	monitor_info_t		*mi,
 	BackendDB		*be,
-	monitor_subsys_t	*ms
-)
+	monitor_subsys_t	*ms,
+	monitor_subsys_t	*ms_backend,
+	monitor_subsys_t	*ms_overlay,
+	struct berval		*rdn,
+	Entry			*e_database,
+	Entry			***epp )
+{
+	char			buf[ BACKMONITOR_BUFSIZE ];
+	int			j;
+	slap_overinfo		*oi = NULL;
+	BackendInfo		*bi, *bi2;
+	Entry			*e;
+	monitor_entry_t		*mp;
+	char			*rdnval = strchr( rdn->bv_val, '=' ) + 1;
+	struct berval		bv;
+
+	bi = be->bd_info;
+
+	if ( overlay_is_over( be ) ) {
+		oi = (slap_overinfo *)be->bd_info->bi_private;
+		bi = oi->oi_orig;
+	}
+
+	/* Subordinates are not exposed as their own naming context */
+	if ( SLAP_GLUE_SUBORDINATE( be ) ) {
+		return 0;
+	}
+
+	e = monitor_entry_stub( &ms->mss_dn, &ms->mss_ndn, rdn,
+		mi->mi_oc_monitoredObject, mi, NULL, NULL );
+
+	if ( e == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_database_init: "
+			"unable to create entry \"%s,%s\"\n",
+			rdn->bv_val, ms->mss_dn.bv_val, 0 );
+		return( -1 );
+	}
+
+	ber_str2bv( bi->bi_type, 0, 0, &bv );
+	attr_merge_one( e, mi->mi_ad_monitoredInfo, &bv, NULL );
+	attr_merge_one( e, mi->mi_ad_monitorIsShadow,
+		SLAP_SHADOW( be ) ? (struct berval *)&slap_true_bv :
+			(struct berval *)&slap_false_bv, NULL );
+
+	if ( SLAP_MONITOR( be ) ) {
+		attr_merge( e, slap_schema.si_ad_monitorContext,
+				be->be_suffix, be->be_nsuffix );
+		attr_merge( e_database, slap_schema.si_ad_monitorContext,
+				be->be_suffix, be->be_nsuffix );
+
+	} else {
+		if ( be->be_suffix == NULL ) {
+			Debug( LDAP_DEBUG_ANY,
+				"monitor_subsys_database_init: "
+				"missing suffix for %s\n",
+				rdnval, 0, 0 );
+			return -1;
+		}
+		attr_merge( e, slap_schema.si_ad_namingContexts,
+				be->be_suffix, be->be_nsuffix );
+		attr_merge( e_database, slap_schema.si_ad_namingContexts,
+				be->be_suffix, be->be_nsuffix );
+	}
+
+	(void)init_readOnly( mi, e, be->be_restrictops );
+	(void)init_restrictedOperation( mi, e, be->be_restrictops );
+
+	if ( SLAP_SHADOW( be ) && be->be_update_refs ) {
+		attr_merge_normalize( e, mi->mi_ad_monitorUpdateRef,
+				be->be_update_refs, NULL );
+	}
+
+	if ( oi != NULL ) {
+		slap_overinst	*on = oi->oi_list,
+				*on1 = on;
+
+		for ( ; on; on = on->on_next ) {
+			slap_overinst		*on2;
+
+			for ( on2 = on1; on2 != on; on2 = on2->on_next ) {
+				if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
+					break;
+				}
+			}
+
+			if ( on2 != on ) {
+				break;
+			}
+			
+			ber_str2bv( on->on_bi.bi_type, 0, 0, &bv );
+			attr_merge_normalize_one( e, mi->mi_ad_monitorOverlay,
+					&bv, NULL );
+
+			/* find the overlay number, j */
+			for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
+				if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
+					break;
+				}
+			}
+			assert( on2 != NULL );
+
+			snprintf( buf, sizeof( buf ), 
+				"cn=Overlay %d,%s", 
+				j, ms_overlay->mss_dn.bv_val );
+			ber_str2bv( buf, 0, 0, &bv );
+			attr_merge_normalize_one( e,
+					slap_schema.si_ad_seeAlso,
+					&bv, NULL );
+		}
+	}
+
+	j = -1;
+	LDAP_STAILQ_FOREACH( bi2, &backendInfo, bi_next ) {
+		j++;
+		if ( bi2->bi_type == bi->bi_type ) {
+			snprintf( buf, sizeof( buf ), 
+				"cn=Backend %d,%s", 
+				j, ms_backend->mss_dn.bv_val );
+			bv.bv_val = buf;
+			bv.bv_len = strlen( buf );
+			attr_merge_normalize_one( e,
+					slap_schema.si_ad_seeAlso,
+					&bv, NULL );
+			break;
+		}
+	}
+	/* we must find it! */
+	assert( j >= 0 );
+
+	mp = monitor_entrypriv_create();
+	if ( mp == NULL ) {
+		return -1;
+	}
+	e->e_private = ( void * )mp;
+	mp->mp_info = ms;
+	mp->mp_flags = ms->mss_flags
+		| MONITOR_F_SUB;
+
+	if ( monitor_cache_add( mi, e ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_database_init: "
+			"unable to add entry \"%s,%s\"\n",
+			rdn->bv_val, ms->mss_dn.bv_val, 0 );
+		return( -1 );
+	}
+
+#if defined(LDAP_SLAPI)
+	monitor_back_add_plugin( mi, be, e );
+#endif /* defined(LDAP_SLAPI) */
+
+	if ( oi != NULL ) {
+		Entry		**ep_overlay = &mp->mp_children;
+		monitor_entry_t	*mp_overlay;
+		slap_overinst	*on = oi->oi_list;
+		int		o;
+
+		for ( o = 0; on; o++, on = on->on_next ) {
+			Entry			*e_overlay;
+			slap_overinst		*on2;
+
+			/* find the overlay number, j */
+			for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
+				if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
+					break;
+				}
+			}
+			assert( on2 != NULL );
+
+			bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d", o );
+			bv.bv_val = buf;
+
+			e_overlay = monitor_entry_stub( &e->e_name, &e->e_nname, &bv,
+				mi->mi_oc_monitoredObject, mi, NULL, NULL );
+
+			if ( e_overlay == NULL ) {
+				Debug( LDAP_DEBUG_ANY,
+					"monitor_subsys_database_init: "
+					"unable to create entry "
+					"\"cn=Overlay %d,%s,%s\"\n",
+					o, rdn->bv_val, ms->mss_dn.bv_val );
+				return( -1 );
+			}
+			ber_str2bv( on->on_bi.bi_type, 0, 0, &bv );
+			attr_merge_one( e_overlay, mi->mi_ad_monitoredInfo, &bv, NULL );
+
+			bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d,%s",
+				j, ms_overlay->mss_dn.bv_val );
+			bv.bv_val = buf;
+			attr_merge_normalize_one( e_overlay, slap_schema.si_ad_seeAlso,
+				&bv, NULL );
+
+			if ( SLAP_MONITOR( be ) ) {
+				attr_merge( e_overlay, slap_schema.si_ad_monitorContext,
+						be->be_suffix, be->be_nsuffix );
+
+			} else {
+				attr_merge( e_overlay, slap_schema.si_ad_namingContexts,
+						be->be_suffix, be->be_nsuffix );
+			}
+
+			mp_overlay = monitor_entrypriv_create();
+			if ( mp_overlay == NULL ) {
+				return -1;
+			}
+			e_overlay->e_private = ( void * )mp_overlay;
+			mp_overlay->mp_info = ms;
+			mp_overlay->mp_flags = ms->mss_flags
+				| MONITOR_F_SUB;
+	
+			if ( monitor_cache_add( mi, e_overlay ) ) {
+				Debug( LDAP_DEBUG_ANY,
+					"monitor_subsys_database_init: "
+					"unable to add entry "
+					"\"cn=Overlay %d,%s,%s\"\n",
+					o, rdn->bv_val, ms->mss_dn.bv_val );
+				return( -1 );
+			}
+
+			*ep_overlay = e_overlay;
+			ep_overlay = &mp_overlay->mp_next;
+		}
+	}
+
+	**epp = e;
+	*epp = &mp->mp_next;
+
+	return 0;
+}
+
+int
+monitor_back_register_database(
+	BackendDB		*be )
 {
 	monitor_info_t		*mi;
 	Entry			*e_database, **ep;
-	int			i;
+	int			i, rc;
+	monitor_entry_t		*mp;
+	monitor_subsys_t	*ms_backend,
+				*ms_database,
+				*ms_overlay;
+	struct berval		bv;
+	char			buf[ BACKMONITOR_BUFSIZE ];
+
+	assert( be_monitor != NULL );
+
+	if ( !monitor_subsys_is_opened() ) {
+		return monitor_back_register_database_limbo( be );
+	}
+
+	mi = ( monitor_info_t * )be_monitor->be_private;
+
+	ms_backend = monitor_back_get_subsys( SLAPD_MONITOR_BACKEND_NAME );
+	if ( ms_backend == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_BACKEND_NAME "\" "
+			"subsystem\n",
+			0, 0, 0 );
+		return -1;
+	}
+
+	ms_database = monitor_back_get_subsys( SLAPD_MONITOR_DATABASE_NAME );
+	if ( ms_database == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_DATABASE_NAME "\" "
+			"subsystem\n",
+			0, 0, 0 );
+		return -1;
+	}
+
+	ms_overlay = monitor_back_get_subsys( SLAPD_MONITOR_OVERLAY_NAME );
+	if ( ms_overlay == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_OVERLAY_NAME "\" "
+			"subsystem\n",
+			0, 0, 0 );
+		return -1;
+	}
+
+	if ( monitor_cache_get( mi, &ms_database->mss_ndn, &e_database ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_database_init: "
+			"unable to get entry \"%s\"\n",
+			ms_database->mss_ndn.bv_val, 0, 0 );
+		return( -1 );
+	}
+
+	mp = ( monitor_entry_t * )e_database->e_private;
+	for ( i = -1, ep = &mp->mp_children; *ep; i++ ) {
+		Attribute	*a;
+
+		a = attr_find( (*ep)->e_attrs, slap_schema.si_ad_namingContexts );
+		if ( a ) {
+			int		j, k;
+
+			for ( j = 0; !BER_BVISNULL( &a->a_nvals[ j ] ); j++ ) {
+				for ( k = 0; !BER_BVISNULL( &be->be_nsuffix[ k ] ); k++ ) {
+					if ( dn_match( &a->a_nvals[ j ], &be->be_nsuffix[ k ] ) ) {
+						rc = 0;
+						goto done;
+					}
+				}
+			}
+		}
+
+		mp = ( monitor_entry_t * )(*ep)->e_private;
+
+		assert( mp != NULL );
+		ep = &mp->mp_next;
+	}
+
+	bv.bv_val = buf;
+	bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Database %d", i );
+	if ( bv.bv_len >= sizeof( buf ) ) {
+		rc = -1;
+		goto done;
+	}
+	
+	rc = monitor_subsys_database_init_one( mi, be,
+		ms_database, ms_backend, ms_overlay, &bv, e_database, &ep );
+	if ( rc != 0 ) {
+		goto done;
+	}
+
+done:;
+	monitor_cache_release( mi, e_database );
+
+	return rc;
+}
+
+int
+monitor_subsys_database_init(
+	BackendDB		*be,
+	monitor_subsys_t	*ms )
+{
+	monitor_info_t		*mi;
+	Entry			*e_database, **ep;
+	int			i, rc;
 	monitor_entry_t		*mp;
 	monitor_subsys_t	*ms_backend,
 				*ms_overlay;
+	struct berval		bv;
 
 	assert( be != NULL );
 
+	ms->mss_modify = monitor_subsys_database_modify;
+
 	mi = ( monitor_info_t * )be->be_private;
 
 	ms_backend = monitor_back_get_subsys( SLAPD_MONITOR_BACKEND_NAME );
@@ -168,325 +503,28 @@ monitor_subsys_database_init(
 	mp->mp_children = NULL;
 	ep = &mp->mp_children;
 
+	BER_BVSTR( &bv, "cn=Frontend" );
+	rc = monitor_subsys_database_init_one( mi, frontendDB,
+		ms, ms_backend, ms_overlay, &bv, e_database, &ep );
+	if ( rc != 0 ) {
+		return rc;
+	}
+
 	i = -1;
 	LDAP_STAILQ_FOREACH( be, &backendDB, be_next ) {
 		char		buf[ BACKMONITOR_BUFSIZE ];
-		int		j;
-		slap_overinfo	*oi = NULL;
-		BackendInfo	*bi, *bi2;
-		Entry		*e;
 
-		i++;
-
-		bi = be->bd_info;
-
-		if ( overlay_is_over( be ) ) {
-			oi = (slap_overinfo *)be->bd_info->bi_private;
-			bi = oi->oi_orig;
-		}
-
-		/* Subordinates are not exposed as their own naming context */
-		if ( SLAP_GLUE_SUBORDINATE( be ) ) {
-			continue;
-		}
-
-		snprintf( buf, sizeof( buf ),
-				"dn: cn=Database %d,%s\n"
-				"objectClass: %s\n"
-				"structuralObjectClass: %s\n"
-				"cn: Database %d\n"
-				"%s: %s\n"
-				"%s: %s\n"
-				"creatorsName: %s\n"
-				"modifiersName: %s\n"
-				"createTimestamp: %s\n"
-				"modifyTimestamp: %s\n",
-				i,
-					ms->mss_dn.bv_val,
-				mi->mi_oc_monitoredObject->soc_cname.bv_val,
-				mi->mi_oc_monitoredObject->soc_cname.bv_val,
-				i,
-				mi->mi_ad_monitoredInfo->ad_cname.bv_val,
-					bi->bi_type,
-				mi->mi_ad_monitorIsShadow->ad_cname.bv_val,
-					SLAP_SHADOW( be ) ? slap_true_bv.bv_val : slap_false_bv.bv_val,
-				mi->mi_creatorsName.bv_val,
-				mi->mi_creatorsName.bv_val,
-				mi->mi_startTime.bv_val,
-				mi->mi_startTime.bv_val );
-		
-		e = str2entry( buf );
-		if ( e == NULL ) {
-			Debug( LDAP_DEBUG_ANY,
-				"monitor_subsys_database_init: "
-				"unable to create entry \"cn=Database %d,%s\"\n",
-				i, ms->mss_dn.bv_val, 0 );
-			return( -1 );
-		}
-		
-		if ( SLAP_MONITOR( be ) ) {
-			attr_merge( e, slap_schema.si_ad_monitorContext,
-					be->be_suffix, be->be_nsuffix );
-			attr_merge( e_database, slap_schema.si_ad_monitorContext,
-					be->be_suffix, be->be_nsuffix );
-
-		} else {
-			if ( be->be_suffix == NULL ) {
-				Debug( LDAP_DEBUG_ANY,
-					"monitor_subsys_database_init: "
-					"missing suffix for database %d\n",
-					i, 0, 0 );
-				return -1;
-			}
-			attr_merge( e, slap_schema.si_ad_namingContexts,
-					be->be_suffix, be->be_nsuffix );
-			attr_merge( e_database, slap_schema.si_ad_namingContexts,
-					be->be_suffix, be->be_nsuffix );
-		}
-
-		(void)init_readOnly( mi, e, be->be_restrictops );
-		(void)init_restrictedOperation( mi, e, be->be_restrictops );
-
-		if ( oi != NULL ) {
-			slap_overinst	*on = oi->oi_list,
-					*on1 = on;
-
-			for ( ; on; on = on->on_next ) {
-				struct berval		bv;
-				slap_overinst		*on2;
-
-				for ( on2 = on1; on2 != on; on2 = on2->on_next ) {
-					if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
-						break;
-					}
-				}
-
-				if ( on2 != on ) {
-					break;
-				}
-				
-				ber_str2bv( on->on_bi.bi_type, 0, 0, &bv );
-				attr_merge_normalize_one( e, mi->mi_ad_monitorOverlay,
-						&bv, NULL );
-
-				/* find the overlay number, j */
-				for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
-					if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
-						break;
-					}
-				}
-				assert( on2 );
-
-				snprintf( buf, sizeof( buf ), 
-					"cn=Overlay %d,%s", 
-					j, ms_overlay->mss_dn.bv_val );
-				ber_str2bv( buf, 0, 0, &bv );
-				attr_merge_normalize_one( e,
-						slap_schema.si_ad_seeAlso,
-						&bv, NULL );
-			}
-		}
-
-#if defined(SLAPD_BDB) || defined(SLAPD_HDB) 
-		if ( strcmp( bi->bi_type, "bdb" ) == 0
-				|| strcmp( bi->bi_type, "hdb" ) == 0 )
-		{
-			struct berval	bv;
-			ber_len_t	pathlen = 0, len = 0;
-			char		path[ PATH_MAX ] = { '\0' };
-			char		*fname = NULL;
-
-			if ( strcmp( bi->bi_type, "bdb" ) == 0
-					|| strcmp( bi->bi_type, "hdb" ) == 0 )
-			{
-				struct bdb_info *bdb = (struct bdb_info *) be->be_private;
-
-				fname = bdb->bi_dbenv_home;
-#if 0
-			} else if ( strcmp( bi->bi_type, "ldbm" ) == 0 ) {
-				struct ldbminfo *ldbm = (struct ldbminfo *) be->be_private;
-
-				/* FIXME: there's a conflict
-				 * between back-bdb.h and back.ldbm.h;
-				 * anyway, this code will be moved
-				 * to the backends as soon as the
-				 * issue with filtering on namingContexts
-				 * is fixed */
-				fname = ldbm->li_directory;
-#endif
-			}
-
-			len = strlen( fname );
-			if ( fname[ 0 ] != '/' ) {
-				/* get full path name */
-				getcwd( path, sizeof( path ) );
-				pathlen = strlen( path );
-
-				if ( fname[ 0 ] == '.' && fname[ 1 ] == '/' ) {
-					fname += 2;
-					len -= 2;
-				}
-			}
-
-			bv.bv_len = STRLENOF( "file://" ) + pathlen
-				+ STRLENOF( "/" ) + len;
-			bv.bv_val = ch_malloc( bv.bv_len + STRLENOF( "/" ) + 1 );
-			AC_MEMCPY( bv.bv_val, "file://", STRLENOF( "file://" ) );
-			if ( pathlen ) {
-				AC_MEMCPY( &bv.bv_val[ STRLENOF( "file://" ) ],
-						path, pathlen );
-				bv.bv_val[ STRLENOF( "file://" ) + pathlen ] = '/';
-				pathlen++;
-			}
-			AC_MEMCPY( &bv.bv_val[ STRLENOF( "file://" ) + pathlen ],
-					fname, len );
-			if ( bv.bv_val[ bv.bv_len - 1 ] != '/' ) {
-				bv.bv_val[ bv.bv_len ] = '/';
-				bv.bv_len++;
-			}
-			bv.bv_val[ bv.bv_len ] = '\0';
-
-			attr_merge_normalize_one( e, slap_schema.si_ad_labeledURI,
-					&bv, NULL );
-
-			ch_free( bv.bv_val );
-		}
-#endif /* defined(SLAPD_LDAP) || defined(SLAPD_HDB) */
-
-#if defined(SLAPD_LDAP) 
-		if ( strcmp( bi->bi_type, "ldap" ) == 0 ) {
-			struct ldapinfo		*li =
-				(struct ldapinfo *)be->be_private;
-			struct berval		bv;
-
-			ber_str2bv( li->url, 0, 0, &bv );
-
-			attr_merge_normalize_one( e,
-					slap_schema.si_ad_labeledURI,
-					&bv, NULL );
-		}
-#endif /* defined(SLAPD_LDAP) */
-
-		j = -1;
-		LDAP_STAILQ_FOREACH( bi2, &backendInfo, bi_next ) {
-			j++;
-			if ( bi2->bi_type == bi->bi_type ) {
-				struct berval 		bv;
-
-				snprintf( buf, sizeof( buf ), 
-					"cn=Backend %d,%s", 
-					j, ms_backend->mss_dn.bv_val );
-				bv.bv_val = buf;
-				bv.bv_len = strlen( buf );
-				attr_merge_normalize_one( e,
-						slap_schema.si_ad_seeAlso,
-						&bv, NULL );
-				break;
-			}
-		}
-		/* we must find it! */
-		assert( j >= 0 );
-
-		mp = monitor_entrypriv_create();
-		if ( mp == NULL ) {
+		bv.bv_val = buf;
+		bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Database %d", ++i );
+		if ( bv.bv_len >= sizeof( buf ) ) {
 			return -1;
 		}
-		e->e_private = ( void * )mp;
-		mp->mp_info = ms;
-		mp->mp_flags = ms->mss_flags
-			| MONITOR_F_SUB;
-
-		if ( monitor_cache_add( mi, e ) ) {
-			Debug( LDAP_DEBUG_ANY,
-				"monitor_subsys_database_init: "
-				"unable to add entry \"cn=Database %d,%s\"\n",
-				i, ms->mss_dn.bv_val, 0 );
-			return( -1 );
-		}
-
-#if defined(LDAP_SLAPI)
-		monitor_back_add_plugin( mi, be, e );
-#endif /* defined(LDAP_SLAPI) */
-
-		if ( oi != NULL ) {
-			Entry		**ep_overlay = &mp->mp_children;
-			monitor_entry_t	*mp_overlay;
-			slap_overinst	*on = oi->oi_list;
-			int		o;
-
-			for ( o = 0; on; o++, on = on->on_next ) {
-				Entry			*e_overlay;
-				slap_overinst		*on2;
-
-				/* find the overlay number, j */
-				for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
-					if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
-						break;
-					}
-				}
-				assert( on2 );
-
-				snprintf( buf, sizeof( buf ),
-						"dn: cn=Overlay %d,cn=Database %d,%s\n"
-						"objectClass: %s\n"
-						"structuralObjectClass: %s\n"
-						"cn: Overlay %d\n"
-						"%s: %s\n"
-						"seeAlso: cn=Overlay %d,%s\n"
-						"creatorsName: %s\n"
-						"modifiersName: %s\n"
-						"createTimestamp: %s\n"
-						"modifyTimestamp: %s\n",
-						o,
-						i,
-						ms->mss_dn.bv_val,
-						mi->mi_oc_monitoredObject->soc_cname.bv_val,
-						mi->mi_oc_monitoredObject->soc_cname.bv_val,
-						o,
-						mi->mi_ad_monitoredInfo->ad_cname.bv_val,
-						on->on_bi.bi_type,
-						j,
-						ms_overlay->mss_dn.bv_val,
-						mi->mi_creatorsName.bv_val,
-						mi->mi_creatorsName.bv_val,
-						mi->mi_startTime.bv_val,
-						mi->mi_startTime.bv_val );
-				
-				e_overlay = str2entry( buf );
-				if ( e_overlay == NULL ) {
-					Debug( LDAP_DEBUG_ANY,
-						"monitor_subsys_database_init: "
-						"unable to create entry "
-						"\"cn=Overlay %d,cn=Database %d,%s\"\n",
-						o, i, ms->mss_dn.bv_val );
-					return( -1 );
-				}
-
-				mp_overlay = monitor_entrypriv_create();
-				if ( mp_overlay == NULL ) {
-					return -1;
-				}
-				e_overlay->e_private = ( void * )mp_overlay;
-				mp_overlay->mp_info = ms;
-				mp_overlay->mp_flags = ms->mss_flags
-					| MONITOR_F_SUB;
 		
-				if ( monitor_cache_add( mi, e_overlay ) ) {
-					Debug( LDAP_DEBUG_ANY,
-						"monitor_subsys_database_init: "
-						"unable to add entry "
-						"\"cn=Overlay %d,cn=Database %d,%s\"\n",
-						o, i, ms->mss_dn.bv_val );
-					return( -1 );
-				}
-
-				*ep_overlay = e_overlay;
-				ep_overlay = &mp_overlay->mp_next;
-			}
+		rc = monitor_subsys_database_init_one( mi, be,
+			ms, ms_backend, ms_overlay, &bv, e_database, &ep );
+		if ( rc != 0 ) {
+			return rc;
 		}
-
-		*ep = e;
-		ep = &mp->mp_next;
 	}
 	
 	monitor_cache_release( mi, e_database );
@@ -540,12 +578,11 @@ value_mask( BerVarray v, slap_mask_t cur, slap_mask_t *delta )
 	return LDAP_SUCCESS;
 }
 
-int
+static int
 monitor_subsys_database_modify(
 	Operation	*op,
 	SlapReply	*rs,
-	Entry		*e
-)
+	Entry		*e )
 {
 	monitor_info_t	*mi = (monitor_info_t *)op->o_bd->be_private;
 	int		rc = LDAP_OTHER;
