@@ -1,8 +1,8 @@
 /* operational.c - monitor backend operational attributes function */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/operational.c,v 1.14.2.1 2005/01/20 18:04:00 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/operational.c,v 1.17.2.2 2007/08/31 23:14:03 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2005 The OpenLDAP Foundation.
+ * Copyright 2001-2007 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -41,7 +41,7 @@ monitor_back_operational(
 {
 	Attribute	**ap;
 
-	assert( rs->sr_entry );
+	assert( rs->sr_entry != NULL );
 
 	for ( ap = &rs->sr_operational_attrs; *ap; ap = &(*ap)->a_next )
 		/* just count */ ;
@@ -54,11 +54,11 @@ monitor_back_operational(
 
 		mp = ( monitor_entry_t * )rs->sr_entry->e_private;
 
-		assert( mp );
+		assert( mp != NULL );
 
 		hs = MONITOR_HAS_CHILDREN( mp );
 		*ap = slap_operational_hasSubordinate( hs );
-		assert( *ap );
+		assert( *ap != NULL );
 		ap = &(*ap)->a_next;
 	}
 	
