@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/add.c,v 1.20.2.2 2005/04/29 21:29:09 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/add.c,v 1.50.2.4 2007/08/31 23:14:05 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Copyright 1999-2007 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -66,7 +66,7 @@ backsql_modify_delete_all_values(
 	SQLHSTMT	asth = SQL_NULL_HSTMT;
 	BACKSQL_ROW_NTS	row;
 
-	assert( at );
+	assert( at != NULL );
 	if ( at->bam_delete_proc == NULL ) {
 		Debug( LDAP_DEBUG_TRACE,
 			"   backsql_modify_delete_all_values(): "
@@ -91,11 +91,8 @@ backsql_modify_delete_all_values(
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 				asth, rc );
 
-		if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-			rs->sr_text = "SQL-backend error";
-			return rs->sr_err = LDAP_OTHER;
-		}
-		return LDAP_SUCCESS;
+		rs->sr_text = "SQL-backend error";
+		return rs->sr_err = LDAP_OTHER;
 	}
 
 	rc = backsql_BindParamID( asth, 1, SQL_PARAM_INPUT, &e_id->eid_keyval );
@@ -109,12 +106,8 @@ backsql_modify_delete_all_values(
 				asth, rc );
 		SQLFreeStmt( asth, SQL_DROP );
 
-		if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-			rs->sr_text = "SQL-backend error";
-			return rs->sr_err = LDAP_OTHER;
-		}
-
-		return LDAP_SUCCESS;
+		rs->sr_text = "SQL-backend error";
+		return rs->sr_err = LDAP_OTHER;
 	}
 			
 	rc = SQLExecute( asth );
@@ -127,24 +120,21 @@ backsql_modify_delete_all_values(
 				asth, rc );
 		SQLFreeStmt( asth, SQL_DROP );
 
-		if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-			rs->sr_text = "SQL-backend error";
-			return rs->sr_err = LDAP_OTHER;
-		}
-
-		return LDAP_SUCCESS;
+		rs->sr_text = "SQL-backend error";
+		return rs->sr_err = LDAP_OTHER;
 	}
 
-	backsql_BindRowAsStrings( asth, &row );
+	backsql_BindRowAsStrings_x( asth, &row, op->o_tmpmemctx );
 	for ( rc = SQLFetch( asth );
 			BACKSQL_SUCCESS( rc );
 			rc = SQLFetch( asth ) )
 	{
-		int			i;
+		int		i;
 		/* first parameter no, parameter order */
-		SQLUSMALLINT		pno, po;
+		SQLUSMALLINT	pno = 0,
+				po = 0;
 		/* procedure return code */
-		int			prc = LDAP_SUCCESS;
+		int		prc = LDAP_SUCCESS;
 		
 		for ( i = 0; i < row.ncols; i++ ) {
 			SQLHSTMT	sth = SQL_NULL_HSTMT;
@@ -161,12 +151,9 @@ backsql_modify_delete_all_values(
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 
-				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-					rs->sr_text = "SQL-backend error";
-					return rs->sr_err = LDAP_OTHER;
-				}
-
-				continue;
+				rs->sr_text = "SQL-backend error";
+				rs->sr_err = LDAP_OTHER;
+				goto done;
 			}
 
 	   		if ( BACKSQL_IS_DEL( at->bam_expect_return ) ) {
@@ -182,16 +169,10 @@ backsql_modify_delete_all_values(
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
 
-					if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-						rs->sr_text = "SQL-backend error";
-						return rs->sr_err = LDAP_OTHER;
-					}
-
-					continue;
+					rs->sr_text = "SQL-backend error";
+					rs->sr_err = LDAP_OTHER;
+					goto done;
 				}
-
-			} else {
-				pno = 0;
 			}
 			po = ( BACKSQL_IS_DEL( at->bam_param_order ) ) > 0;
 			rc = backsql_BindParamID( sth, pno + 1 + po,
@@ -205,22 +186,19 @@ backsql_modify_delete_all_values(
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
 
-				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-					rs->sr_text = "SQL-backend error";
-					return rs->sr_err = LDAP_OTHER;
-				}
-
-				continue;
+				rs->sr_text = "SQL-backend error";
+				rs->sr_err = LDAP_OTHER;
+				goto done;
 			}
 #ifdef BACKSQL_ARBITRARY_KEY
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_modify_delete_all_values() "
-				"arg%d=%s\n",
+				"arg(%d)=%s\n",
 				pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
 #else /* ! BACKSQL_ARBITRARY_KEY */
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_modify_delete_all_values() "
-				"arg%d=%lu\n",
+				"arg(%d)=%lu\n",
 				pno + 1 + po, e_id->eid_keyval, 0 );
 #endif /* ! BACKSQL_ARBITRARY_KEY */
 
@@ -240,17 +218,14 @@ backsql_modify_delete_all_values(
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
 
-				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-					rs->sr_text = "SQL-backend error";
-					return rs->sr_err = LDAP_OTHER;
-				}
-
-				continue;
+				rs->sr_text = "SQL-backend error";
+				rs->sr_err = LDAP_OTHER;
+				goto done;
 			}
 	 
 			Debug( LDAP_DEBUG_TRACE, 
 				"   backsql_modify_delete_all_values(): "
-				"arg%d=%s; executing \"%s\"\n",
+				"arg(%d)=%s; executing \"%s\"\n",
 				pno + 2 - po, row.cols[ i ],
 				at->bam_delete_proc );
 			rc = SQLExecute( sth );
@@ -267,29 +242,27 @@ backsql_modify_delete_all_values(
 					/* SQL procedure executed fine 
 					 * but returned an error */
 					rs->sr_err = BACKSQL_SANITIZE_ERROR( prc );
-					rs->sr_text = op->ora_e->e_name.bv_val;
-					SQLFreeStmt( sth, SQL_DROP );
-					return rs->sr_err;
 
 				} else {
 					backsql_PrintErrors( bi->sql_db_env, dbh,
 							sth, rc );
-					if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) 
-					{
-						rs->sr_err = LDAP_OTHER;
-						rs->sr_text = op->ora_e->e_name.bv_val;
-						SQLFreeStmt( sth, SQL_DROP );
-						return rs->sr_err;
-					}
+					rs->sr_err = LDAP_OTHER;
 				}
+				rs->sr_text = op->o_req_dn.bv_val;
+				SQLFreeStmt( sth, SQL_DROP );
+				goto done;
 			}
 			SQLFreeStmt( sth, SQL_DROP );
 		}
 	}
-	backsql_FreeRow( &row );
+
+	rs->sr_err = LDAP_SUCCESS;
+
+done:;
+	backsql_FreeRow_x( &row, op->o_tmpmemctx );
 	SQLFreeStmt( asth, SQL_DROP );
 
-	return LDAP_SUCCESS;
+	return rs->sr_err;
 }
 
 int
@@ -301,9 +274,8 @@ backsql_modify_internal(
 	backsql_entryID		*e_id,
 	Modifications		*modlist )
 {
-	backsql_info	*bi = (backsql_info*)op->o_bd->be_private;
+	backsql_info	*bi = (backsql_info *)op->o_bd->be_private;
 	RETCODE		rc;
-	SQLHSTMT	sth = SQL_NULL_HSTMT;
 	Modifications	*ml;
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_modify_internal(): "
@@ -323,10 +295,6 @@ backsql_modify_internal(
 		backsql_at_map_rec	*at = NULL;
 		struct berval		*at_val;
 		int			i;
-		/* first parameter position, parameter order */
-		SQLUSMALLINT		pno, po;
-		/* procedure return code */
-		int			prc = LDAP_SUCCESS;
 		
 		ad = ml->sml_mod.sm_desc;
 		sm_op = ( ml->sml_mod.sm_op & LDAP_MOD_OP );
@@ -455,6 +423,13 @@ add_only:;
 					!BER_BVISNULL( at_val ); 
 					i++, at_val++ )
 			{
+				SQLHSTMT	sth = SQL_NULL_HSTMT;
+				/* first parameter position, parameter order */
+				SQLUSMALLINT	pno = 0,
+						po;
+				/* procedure return code */
+				int		prc = LDAP_SUCCESS;
+
 				rc = backsql_Prepare( dbh, &sth, at->bam_add_proc, 0 );
 				if ( rc != SQL_SUCCESS ) {
 					Debug( LDAP_DEBUG_TRACE,
@@ -485,9 +460,6 @@ add_only:;
 						rs->sr_err = LDAP_OTHER;
 						goto done;
 					}
-	 
-				} else {
-	      				pno = 0;
 				}
 				po = ( BACKSQL_IS_ADD( at->bam_param_order ) ) > 0;
 				rc = backsql_BindParamID( sth, pno + 1 + po,
@@ -508,12 +480,12 @@ add_only:;
 #ifdef BACKSQL_ARBITRARY_KEY
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg%d=\"%s\"\n", 
+					"arg(%d)=\"%s\"\n", 
 					pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
 #else /* ! BACKSQL_ARBITRARY_KEY */
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg%d=\"%lu\"\n", 
+					"arg(%d)=\"%lu\"\n", 
 					pno + 1 + po, e_id->eid_keyval, 0 );
 #endif /* ! BACKSQL_ARBITRARY_KEY */
 
@@ -538,24 +510,40 @@ add_only:;
 				}
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg%d=\"%s\"; executing \"%s\"\n", 
+					"arg(%d)=\"%s\"; executing \"%s\"\n", 
 					pno + 2 - po, at_val->bv_val,
 					at->bam_add_proc );
 
 				rc = SQLExecute( sth );
-				if ( rc != SQL_SUCCESS ) {
+				if ( rc == SQL_SUCCESS && prc == LDAP_SUCCESS ) {
+					rs->sr_err = LDAP_SUCCESS;
+
+				} else {
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
-						"add_proc execution failed\n",
-						0, 0, 0 );
-					backsql_PrintErrors( bi->sql_db_env,
-							dbh, sth, rc );
-
-					if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
+						"add_proc execution failed "
+						"(rc=%d, prc=%d)\n",
+						rc, prc, 0 );
+					if ( prc != LDAP_SUCCESS ) {
+						/* SQL procedure executed fine 
+						 * but returned an error */
 						SQLFreeStmt( sth, SQL_DROP );
-						rs->sr_err = LDAP_OTHER;
-						rs->sr_text = "SQL-backend error";
-						goto done;
+
+						rs->sr_err = BACKSQL_SANITIZE_ERROR( prc );
+						rs->sr_text = at->bam_ad->ad_cname.bv_val;
+						return rs->sr_err;
+					
+					} else {
+						backsql_PrintErrors( bi->sql_db_env, dbh,
+								sth, rc );
+						if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) 
+						{
+							SQLFreeStmt( sth, SQL_DROP );
+
+							rs->sr_err = LDAP_OTHER;
+							rs->sr_text = "SQL-backend error";
+							goto done;
+						}
 					}
 				}
 				SQLFreeStmt( sth, SQL_DROP );
@@ -598,7 +586,13 @@ add_only:;
 					!BER_BVISNULL( at_val );
 					i++, at_val++ )
 			{
-				prc = LDAP_SUCCESS;
+				SQLHSTMT	sth = SQL_NULL_HSTMT;
+				/* first parameter position, parameter order */
+				SQLUSMALLINT	pno = 0,
+						po;
+				/* procedure return code */
+				int		prc = LDAP_SUCCESS;
+
 				rc = backsql_Prepare( dbh, &sth, at->bam_delete_proc, 0 );
 				if ( rc != SQL_SUCCESS ) {
 					Debug( LDAP_DEBUG_TRACE,
@@ -629,9 +623,6 @@ add_only:;
 						rs->sr_err = LDAP_OTHER;
 						goto done;
 					}
-
-				} else {
-					pno = 0;
 				}
 				po = ( BACKSQL_IS_DEL( at->bam_param_order ) ) > 0;
 				rc = backsql_BindParamID( sth, pno + 1 + po,
@@ -652,12 +643,12 @@ add_only:;
 #ifdef BACKSQL_ARBITRARY_KEY
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg%d=\"%s\"\n", 
+					"arg(%d)=\"%s\"\n", 
 					pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
 #else /* ! BACKSQL_ARBITRARY_KEY */
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg%d=\"%lu\"\n", 
+					"arg(%d)=\"%lu\"\n", 
 					pno + 1 + po, e_id->eid_keyval, 0 );
 #endif /* ! BACKSQL_ARBITRARY_KEY */
 
@@ -676,11 +667,9 @@ add_only:;
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
 
-					if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-						rs->sr_text = "SQL-backend error";
-						rs->sr_err = LDAP_OTHER;
-						goto done;
-					}
+					rs->sr_text = "SQL-backend error";
+					rs->sr_err = LDAP_OTHER;
+					goto done;
 				}
 
 				Debug( LDAP_DEBUG_TRACE,
@@ -709,13 +698,10 @@ add_only:;
 					} else {
 						backsql_PrintErrors( bi->sql_db_env,
 								dbh, sth, rc );
-						if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) )
-						{
-							SQLFreeStmt( sth, SQL_DROP );
-							rs->sr_err = LDAP_OTHER;
-							rs->sr_text = at->bam_ad->ad_cname.bv_val;
-							goto done;
-						}
+						SQLFreeStmt( sth, SQL_DROP );
+						rs->sr_err = LDAP_OTHER;
+						rs->sr_text = at->bam_ad->ad_cname.bv_val;
+						goto done;
 					}
 				}
 				SQLFreeStmt( sth, SQL_DROP );
@@ -823,13 +809,8 @@ backsql_add_attr(
 
 		rc = backsql_Prepare( dbh, &sth, at_rec->bam_add_proc, 0 );
 		if ( rc != SQL_SUCCESS ) {
-
-			if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-				rs->sr_text = "SQL-backend error";
-				return rs->sr_err = LDAP_OTHER;
-			}
-
-			return LDAP_SUCCESS;
+			rs->sr_text = "SQL-backend error";
+			return rs->sr_err = LDAP_OTHER;
 		}
 
 		if ( BACKSQL_IS_ADD( at_rec->bam_expect_return ) ) {
@@ -838,18 +819,14 @@ backsql_add_attr(
 			if ( rc != SQL_SUCCESS ) {
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_add_attr(): "
-					"error binding output parameter for %s[%d]\n",
+					"error binding output parameter for %s[%lu]\n",
 					at_rec->bam_ad->ad_cname.bv_val, i, 0 );
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
 
-				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-					rs->sr_text = "SQL-backend error";
-					return rs->sr_err = LDAP_OTHER;
-				}
-
-				return LDAP_SUCCESS;
+				rs->sr_text = "SQL-backend error";
+				return rs->sr_err = LDAP_OTHER;
 			}
 
 		} else {
@@ -863,18 +840,14 @@ backsql_add_attr(
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_add_attr(): "
-				"error binding keyval parameter for %s[%d]\n",
+				"error binding keyval parameter for %s[%lu]\n",
 				at_rec->bam_ad->ad_cname.bv_val, i, 0 );
 			backsql_PrintErrors( bi->sql_db_env, dbh, 
 				sth, rc );
 			SQLFreeStmt( sth, SQL_DROP );
 
-			if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-				rs->sr_text = "SQL-backend error";
-				return rs->sr_err = LDAP_OTHER;
-			}
-
-			return LDAP_SUCCESS;
+			rs->sr_text = "SQL-backend error";
+			return rs->sr_err = LDAP_OTHER;
 		}
 
 		currpos = pno + 2 - po;
@@ -888,18 +861,14 @@ backsql_add_attr(
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_add_attr(): "
-				"error binding value parameter for %s[%d]\n",
+				"error binding value parameter for %s[%lu]\n",
 				at_rec->bam_ad->ad_cname.bv_val, i, 0 );
 			backsql_PrintErrors( bi->sql_db_env, dbh, 
 				sth, rc );
 			SQLFreeStmt( sth, SQL_DROP );
 
-			if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-				rs->sr_text = "SQL-backend error";
-				return rs->sr_err = LDAP_OTHER;
-			}
-
-			return LDAP_SUCCESS;
+			rs->sr_text = "SQL-backend error";
+			return rs->sr_err = LDAP_OTHER;
 		}
 
 #ifdef LDAP_DEBUG
@@ -930,12 +899,10 @@ backsql_add_attr(
 			} else {
 				backsql_PrintErrors( bi->sql_db_env, dbh,
 						sth, rc );
-				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
-					rs->sr_err = LDAP_OTHER;
-					rs->sr_text = op->ora_e->e_name.bv_val;
-					SQLFreeStmt( sth, SQL_DROP );
-					return rs->sr_err;
-				}
+				rs->sr_err = LDAP_OTHER;
+				rs->sr_text = op->ora_e->e_name.bv_val;
+				SQLFreeStmt( sth, SQL_DROP );
+				return rs->sr_err;
 			}
 		}
 		SQLFreeStmt( sth, SQL_DROP );
@@ -957,21 +924,27 @@ backsql_add( Operation *op, SlapReply *rs )
 	Entry			p = { 0 }, *e = NULL;
 	Attribute		*at,
 				*at_objectClass = NULL;
+	ObjectClass		*soc = NULL;
 	struct berval		scname = BER_BVNULL;
 	struct berval		pdn;
 	struct berval		realdn = BER_BVNULL;
 	int			colnum;
 	slap_mask_t		mask;
 
+	char			textbuf[ SLAP_TEXT_BUFLEN ];
+	size_t			textlen = sizeof( textbuf );
+
 #ifdef BACKSQL_SYNCPROV
 	/*
 	 * NOTE: fake successful result to force contextCSN to be bumped up
 	 */
 	if ( op->o_sync ) {
 		char		buf[ LDAP_LUTIL_CSNSTR_BUFSIZE ];
-		struct berval	csn = BER_BVNULL;
+		struct berval	csn;
 
-		slap_get_csn( op, buf, sizeof( buf ), &csn, 1 );
+		csn.bv_val = buf;
+		csn.bv_len = sizeof( buf );
+		slap_get_csn( op, &csn, 1 );
 
 		rs->sr_err = LDAP_SUCCESS;
 		send_ldap_result( op, rs );
@@ -986,12 +959,11 @@ backsql_add( Operation *op, SlapReply *rs )
 			op->ora_e->e_name.bv_val, 0, 0 );
 
 	/* check schema */
-	if ( global_schemacheck ) {
+	if ( BACKSQL_CHECK_SCHEMA( bi ) ) {
 		char		textbuf[ SLAP_TEXT_BUFLEN ] = { '\0' };
 
-		rs->sr_err = entry_schema_check( op->o_bd, op->ora_e,
-				NULL,
-				&rs->sr_text, textbuf, sizeof( textbuf ) );
+		rs->sr_err = entry_schema_check( op, op->ora_e, NULL, 0, 1,
+			&rs->sr_text, textbuf, sizeof( textbuf ) );
 		if ( rs->sr_err != LDAP_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"entry failed schema check -- aborting\n",
@@ -1001,6 +973,8 @@ backsql_add( Operation *op, SlapReply *rs )
 		}
 	}
 
+	slap_add_opattrs( op, &rs->sr_text, textbuf, textlen, 1 );
+
 	/* search structuralObjectClass */
 	for ( at = op->ora_e->e_attrs; at != NULL; at = at->a_next ) {
 		if ( at->a_desc == slap_schema.si_ad_structuralObjectClass ) {
@@ -1029,8 +1003,8 @@ backsql_add( Operation *op, SlapReply *rs )
 			goto done;
 		}
 
-		rs->sr_err = structural_class( at->a_vals, &scname, NULL,
-				&text, buf, sizeof( buf ) );
+		rs->sr_err = structural_class( at->a_vals, &soc, NULL,
+				&text, buf, sizeof( buf ), op->o_tmpmemctx );
 		if ( rs->sr_err != LDAP_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"%s (%d)\n",
@@ -1038,6 +1012,7 @@ backsql_add( Operation *op, SlapReply *rs )
 			e = NULL;
 			goto done;
 		}
+		scname = soc->soc_cname;
 
 	} else {
 		scname = at->a_vals[0];
@@ -1082,6 +1057,16 @@ backsql_add( Operation *op, SlapReply *rs )
 		goto done;
 	}
 
+	/* check write access */
+	if ( !access_allowed_mask( op, op->ora_e,
+				slap_schema.si_ad_entry,
+				NULL, ACL_WADD, NULL, &mask ) )
+	{
+		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
+		e = op->ora_e;
+		goto done;
+	}
+
 	rs->sr_err = backsql_get_db_conn( op, &dbh );
 	if ( rs->sr_err != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
@@ -1095,7 +1080,7 @@ backsql_add( Operation *op, SlapReply *rs )
 
 	/*
 	 * Check if entry exists
-	*
+	 *
 	 * NOTE: backsql_api_dn2odbc() is called explicitly because
 	 * we need the mucked DN to pass it to the create procedure.
 	 */
@@ -1128,53 +1113,34 @@ backsql_add( Operation *op, SlapReply *rs )
 
 	} else {
 		dnParent( &op->ora_e->e_nname, &pdn );
-	}
 
-	/*
-	 * Get the parent
-	 */
-	bsi.bsi_e = &p;
-	rs->sr_err = backsql_init_search( &bsi, &pdn,
-			LDAP_SCOPE_BASE, 
-			SLAP_NO_LIMIT, SLAP_NO_LIMIT,
-			(time_t)(-1), NULL, dbh, op, rs, slap_anlist_no_attrs,
-			( BACKSQL_ISF_MATCHED | BACKSQL_ISF_GET_ENTRY ) );
-	if ( rs->sr_err != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_add(): "
-			"could not retrieve addDN parent "
-			"\"%s\" ID - %s matched=\"%s\"\n", 
-			pdn.bv_val,
-			rs->sr_err == LDAP_REFERRAL ? "referral" : "no such entry",
-			rs->sr_matched ? rs->sr_matched : "(null)" );
-		e = &p;
-		goto done;
-	}
+		/*
+		 * Get the parent
+		 */
+		bsi.bsi_e = &p;
+		rs->sr_err = backsql_init_search( &bsi, &pdn,
+				LDAP_SCOPE_BASE, 
+				(time_t)(-1), NULL, dbh, op, rs, slap_anlist_no_attrs,
+				( BACKSQL_ISF_MATCHED | BACKSQL_ISF_GET_ENTRY ) );
+		if ( rs->sr_err != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_TRACE, "backsql_add(): "
+				"could not retrieve addDN parent "
+				"\"%s\" ID - %s matched=\"%s\"\n", 
+				pdn.bv_val,
+				rs->sr_err == LDAP_REFERRAL ? "referral" : "no such entry",
+				rs->sr_matched ? rs->sr_matched : "(null)" );
+			e = &p;
+			goto done;
+		}
 
-	/* check "children" pseudo-attribute access to parent */
-	if ( !access_allowed( op, &p, slap_schema.si_ad_children,
-				NULL, ACL_WADD, NULL ) )
-	{
-		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
-		e = &p;
-		goto done;
-	}
-
-	if ( get_assert( op ) &&
-		( test_filter( op, op->oq_add.rs_e, get_assertion( op ) )
-			!= LDAP_COMPARE_TRUE ) )
-	{
-		rs->sr_err = LDAP_ASSERTION_FAILED;
-		e = op->ora_e;
-		goto done;
-	}
-
-	if ( !access_allowed_mask( op, op->ora_e,
-				slap_schema.si_ad_entry,
-				NULL, ACL_WADD, NULL, &mask ) )
-	{
-		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
-		e = op->ora_e;
-		goto done;
+		/* check "children" pseudo-attribute access to parent */
+		if ( !access_allowed( op, &p, slap_schema.si_ad_children,
+					NULL, ACL_WADD, NULL ) )
+		{
+			rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
+			e = &p;
+			goto done;
+		}
 	}
 
 	/*
@@ -1497,6 +1463,7 @@ done:;
 		SQLUSMALLINT	CompletionType = SQL_ROLLBACK;
 
 		if ( rs->sr_err == LDAP_SUCCESS && !op->o_noop ) {
+			assert( e == NULL );
 			CompletionType = SQL_COMMIT;
 		}
 
@@ -1514,7 +1481,6 @@ done:;
 	 * in deleting that row.
 	 */
 
-#ifdef SLAP_ACL_HONOR_DISCLOSE
 	if ( e != NULL ) {
 		int	disclose = 1;
 
@@ -1539,9 +1505,13 @@ done:;
 			}
 		}
 	}
-#endif /* SLAP_ACL_HONOR_DISCLOSE */
+
+	if ( op->o_noop && rs->sr_err == LDAP_SUCCESS ) {
+		rs->sr_err = LDAP_X_NO_OPERATION;
+	}
 
 	send_ldap_result( op, rs );
+	slap_graduate_commit_csn( op );
 
 	if ( !BER_BVISNULL( &realdn )
 			&& realdn.bv_val != op->ora_e->e_name.bv_val )
@@ -1550,11 +1520,11 @@ done:;
 	}
 
 	if ( !BER_BVISNULL( &bsi.bsi_base_id.eid_ndn ) ) {
-		(void)backsql_free_entryID( op, &bsi.bsi_base_id, 0 );
+		(void)backsql_free_entryID( &bsi.bsi_base_id, 0, op->o_tmpmemctx );
 	}
 
 	if ( !BER_BVISNULL( &p.e_nname ) ) {
-		entry_clean( &p );
+		backsql_entry_clean( op, &p );
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_add(\"%s\"): %d \"%s\"\n",
