@@ -1,8 +1,8 @@
 /* filter.c - routines for parsing and dealing with filters */
-/* $OpenLDAP: pkg/ldap/servers/slapd/filter.c,v 1.125.2.1 2005/01/20 18:03:55 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/filter.c,v 1.134.2.3 2007/08/31 23:13:59 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,14 +42,9 @@ static int	get_filter_list(
 static int	get_ssa(
 	Operation *op,
 	BerElement *ber,
-	SubstringsAssertion **s,
+	Filter *f,
 	const char **text );
 
-static int filter_escape_value_x(
-	struct berval *in,
-	struct berval *out,
-	void *ctx );
-
 static void simple_vrFilter2bv(
 	Operation *op,
 	ValuesReturnFilter *f,
@@ -84,8 +79,8 @@ get_filter(
 	 *		substrings	[4]	SubstringFilter,
 	 *		greaterOrEqual	[5]	AttributeValueAssertion,
 	 *		lessOrEqual	[6]	AttributeValueAssertion,
-	 *		present		[7]	AttributeType,,
-	 *		approxMatch	[8]	AttributeValueAssertion
+	 *		present		[7]	AttributeType,
+	 *		approxMatch	[8]	AttributeValueAssertion,
 	 *		extensibleMatch [9]	MatchingRuleAssertion
 	 *	}
 	 *
@@ -122,7 +117,7 @@ get_filter(
 	switch ( f.f_choice ) {
 	case LDAP_FILTER_EQUALITY:
 		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n", 0, 0, 0 );
-		err = get_ava( op, ber, &f.f_ava, SLAP_MR_EQUALITY, text );
+		err = get_ava( op, ber, &f, SLAP_MR_EQUALITY, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -132,7 +127,7 @@ get_filter(
 
 	case LDAP_FILTER_SUBSTRINGS:
 		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n", 0, 0, 0 );
-		err = get_ssa( op, ber, &f.f_sub, text );
+		err = get_ssa( op, ber, &f, text );
 		if( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -141,7 +136,7 @@ get_filter(
 
 	case LDAP_FILTER_GE:
 		Debug( LDAP_DEBUG_FILTER, "GE\n", 0, 0, 0 );
-		err = get_ava( op, ber, &f.f_ava, SLAP_MR_ORDERING, text );
+		err = get_ava( op, ber, &f, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -150,7 +145,7 @@ get_filter(
 
 	case LDAP_FILTER_LE:
 		Debug( LDAP_DEBUG_FILTER, "LE\n", 0, 0, 0 );
-		err = get_ava( op, ber, &f.f_ava, SLAP_MR_ORDERING, text );
+		err = get_ava( op, ber, &f, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -171,17 +166,21 @@ get_filter(
 		err = slap_bv2ad( &type, &f.f_desc, text );
 
 		if( err != LDAP_SUCCESS ) {
-			/* unrecognized attribute description or other error */
-			Debug( LDAP_DEBUG_ANY, 
-				"get_filter: conn %d unknown attribute "
-				"type=%s (%d)\n",
-				op->o_connid, type.bv_val, err );
+			f.f_choice |= SLAPD_FILTER_UNDEFINED;
+			err = slap_bv2undef_ad( &type, &f.f_desc, text,
+				SLAP_AD_PROXIED|SLAP_AD_NOINSERT );
 
-			f.f_choice = SLAPD_FILTER_COMPUTED;
-			f.f_result = LDAP_COMPARE_FALSE;
-			err = LDAP_SUCCESS;
+			if ( err != LDAP_SUCCESS ) {
+				/* unrecognized attribute description or other error */
+				Debug( LDAP_DEBUG_ANY, 
+					"get_filter: conn %lu unknown attribute "
+					"type=%s (%d)\n",
+					op->o_connid, type.bv_val, err );
+
+				err = LDAP_SUCCESS;
+				f.f_desc = slap_bv2tmp_ad( &type, op->o_tmpmemctx );
+			}
 			*text = NULL;
-			break;
 		}
 
 		assert( f.f_desc != NULL );
@@ -189,7 +188,7 @@ get_filter(
 
 	case LDAP_FILTER_APPROX:
 		Debug( LDAP_DEBUG_FILTER, "APPROX\n", 0, 0, 0 );
-		err = get_ava( op, ber, &f.f_ava, SLAP_MR_EQUALITY_APPROX, text );
+		err = get_ava( op, ber, &f, SLAP_MR_EQUALITY_APPROX, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -253,7 +252,7 @@ get_filter(
 	case LDAP_FILTER_EXT:
 		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n", 0, 0, 0 );
 
-		err = get_mra( op, ber, &f.f_mra, text );
+		err = get_mra( op, ber, &f, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -320,7 +319,7 @@ static int
 get_ssa(
 	Operation *op,
 	BerElement	*ber,
-	SubstringsAssertion	**out,
+	Filter 		*f,
 	const char	**text )
 {
 	ber_tag_t	tag;
@@ -331,7 +330,6 @@ get_ssa(
 	SubstringsAssertion ssa;
 
 	*text = "error decoding filter";
-	*out = NULL;
 
 	Debug( LDAP_DEBUG_FILTER, "begin get_ssa\n", 0, 0, 0 );
 	if ( ber_scanf( ber, "{m" /*}*/, &desc ) == LBER_ERROR ) {
@@ -348,17 +346,17 @@ get_ssa(
 	rc = slap_bv2ad( &desc, &ssa.sa_desc, text );
 
 	if( rc != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY, 
-			"get_ssa: conn %d unknown attribute type=%s (%d)\n",
-			op->o_connid, desc.bv_val, rc );
+		f->f_choice |= SLAPD_FILTER_UNDEFINED;
+		rc = slap_bv2undef_ad( &desc, &ssa.sa_desc, text,
+			SLAP_AD_PROXIED|SLAP_AD_NOINSERT );
 
-		/* skip over the rest of this filter */
-		for ( tag = ber_first_element( ber, &len, &last );
-			tag != LBER_DEFAULT;
-			tag = ber_next_element( ber, &len, last ) ) {
-			ber_scanf( ber, "x" );
+		if( rc != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY, 
+				"get_ssa: conn %lu unknown attribute type=%s (%ld)\n",
+				op->o_connid, desc.bv_val, (long) rc );
+	
+			ssa.sa_desc = slap_bv2tmp_ad( &desc, op->o_tmpmemctx );
 		}
-		return rc;
 	}
 
 	rc = LDAP_PROTOCOL_ERROR;
@@ -382,24 +380,39 @@ get_ssa(
 
 		switch ( tag ) {
 		case LDAP_SUBSTRING_INITIAL:
+			if ( ssa.sa_initial.bv_val != NULL
+				|| ssa.sa_any != NULL 
+				|| ssa.sa_final.bv_val != NULL )
+			{
+				rc = LDAP_PROTOCOL_ERROR;
+				goto return_error;
+			}
 			usage = SLAP_MR_SUBSTR_INITIAL;
 			break;
 
 		case LDAP_SUBSTRING_ANY:
+			if ( ssa.sa_final.bv_val != NULL ) {
+				rc = LDAP_PROTOCOL_ERROR;
+				goto return_error;
+			}
 			usage = SLAP_MR_SUBSTR_ANY;
 			break;
 
 		case LDAP_SUBSTRING_FINAL:
+			if ( ssa.sa_final.bv_val != NULL ) {
+				rc = LDAP_PROTOCOL_ERROR;
+				goto return_error;
+			}
+
 			usage = SLAP_MR_SUBSTR_FINAL;
 			break;
 
 		default:
-			rc = LDAP_PROTOCOL_ERROR;
-
 			Debug( LDAP_DEBUG_FILTER,
 				"  unknown substring choice=%ld\n",
 				(long) tag, 0, 0 );
 
+			rc = LDAP_PROTOCOL_ERROR;
 			goto return_error;
 		}
 
@@ -407,63 +420,36 @@ get_ssa(
 		rc = asserted_value_validate_normalize(
 			ssa.sa_desc, ssa.sa_desc->ad_type->sat_equality,
 			usage, &value, &nvalue, text, op->o_tmpmemctx );
-
-		if( rc != LDAP_SUCCESS ) {
-			goto return_error;
-		}
-
-		rc = LDAP_PROTOCOL_ERROR;
+		if( rc != LDAP_SUCCESS ) goto return_error;
 
 		switch ( tag ) {
 		case LDAP_SUBSTRING_INITIAL:
 			Debug( LDAP_DEBUG_FILTER, "  INITIAL\n", 0, 0, 0 );
-
-			if ( ssa.sa_initial.bv_val != NULL
-				|| ssa.sa_any != NULL 
-				|| ssa.sa_final.bv_val != NULL )
-			{
-				slap_sl_free( nvalue.bv_val, op->o_tmpmemctx );
-				goto return_error;
-			}
-
 			ssa.sa_initial = nvalue;
 			break;
 
 		case LDAP_SUBSTRING_ANY:
 			Debug( LDAP_DEBUG_FILTER, "  ANY\n", 0, 0, 0 );
-
-			if ( ssa.sa_final.bv_val != NULL ) {
-				slap_sl_free( nvalue.bv_val, op->o_tmpmemctx );
-				goto return_error;
-			}
-
 			ber_bvarray_add_x( &ssa.sa_any, &nvalue, op->o_tmpmemctx );
 			break;
 
 		case LDAP_SUBSTRING_FINAL:
 			Debug( LDAP_DEBUG_FILTER, "  FINAL\n", 0, 0, 0 );
-
-			if ( ssa.sa_final.bv_val != NULL ) {
-				slap_sl_free( nvalue.bv_val, op->o_tmpmemctx );
-				goto return_error;
-			}
-
 			ssa.sa_final = nvalue;
 			break;
 
 		default:
-			Debug( LDAP_DEBUG_FILTER,
-				"  unknown substring type=%ld\n",
-				(long) tag, 0, 0 );
-
 			assert( 0 );
 			slap_sl_free( nvalue.bv_val, op->o_tmpmemctx );
+			rc = LDAP_PROTOCOL_ERROR;
 
 return_error:
 			Debug( LDAP_DEBUG_FILTER, "  error=%ld\n",
 				(long) rc, 0, 0 );
 			slap_sl_free( ssa.sa_initial.bv_val, op->o_tmpmemctx );
 			ber_bvarray_free_x( ssa.sa_any, op->o_tmpmemctx );
+			if ( ssa.sa_desc->ad_flags & SLAP_DESC_TEMPORARY )
+				op->o_tmpfree( ssa.sa_desc, op->o_tmpmemctx );
 			slap_sl_free( ssa.sa_final.bv_val, op->o_tmpmemctx );
 			return rc;
 		}
@@ -472,12 +458,11 @@ return_error:
 	}
 
 	if( rc == LDAP_SUCCESS ) {
-		*out = op->o_tmpalloc( sizeof( ssa ), op->o_tmpmemctx );
-		**out = ssa;
+		f->f_sub = op->o_tmpalloc( sizeof( ssa ), op->o_tmpmemctx );
+		*f->f_sub = ssa;
 	}
 
 	Debug( LDAP_DEBUG_FILTER, "end get_ssa\n", 0, 0, 0 );
-
 	return rc /* LDAP_SUCCESS */ ;
 }
 
@@ -490,6 +475,8 @@ filter_free_x( Operation *op, Filter *f )
 		return;
 	}
 
+	f->f_choice &= SLAPD_FILTER_MASK;
+
 	switch ( f->f_choice ) {
 	case LDAP_FILTER_PRESENT:
 		break;
@@ -509,6 +496,8 @@ filter_free_x( Operation *op, Filter *f )
 		if ( f->f_sub_final.bv_val != NULL ) {
 			op->o_tmpfree( f->f_sub_final.bv_val, op->o_tmpmemctx );
 		}
+		if ( f->f_sub->sa_desc->ad_flags & SLAP_DESC_TEMPORARY )
+			op->o_tmpfree( f->f_sub->sa_desc, op->o_tmpmemctx );
 		op->o_tmpfree( f->f_sub, op->o_tmpmemctx );
 		break;
 
@@ -563,121 +552,123 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 			ber_bvunknown = BER_BVC( "(?=unknown)" ),
 			ber_bvnone = BER_BVC( "(?=none)" );
 	ber_len_t	len;
+	ber_tag_t	choice;
+	int undef;
+	char *sign;
 
 	if ( f == NULL ) {
 		ber_dupbv_x( fstr, &ber_bvnone, op->o_tmpmemctx );
 		return;
 	}
 
-	switch ( f->f_choice ) {
+	undef = f->f_choice & SLAPD_FILTER_UNDEFINED;
+	choice = f->f_choice & SLAPD_FILTER_MASK;
+
+	switch ( choice ) {
 	case LDAP_FILTER_EQUALITY:
-		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
-
-		fstr->bv_len = f->f_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(=)") - 1 );
-		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
-
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=%s)",
-			f->f_av_desc->ad_cname.bv_val,
-			tmp.bv_val );
-
-		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
-		break;
-
+		fstr->bv_len = STRLENOF("(=)");
+		sign = "=";
+		goto simple;
 	case LDAP_FILTER_GE:
-		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
-
-		fstr->bv_len = f->f_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(>=)") - 1 );
-		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
-
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s>=%s)",
-			f->f_av_desc->ad_cname.bv_val,
-			tmp.bv_val );
-
-		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
-		break;
-
+		fstr->bv_len = STRLENOF("(>=)");
+		sign = ">=";
+		goto simple;
 	case LDAP_FILTER_LE:
-		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
-
-		fstr->bv_len = f->f_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(<=)") - 1 );
-		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
-
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s<=%s)",
-			f->f_av_desc->ad_cname.bv_val,
-			tmp.bv_val );
-
-		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
-		break;
-
+		fstr->bv_len = STRLENOF("(<=)");
+		sign = "<=";
+		goto simple;
 	case LDAP_FILTER_APPROX:
-		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
+		fstr->bv_len = STRLENOF("(~=)");
+		sign = "~=";
 
-		fstr->bv_len = f->f_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(~=)") - 1 );
+simple:
+		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
+		/* NOTE: tmp can legitimately be NULL (meaning empty) 
+		 * since in a Filter values in AVAs are supposed
+		 * to have been normalized, meaning that an empty value
+		 * is legal for that attribute's syntax */
+
+		fstr->bv_len += f->f_av_desc->ad_cname.bv_len + tmp.bv_len;
+		if ( undef )
+			fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s~=%s)",
-			f->f_av_desc->ad_cname.bv_val,
-			tmp.bv_val );
+		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s)",
+			undef ? "?" : "",
+			f->f_av_desc->ad_cname.bv_val, sign,
+			tmp.bv_len ? tmp.bv_val : "" );
+
 		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
 		fstr->bv_len = f->f_sub_desc->ad_cname.bv_len +
-			( sizeof("(=*)") - 1 );
+			STRLENOF("(=*)");
+		if ( undef )
+			fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 128, op->o_tmpmemctx );
 
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=*)",
+		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s=*)",
+			undef ? "?" : "",
 			f->f_sub_desc->ad_cname.bv_val );
 
 		if ( f->f_sub_initial.bv_val != NULL ) {
+			ber_len_t tmplen;
+
 			len = fstr->bv_len;
 
 			filter_escape_value_x( &f->f_sub_initial, &tmp, op->o_tmpmemctx );
+			tmplen = tmp.bv_len;
 
-			fstr->bv_len += tmp.bv_len;
-			fstr->bv_val = op->o_tmprealloc( fstr->bv_val, fstr->bv_len + 1,
-				op->o_tmpmemctx );
+			fstr->bv_len += tmplen;
+			fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
+				fstr->bv_len + 1, op->o_tmpmemctx );
 
-			snprintf( &fstr->bv_val[len-2], tmp.bv_len+3,
+			snprintf( &fstr->bv_val[len - 2],
+				tmplen + STRLENOF( /*(*/ "*)" ) + 1,
 				/* "(attr=" */ "%s*)",
-				tmp.bv_val );
+				tmp.bv_len ? tmp.bv_val : "");
 
 			ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		}
 
 		if ( f->f_sub_any != NULL ) {
 			for ( i = 0; f->f_sub_any[i].bv_val != NULL; i++ ) {
+				ber_len_t tmplen;
+
 				len = fstr->bv_len;
 				filter_escape_value_x( &f->f_sub_any[i],
 					&tmp, op->o_tmpmemctx );
+				tmplen = tmp.bv_len;
 
-				fstr->bv_len += tmp.bv_len + 1;
-				fstr->bv_val = op->o_tmprealloc( fstr->bv_val, fstr->bv_len + 1,
-					op->o_tmpmemctx );
+				fstr->bv_len += tmplen + STRLENOF( /*(*/ ")" );
+				fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
+					fstr->bv_len + 1, op->o_tmpmemctx );
 
-				snprintf( &fstr->bv_val[len-1], tmp.bv_len+3,
+				snprintf( &fstr->bv_val[len - 1],
+					tmplen + STRLENOF( /*(*/ "*)" ) + 1,
 					/* "(attr=[init]*[any*]" */ "%s*)",
-					tmp.bv_val );
+					tmp.bv_len ? tmp.bv_val : "");
 				ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 			}
 		}
 
 		if ( f->f_sub_final.bv_val != NULL ) {
+			ber_len_t tmplen;
+
 			len = fstr->bv_len;
 
 			filter_escape_value_x( &f->f_sub_final, &tmp, op->o_tmpmemctx );
+			tmplen = tmp.bv_len;
 
-			fstr->bv_len += tmp.bv_len;
-			fstr->bv_val = op->o_tmprealloc( fstr->bv_val, fstr->bv_len + 1,
-				op->o_tmpmemctx );
+			fstr->bv_len += tmplen;
+			fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
+				fstr->bv_len + 1, op->o_tmpmemctx );
 
-			snprintf( &fstr->bv_val[len-1], tmp.bv_len+3,
+			snprintf( &fstr->bv_val[len - 1],
+				tmplen + STRLENOF( /*(*/ ")" ) + 1,
 				/* "(attr=[init*][any*]" */ "%s)",
-				tmp.bv_val );
+				tmp.bv_len ? tmp.bv_val : "");
 
 			ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		}
@@ -686,17 +677,21 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 
 	case LDAP_FILTER_PRESENT:
 		fstr->bv_len = f->f_desc->ad_cname.bv_len +
-			( sizeof("(=*)") - 1 );
+			STRLENOF("(=*)");
+		if ( undef )
+			fstr->bv_len++;
+
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
-		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=*)",
+		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s=*)",
+			undef ? "?" : "",
 			f->f_desc->ad_cname.bv_val );
 		break;
 
 	case LDAP_FILTER_AND:
 	case LDAP_FILTER_OR:
 	case LDAP_FILTER_NOT:
-		fstr->bv_len = sizeof("(%)") - 1;
+		fstr->bv_len = STRLENOF("(%)");
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 128, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%c)",
@@ -712,7 +707,8 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 			fstr->bv_val = op->o_tmprealloc( fstr->bv_val, fstr->bv_len + 1,
 				op->o_tmpmemctx );
 
-			snprintf( &fstr->bv_val[len-1], tmp.bv_len + 2, 
+			snprintf( &fstr->bv_val[len-1],
+				tmp.bv_len + STRLENOF( /*(*/ ")" ) + 1, 
 				/*"("*/ "%s)", tmp.bv_val );
 
 			op->o_tmpfree( tmp.bv_val, op->o_tmpmemctx );
@@ -721,29 +717,35 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 		break;
 
 	case LDAP_FILTER_EXT: {
-			struct berval ad;
-			filter_escape_value_x( &f->f_mr_value, &tmp, op->o_tmpmemctx );
+		struct berval ad;
 
-			if ( f->f_mr_desc ) {
-				ad = f->f_mr_desc->ad_cname;
-			} else {
-				ad.bv_len = 0;
-				ad.bv_val = "";
-			}
-			
-			fstr->bv_len = ad.bv_len +
-				( f->f_mr_dnattrs ? sizeof(":dn")-1 : 0 ) +
-				( f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_len+1 : 0 ) +
-				tmp.bv_len + ( sizeof("(:=)") - 1 );
-			fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
+		filter_escape_value_x( &f->f_mr_value, &tmp, op->o_tmpmemctx );
+		/* NOTE: tmp can legitimately be NULL (meaning empty) 
+		 * since in a Filter values in MRAs are supposed
+		 * to have been normalized, meaning that an empty value
+		 * is legal for that attribute's syntax */
 
-			snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s:=%s)",
-				ad.bv_val,
-				f->f_mr_dnattrs ? ":dn" : "",
-				f->f_mr_rule_text.bv_len ? ":" : "",
-				f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_val : "",
-				tmp.bv_val );
-			ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
+		if ( f->f_mr_desc ) {
+			ad = f->f_mr_desc->ad_cname;
+		} else {
+			ad.bv_len = 0;
+			ad.bv_val = "";
+		}
+		
+		fstr->bv_len = ad.bv_len +
+			( f->f_mr_dnattrs ? STRLENOF(":dn") : 0 ) +
+			( f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_len+1 : 0 ) +
+			tmp.bv_len + STRLENOF("(:=)");
+		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
+
+		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s%s:=%s)",
+			undef ? "?" : "",
+			ad.bv_val,
+			f->f_mr_dnattrs ? ":dn" : "",
+			f->f_mr_rule_text.bv_len ? ":" : "",
+			f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_val : "",
+			tmp.bv_len ? tmp.bv_val : "" );
+		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		} break;
 
 	case SLAPD_FILTER_COMPUTED:
@@ -787,40 +789,89 @@ filter2bv( Filter *f, struct berval *fstr )
 	filter2bv_x( &op, f, fstr );
 }
 
-static int
-filter_escape_value_x(
-	struct berval *in,
-	struct berval *out,
-	void *ctx )
+Filter *
+filter_dup( Filter *f, void *memctx )
 {
-	ber_len_t i;
-	assert( in );
-	assert( out );
+	BerMemoryFunctions *mf = &slap_sl_mfuncs;
+	Filter *n;
 
-	i = in->bv_len * 3 + 1;
-	out->bv_val = ctx ? slap_sl_malloc( i, ctx ) : ch_malloc( i );
-	out->bv_len = 0;
+	if ( !f )
+		return NULL;
 
-	for( i=0; i < in->bv_len ; i++ ) {
-		if( FILTER_ESCAPE(in->bv_val[i]) ) {
-			out->bv_val[out->bv_len++] = SLAP_ESCAPE_CHAR;
-			out->bv_val[out->bv_len++] = SLAP_ESCAPE_HI( in->bv_val[i] );
-			out->bv_val[out->bv_len++] = SLAP_ESCAPE_LO( in->bv_val[i] );
-		} else {
-			out->bv_val[out->bv_len++] = in->bv_val[i];
+	n = mf->bmf_malloc( sizeof(Filter), memctx );
+	n->f_choice = f->f_choice;
+	n->f_next = NULL;
+
+	switch( f->f_choice & SLAPD_FILTER_MASK ) {
+	case SLAPD_FILTER_COMPUTED:
+		n->f_result = f->f_result;
+		break;
+	case LDAP_FILTER_PRESENT:
+		if ( f->f_desc->ad_flags & SLAP_DESC_TEMPORARY )
+			n->f_desc = slap_bv2tmp_ad( &f->f_desc->ad_cname, memctx );
+		else
+			n->f_desc = f->f_desc;
+		break;
+	case LDAP_FILTER_EQUALITY:
+	case LDAP_FILTER_GE:
+	case LDAP_FILTER_LE:
+	case LDAP_FILTER_APPROX:
+		/* Should this be ava_dup() ? */
+		n->f_ava = mf->bmf_calloc( 1, sizeof(AttributeAssertion), memctx );
+		*n->f_ava = *f->f_ava;
+		if ( f->f_av_desc->ad_flags & SLAP_DESC_TEMPORARY )
+			n->f_av_desc = slap_bv2tmp_ad( &f->f_av_desc->ad_cname, memctx );
+		ber_dupbv_x( &n->f_av_value, &f->f_av_value, memctx );
+		break;
+	case LDAP_FILTER_SUBSTRINGS:
+		n->f_sub = mf->bmf_calloc( 1, sizeof(SubstringsAssertion), memctx );
+		if ( f->f_sub_desc->ad_flags & SLAP_DESC_TEMPORARY )
+			n->f_sub_desc = slap_bv2tmp_ad( &f->f_sub_desc->ad_cname, memctx );
+		else
+			n->f_sub_desc = f->f_sub_desc;
+		if ( !BER_BVISNULL( &f->f_sub_initial ))
+			ber_dupbv_x( &n->f_sub_initial, &f->f_sub_initial, memctx );
+		if ( f->f_sub_any ) {
+			int i;
+			for ( i = 0; !BER_BVISNULL( &f->f_sub_any[i] ); i++ );
+			n->f_sub_any = mf->bmf_malloc(( i+1 )*sizeof( struct berval ),
+				memctx );
+			for ( i = 0; !BER_BVISNULL( &f->f_sub_any[i] ); i++ ) {
+				ber_dupbv_x( &n->f_sub_any[i], &f->f_sub_any[i], memctx );
+			}
+			BER_BVZERO( &n->f_sub_any[i] );
 		}
+		if ( !BER_BVISNULL( &f->f_sub_final ))
+			ber_dupbv_x( &n->f_sub_final, &f->f_sub_final, memctx );
+		break;
+	case LDAP_FILTER_EXT: {
+		/* Should this be mra_dup() ? */
+		ber_len_t length;
+		length = sizeof(MatchingRuleAssertion);
+		if ( !BER_BVISNULL( &f->f_mr_rule_text ))
+			length += f->f_mr_rule_text.bv_len + 1;
+		n->f_mra = mf->bmf_calloc( 1, length, memctx );
+		*n->f_mra = *f->f_mra;
+		if ( f->f_mr_desc && ( f->f_sub_desc->ad_flags & SLAP_DESC_TEMPORARY ))
+			n->f_mr_desc = slap_bv2tmp_ad( &f->f_mr_desc->ad_cname, memctx );
+		ber_dupbv_x( &n->f_mr_value, &f->f_mr_value, memctx );
+		if ( !BER_BVISNULL( &f->f_mr_rule_text )) {
+			n->f_mr_rule_text.bv_val = (char *)(n->f_mra+1);
+			AC_MEMCPY(n->f_mr_rule_text.bv_val,
+				f->f_mr_rule_text.bv_val, f->f_mr_rule_text.bv_len );
+		}
+		} break;
+	case LDAP_FILTER_AND:
+	case LDAP_FILTER_OR:
+	case LDAP_FILTER_NOT: {
+		Filter **p;
+		for ( p = &n->f_list, f = f->f_list; f; f = f->f_next ) {
+			*p = filter_dup( f, memctx );
+			p = &(*p)->f_next;
+		}
+		} break;
 	}
-
-	out->bv_val[out->bv_len] = '\0';
-	return LDAP_SUCCESS;
-}
-
-int
-filter_escape_value(
-	struct berval *in,
-	struct berval *out )
-{
-	return filter_escape_value_x( in, out, NULL );
+	return n;
 }
 
 static int
@@ -852,7 +903,7 @@ get_simple_vrFilter(
 	switch ( vrf.vrf_choice ) {
 	case LDAP_FILTER_EQUALITY:
 		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n", 0, 0, 0 );
-		err = get_ava( op, ber, &vrf.vrf_ava, SLAP_MR_EQUALITY, text );
+		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_EQUALITY, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -862,12 +913,12 @@ get_simple_vrFilter(
 
 	case LDAP_FILTER_SUBSTRINGS:
 		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n", 0, 0, 0 );
-		err = get_ssa( op, ber, &vrf.vrf_sub, text );
+		err = get_ssa( op, ber, (Filter *)&vrf, text );
 		break;
 
 	case LDAP_FILTER_GE:
 		Debug( LDAP_DEBUG_FILTER, "GE\n", 0, 0, 0 );
-		err = get_ava( op, ber, &vrf.vrf_ava, SLAP_MR_ORDERING, text );
+		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -875,7 +926,7 @@ get_simple_vrFilter(
 
 	case LDAP_FILTER_LE:
 		Debug( LDAP_DEBUG_FILTER, "LE\n", 0, 0, 0 );
-		err = get_ava( op, ber, &vrf.vrf_ava, SLAP_MR_ORDERING, text );
+		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -895,22 +946,28 @@ get_simple_vrFilter(
 		err = slap_bv2ad( &type, &vrf.vrf_desc, text );
 
 		if( err != LDAP_SUCCESS ) {
-			/* unrecognized attribute description or other error */
-			Debug( LDAP_DEBUG_ANY, 
-				"get_simple_vrFilter: conn %d unknown "
-				"attribute type=%s (%d)\n",
-				op->o_connid, type.bv_val, err );
+			vrf.vrf_choice |= SLAPD_FILTER_UNDEFINED;
+			err = slap_bv2undef_ad( &type, &vrf.vrf_desc, text,
+				SLAP_AD_PROXIED);
 
-			vrf.vrf_choice = SLAPD_FILTER_COMPUTED;
-			vrf.vrf_result = LDAP_COMPARE_FALSE;
-			err = LDAP_SUCCESS;
-			break;
+			if( err != LDAP_SUCCESS ) {
+				/* unrecognized attribute description or other error */
+				Debug( LDAP_DEBUG_ANY, 
+					"get_simple_vrFilter: conn %lu unknown "
+					"attribute type=%s (%d)\n",
+					op->o_connid, type.bv_val, err );
+	
+				vrf.vrf_choice = SLAPD_FILTER_COMPUTED;
+				vrf.vrf_result = LDAP_COMPARE_FALSE;
+				err = LDAP_SUCCESS;
+				break;
+			}
 		}
 		} break;
 
 	case LDAP_FILTER_APPROX:
 		Debug( LDAP_DEBUG_FILTER, "APPROX\n", 0, 0, 0 );
-		err = get_ava( op, ber, &vrf.vrf_ava, SLAP_MR_EQUALITY_APPROX, text );
+		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_EQUALITY_APPROX, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -919,7 +976,7 @@ get_simple_vrFilter(
 	case LDAP_FILTER_EXT:
 		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n", 0, 0, 0 );
 
-		err = get_mra( op, ber, &vrf.vrf_mra, text );
+		err = get_mra( op, ber, (Filter *)&vrf, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -944,7 +1001,7 @@ get_simple_vrFilter(
 	}
 
 	if ( err == LDAP_SUCCESS ) {
-		*filt = ch_malloc( sizeof vrf );
+		*filt = op->o_tmpalloc( sizeof vrf, op->o_tmpmemctx );
 		**filt = vrf;
 	}
 
@@ -1035,7 +1092,7 @@ vrFilter_free( Operation *op, ValuesReturnFilter *vrf )
 	for ( p = vrf; p != NULL; p = next ) {
 		next = p->vrf_next;
 
-		switch ( vrf->vrf_choice ) {
+		switch ( vrf->vrf_choice & SLAPD_FILTER_MASK ) {
 		case LDAP_FILTER_PRESENT:
 			break;
 
@@ -1082,12 +1139,12 @@ vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr )
 	ber_len_t len;
 
 	if ( vrf == NULL ) {
-		ber_str2bv_x( "No filter!", sizeof("No filter!")-1,
+		ber_str2bv_x( "No filter!", STRLENOF("No filter!"),
 			1, fstr, op->o_tmpmemctx );
 		return;
 	}
 
-	fstr->bv_len = sizeof("()") - 1;
+	fstr->bv_len = STRLENOF("()");
 	fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 128, op->o_tmpmemctx );
 
 	snprintf( fstr->bv_val, fstr->bv_len + 1, "()");
@@ -1113,19 +1170,22 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 {
 	struct berval tmp;
 	ber_len_t len;
+	int undef;
 
 	if ( vrf == NULL ) {
-		ber_str2bv_x( "No filter!", sizeof("No filter!")-1, 1, fstr,
+		ber_str2bv_x( "No filter!", STRLENOF("No filter!"), 1, fstr,
 			op->o_tmpmemctx );
 		return;
 	}
+	undef = vrf->vrf_choice & SLAPD_FILTER_UNDEFINED;
 
-	switch ( vrf->vrf_choice ) {
+	switch ( vrf->vrf_choice & SLAPD_FILTER_MASK ) {
 	case LDAP_FILTER_EQUALITY:
 		filter_escape_value_x( &vrf->vrf_av_value, &tmp, op->o_tmpmemctx );
 
 		fstr->bv_len = vrf->vrf_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(=)") - 1 );
+			tmp.bv_len + STRLENOF("(=)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=%s)",
@@ -1139,7 +1199,8 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 		filter_escape_value_x( &vrf->vrf_av_value, &tmp, op->o_tmpmemctx );
 
 		fstr->bv_len = vrf->vrf_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(>=)") - 1 );
+			tmp.bv_len + STRLENOF("(>=)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s>=%s)",
@@ -1153,7 +1214,8 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 		filter_escape_value_x( &vrf->vrf_av_value, &tmp, op->o_tmpmemctx );
 
 		fstr->bv_len = vrf->vrf_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(<=)") - 1 );
+			tmp.bv_len + STRLENOF("(<=)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s<=%s)",
@@ -1167,7 +1229,8 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 		filter_escape_value_x( &vrf->vrf_av_value, &tmp, op->o_tmpmemctx );
 
 		fstr->bv_len = vrf->vrf_av_desc->ad_cname.bv_len +
-			tmp.bv_len + ( sizeof("(~=)") - 1 );
+			tmp.bv_len + STRLENOF("(~=)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s~=%s)",
@@ -1178,7 +1241,8 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 
 	case LDAP_FILTER_SUBSTRINGS:
 		fstr->bv_len = vrf->vrf_sub_desc->ad_cname.bv_len +
-			( sizeof("(=*)") - 1 );
+			STRLENOF("(=*)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 128, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=*)",
@@ -1238,7 +1302,8 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 
 	case LDAP_FILTER_PRESENT:
 		fstr->bv_len = vrf->vrf_desc->ad_cname.bv_len +
-			( sizeof("(=*)") - 1 );
+			STRLENOF("(=*)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s=*)",
@@ -1257,10 +1322,11 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 		}
 			
 		fstr->bv_len = ad.bv_len +
-			( vrf->vrf_mr_dnattrs ? sizeof(":dn")-1 : 0 ) +
+			( vrf->vrf_mr_dnattrs ? STRLENOF(":dn") : 0 ) +
 			( vrf->vrf_mr_rule_text.bv_len
 				? vrf->vrf_mr_rule_text.bv_len+1 : 0 ) +
-			tmp.bv_len + ( sizeof("(:=)") - 1 );
+			tmp.bv_len + STRLENOF("(:=)");
+		if ( undef ) fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s:=%s)",
@@ -1279,15 +1345,15 @@ simple_vrFilter2bv( Operation *op, ValuesReturnFilter *vrf, struct berval *fstr
 			vrf->vrf_result == LDAP_COMPARE_TRUE ? "(?=true)" :
 			vrf->vrf_result == SLAPD_COMPARE_UNDEFINED
 				? "(?=undefined)" : "(?=error)",
-			vrf->vrf_result == LDAP_COMPARE_FALSE ? sizeof("(?=false)")-1 :
-			vrf->vrf_result == LDAP_COMPARE_TRUE ? sizeof("(?=true)")-1 :
+			vrf->vrf_result == LDAP_COMPARE_FALSE ? STRLENOF("(?=false)") :
+			vrf->vrf_result == LDAP_COMPARE_TRUE ? STRLENOF("(?=true)") :
 			vrf->vrf_result == SLAPD_COMPARE_UNDEFINED
-				? sizeof("(?=undefined)")-1 : sizeof("(?=error)")-1,
+				? STRLENOF("(?=undefined)") : STRLENOF("(?=error)"),
 			1, fstr, op->o_tmpmemctx );
 		break;
 
 	default:
-		ber_str2bv_x( "(?=unknown)", sizeof("(?=unknown)")-1,
+		ber_str2bv_x( "(?=unknown)", STRLENOF("(?=unknown)"),
 			1, fstr, op->o_tmpmemctx );
 		break;
 	}
