@@ -1,9 +1,18 @@
-/*
- * Copyright 1998-2003 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+/* $OpenLDAP: pkg/ldap/servers/slapd/mods.c,v 1.59.2.2 2007/08/31 23:13:59 quanah Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 1998-2007 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
-/*
- * Copyright (c) 1995 Regents of the University of Michigan.
+/* Portions Copyright (c) 1995 Regents of the University of Michigan.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms are permitted
@@ -19,38 +28,23 @@
 #include <ac/string.h>
 
 #include "slap.h"
-
-int
-modify_check_duplicates(
-	AttributeDescription	*ad,
-	MatchingRule		*mr,
-	BerVarray		vals,
-	BerVarray		mods,
-	int			permissive,
-	const char	**text,
-	char *textbuf, size_t textlen )
-{
-	int		i, j, numvals = 0, nummods,
-			rc = LDAP_SUCCESS, matched;
-	/* this function is no longer used */
-	return rc;
-}
+#include "lutil.h"
 
 int
 modify_add_values(
-	Entry	*e,
+	Entry		*e,
 	Modification	*mod,
-	int	permissive,
+	int		permissive,
 	const char	**text,
-	char *textbuf, size_t textlen )
+	char		*textbuf,
+	size_t		textlen )
 {
-	int		i, j;
-	int		matched;
+	int		rc;
+	const char	*op;
 	Attribute	*a;
-	MatchingRule *mr = mod->sm_desc->ad_type->sat_equality;
-	const char *op;
+	Modification	pmod = *mod;
 
-	switch( mod->sm_op ) {
+	switch ( mod->sm_op ) {
 	case LDAP_MOD_ADD:
 		op = "add";
 		break;
@@ -62,17 +56,14 @@ modify_add_values(
 		assert( 0 );
 	}
 
+	/* check if values to add exist in attribute */
 	a = attr_find( e->e_attrs, mod->sm_desc );
+	if ( a != NULL ) {
+		int		rc, i, j, p;
+		MatchingRule	*mr;
 
-	/*
-	 * With permissive set, as long as the attribute being added
-	 * has the same value(s?) as the existing attribute, then the
-	 * modify will succeed.
-	 */
-
-	/* check if the values we're adding already exist */
-	if( mr == NULL || !mr->smr_match ) {
-		if ( a != NULL ) {
+		mr = mod->sm_desc->ad_type->sat_equality;
+		if( mr == NULL || !mr->smr_match ) {
 			/* do not allow add of additional attribute
 				if no equality rule exists */
 			*text = textbuf;
@@ -82,116 +73,94 @@ modify_add_values(
 			return LDAP_INAPPROPRIATE_MATCHING;
 		}
 
-		for ( i = 0; mod->sm_bvalues[i].bv_val != NULL; i++ ) {
-			/* test asserted values against existing values */
-			if( a ) {
-				for( matched = 0, j = 0; a->a_vals[j].bv_val != NULL; j++ ) {
-					if ( bvmatch( &mod->sm_bvalues[i], &a->a_vals[j] ) ) {
-						if ( permissive ) {
-							matched++;
-							continue;
-						}
-						/* value exists already */
-						*text = textbuf;
-						snprintf( textbuf, textlen,
-							"modify/%s: %s: value #%i already exists",
-							op, mod->sm_desc->ad_cname.bv_val, j );
-						return LDAP_TYPE_OR_VALUE_EXISTS;
-					}
-				}
-				if ( permissive && matched == j ) {
-					/* values already exist; do nothing */
-					return LDAP_SUCCESS;
-				}
+		if ( permissive ) {
+			for ( i = 0; !BER_BVISNULL( &mod->sm_values[i] ); i++ ) {
+				/* EMPTY -- just counting 'em */;
 			}
 
-			/* test asserted values against themselves */
-			for( j = 0; j < i; j++ ) {
-				if ( bvmatch( &mod->sm_bvalues[i], &mod->sm_bvalues[j] ) ) {
-					/* value exists already */
-					*text = textbuf;
-					snprintf( textbuf, textlen,
-						"modify/%s: %s: value #%i already exists",
-						op, mod->sm_desc->ad_cname.bv_val, j );
-					return LDAP_TYPE_OR_VALUE_EXISTS;
-				}
+			pmod.sm_values = (BerVarray)ch_malloc(
+				(i + 1) * sizeof( struct berval ));
+			if ( pmod.sm_nvalues != NULL ) {
+				pmod.sm_nvalues = (BerVarray)ch_malloc(
+					(i + 1) * sizeof( struct berval ));
 			}
 		}
 
-	} else {
 		/* no normalization is done in this routine nor
 		 * in the matching routines called by this routine. 
 		 * values are now normalized once on input to the
 		 * server (whether from LDAP or from the underlying
 		 * database).
-		 * This should outperform the old code.  No numbers
-		 * are available yet.
 		 */
+		for ( p = i = 0; !BER_BVISNULL( &mod->sm_values[i] ); i++ ) {
+			int	match;
 
-		int		rc;
-
-		if ( mod->sm_bvalues[1].bv_val == 0 ) {
-			if ( a != NULL ) {
-				int		i;
-
-				for ( matched = 0, i = 0; a->a_vals[ i ].bv_val; i++ ) {
-					int	match;
-
-					if( mod->sm_nvalues ) {
-						rc = value_match( &match, mod->sm_desc, mr,
-							SLAP_MR_EQUALITY
-								| SLAP_MR_VALUE_OF_ASSERTION_SYNTAX
-								| SLAP_MR_ASSERTED_VALUE_NORMALIZED_MATCH
-								| SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
-							&a->a_nvals[i],
-							&mod->sm_nvalues[0],
-							text );
-
-					} else {
-						rc = value_match( &match, mod->sm_desc, mr,
-							SLAP_MR_EQUALITY
-								| SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
-							&a->a_vals[i],
-							&mod->sm_values[0],
-							text );
-					}
-
-
-					if( rc == LDAP_SUCCESS && match == 0 ) {
-						if ( permissive ) {
-							matched++;
-							continue;
-						}
-						*text = textbuf;
-						snprintf( textbuf, textlen,
-							"modify/%s: %s: value #0 already exists",
-							op, mod->sm_desc->ad_cname.bv_val );
-						return LDAP_TYPE_OR_VALUE_EXISTS;
-					}
+			assert( a->a_vals[0].bv_val != NULL );
+			for ( j = 0; !BER_BVISNULL( &a->a_vals[j] ); j++ ) {
+				if ( mod->sm_nvalues ) {
+					rc = ordered_value_match( &match, mod->sm_desc, mr,
+						SLAP_MR_EQUALITY | SLAP_MR_VALUE_OF_ASSERTION_SYNTAX
+							| SLAP_MR_ASSERTED_VALUE_NORMALIZED_MATCH
+							| SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
+						&a->a_nvals[j], &mod->sm_nvalues[i], text );
+				} else {
+					rc = ordered_value_match( &match, mod->sm_desc, mr,
+						SLAP_MR_EQUALITY | SLAP_MR_VALUE_OF_ASSERTION_SYNTAX,
+						&a->a_vals[j], &mod->sm_values[i], text );
 				}
-				if ( permissive && matched == i ) {
-					/* values already exist; do nothing */
-					return LDAP_SUCCESS;
+
+				if ( rc == LDAP_SUCCESS && match == 0 ) {
+					/* value already exists */
+					if ( permissive ) break;
+
+					*text = textbuf;
+					snprintf( textbuf, textlen,
+						"modify/%s: %s: value #%d already exists",
+						op, mod->sm_desc->ad_cname.bv_val, i );
+					return LDAP_TYPE_OR_VALUE_EXISTS;
+
+				} else if ( rc != LDAP_SUCCESS ) {
+					return rc;
 				}
 			}
 
-		} else {
-			rc = modify_check_duplicates( mod->sm_desc, mr,
-				a ? a->a_vals : NULL, mod->sm_bvalues,
-				permissive, text, textbuf, textlen );
+			if ( permissive && match != 0 ) {
+				if ( pmod.sm_nvalues ) {
+					pmod.sm_nvalues[p] = mod->sm_nvalues[i];
+				}
+				pmod.sm_values[p++] = mod->sm_values[i];
+			}
+		}
 
-			if ( permissive && rc == LDAP_TYPE_OR_VALUE_EXISTS ) {
+		if ( permissive ) {
+			if ( p == 0 ) {
+				/* all new values match exist */
+				ch_free( pmod.sm_values );
+				if ( pmod.sm_nvalues ) ch_free( pmod.sm_nvalues );
 				return LDAP_SUCCESS;
 			}
 
-			if ( rc != LDAP_SUCCESS ) {
-				return rc;
+			BER_BVZERO( &pmod.sm_values[p] );
+			if ( pmod.sm_nvalues ) {
+				BER_BVZERO( &pmod.sm_nvalues[p] );
 			}
 		}
 	}
 
 	/* no - add them */
-	if( attr_merge( e, mod->sm_desc, mod->sm_values, mod->sm_nvalues ) != 0 ) {
+	if ( mod->sm_desc->ad_type->sat_flags & SLAP_AT_ORDERED_VAL ) {
+		rc = ordered_value_add( e, mod->sm_desc, a,
+			pmod.sm_values, pmod.sm_nvalues );
+	} else {
+		rc = attr_merge( e, mod->sm_desc, pmod.sm_values, pmod.sm_nvalues );
+	}
+
+	if ( a != NULL && permissive ) {
+		ch_free( pmod.sm_values );
+		if ( pmod.sm_nvalues ) ch_free( pmod.sm_nvalues );
+	}
+
+	if ( rc != 0 ) {
 		/* this should return result of attr_merge */
 		*text = textbuf;
 		snprintf( textbuf, textlen,
@@ -205,17 +174,29 @@ modify_add_values(
 
 int
 modify_delete_values(
+	Entry	*e,
+	Modification	*m,
+	int	perm,
+	const char	**text,
+	char *textbuf, size_t textlen )
+{
+	return modify_delete_vindex( e, m, perm, text, textbuf, textlen, NULL );
+}
+
+int
+modify_delete_vindex(
 	Entry	*e,
 	Modification	*mod,
 	int	permissive,
 	const char	**text,
-	char *textbuf, size_t textlen
-)
+	char *textbuf, size_t textlen,
+	int *idx )
 {
 	int		i, j, k, rc = LDAP_SUCCESS;
 	Attribute	*a;
 	MatchingRule 	*mr = mod->sm_desc->ad_type->sat_equality;
 	char		dummy = '\0';
+	int		match = 0;
 
 	/*
 	 * If permissive is set, then the non-existence of an 
@@ -223,7 +204,7 @@ modify_delete_values(
 	 */
 
 	/* delete the entire attribute */
-	if ( mod->sm_bvalues == NULL ) {
+	if ( mod->sm_values == NULL ) {
 		rc = attr_delete( &e->e_attrs, mod->sm_desc );
 
 		if( permissive ) {
@@ -260,30 +241,28 @@ modify_delete_values(
 		return LDAP_NO_SUCH_ATTRIBUTE;
 	}
 
-
-	for ( i = 0; mod->sm_values[i].bv_val != NULL; i++ ) {
+	for ( i = 0; !BER_BVISNULL( &mod->sm_values[i] ); i++ ) {
 		int	found = 0;
-		for ( j = 0; a->a_vals[j].bv_val != NULL; j++ ) {
-			int match;
+		for ( j = 0; !BER_BVISNULL( &a->a_vals[j] ); j++ ) {
+			/* skip already deleted values */
+			if ( a->a_vals[j].bv_val == &dummy ) {
+				continue;
+			}
 
 			if( mod->sm_nvalues ) {
-				assert( a->a_nvals );
-				rc = (*mr->smr_match)( &match,
-					SLAP_MR_VALUE_OF_ASSERTION_SYNTAX
+				assert( a->a_nvals != NULL );
+				rc = ordered_value_match( &match, a->a_desc, mr,
+					SLAP_MR_EQUALITY | SLAP_MR_VALUE_OF_ASSERTION_SYNTAX
 						| SLAP_MR_ASSERTED_VALUE_NORMALIZED_MATCH
 						| SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
-					a->a_desc->ad_type->sat_syntax,
-					mr, &a->a_nvals[j],
-					&mod->sm_nvalues[i] );
+					&a->a_nvals[j], &mod->sm_nvalues[i], text );
 			} else {
 #if 0
 				assert( a->a_nvals == NULL );
 #endif
-				rc = (*mr->smr_match)( &match,
-					SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
-					a->a_desc->ad_type->sat_syntax,
-					mr, &a->a_vals[j],
-					&mod->sm_values[i] );
+				rc = ordered_value_match( &match, a->a_desc, mr,
+					SLAP_MR_EQUALITY | SLAP_MR_VALUE_OF_ASSERTION_SYNTAX,
+					&a->a_vals[j], &mod->sm_values[i], text );
 			}
 
 			if ( rc != LDAP_SUCCESS ) {
@@ -291,7 +270,7 @@ modify_delete_values(
 				snprintf( textbuf, textlen,
 					"%s: matching rule failed",
 					mod->sm_desc->ad_cname.bv_val );
-				goto return_results;
+				break;
 			}
 
 			if ( match != 0 ) {
@@ -300,6 +279,9 @@ modify_delete_values(
 
 			found = 1;
 
+			if ( idx )
+				idx[i] = j;
+
 			/* delete value and mark it as dummy */
 			free( a->a_vals[j].bv_val );
 			a->a_vals[j].bv_val = &dummy;
@@ -317,15 +299,19 @@ modify_delete_values(
 				"modify/delete: %s: no such value",
 				mod->sm_desc->ad_cname.bv_val );
 			rc = LDAP_NO_SUCH_ATTRIBUTE;
-			goto return_results;
+			if ( i > 0 ) {
+				break;
+			} else {
+				goto return_results;
+			}
 		}
 	}
 
 	/* compact array skipping dummies */
-	for ( k = 0, j = 0; a->a_vals[k].bv_val != NULL; k++ ) {
+	for ( k = 0, j = 0; !BER_BVISNULL( &a->a_vals[k] ); k++ ) {
 		/* skip dummies */
 		if( a->a_vals[k].bv_val == &dummy ) {
-			assert( a->a_nvals == NULL || a->a_nvals[k].bv_val == &dummy );
+			assert( a->a_nvals[k].bv_val == &dummy );
 			continue;
 		}
 		if ( j != k ) {
@@ -338,11 +324,13 @@ modify_delete_values(
 		j++;
 	}
 
-	a->a_vals[j].bv_val = NULL;
-	if (a->a_nvals != a->a_vals) a->a_nvals[j].bv_val = NULL;
+	BER_BVZERO( &a->a_vals[j] );
+	if (a->a_nvals != a->a_vals) {
+		BER_BVZERO( &a->a_nvals[j] );
+	}
 
 	/* if no values remain, delete the entire attribute */
-	if ( a->a_vals[0].bv_val == NULL ) {
+	if ( BER_BVISNULL( &a->a_vals[0] ) ) {
 		if ( attr_delete( &e->e_attrs, mod->sm_desc ) ) {
 			*text = textbuf;
 			snprintf( textbuf, textlen,
@@ -350,6 +338,9 @@ modify_delete_values(
 				mod->sm_desc->ad_cname.bv_val );
 			rc = LDAP_NO_SUCH_ATTRIBUTE;
 		}
+	} else if ( a->a_desc->ad_type->sat_flags & SLAP_AT_ORDERED_VAL ) {
+		/* For an ordered attribute, renumber the value indices */
+		ordered_value_sort( a, 1 );
 	}
 
 return_results:;
@@ -367,7 +358,7 @@ modify_replace_values(
 {
 	(void) attr_delete( &e->e_attrs, mod->sm_desc );
 
-	if ( mod->sm_bvalues ) {
+	if ( mod->sm_values ) {
 		return modify_add_values( e, mod, permissive, text, textbuf, textlen );
 	}
 
@@ -386,33 +377,50 @@ modify_increment_values(
 
 	a = attr_find( e->e_attrs, mod->sm_desc );
 	if( a == NULL ) {
-		*text = textbuf;
-		snprintf( textbuf, textlen,
-			"modify/increment: %s: no such attribute",
-			mod->sm_desc->ad_cname.bv_val );
-		return LDAP_NO_SUCH_ATTRIBUTE;
-	}
+		if ( permissive ) {
+			Modification modReplace = *mod;
 
+			modReplace.sm_op = LDAP_MOD_REPLACE;
+
+			return modify_add_values(e, &modReplace, permissive, text, textbuf, textlen);
+		} else {
+			*text = textbuf;
+			snprintf( textbuf, textlen,
+				"modify/increment: %s: no such attribute",
+				mod->sm_desc->ad_cname.bv_val );
+			return LDAP_NO_SUCH_ATTRIBUTE;
+		}
+	}
 
 	if ( !strcmp( a->a_desc->ad_type->sat_syntax_oid, SLAPD_INTEGER_SYNTAX )) {
 		int i;
 		char str[sizeof(long)*3 + 2]; /* overly long */
-		long incr = atol( mod->sm_bvalues[0].bv_val );
+		long incr;
+
+		if ( lutil_atol( &incr, mod->sm_values[0].bv_val ) != 0 ) {
+			*text = "modify/increment: invalid syntax of increment";
+			return LDAP_INVALID_SYNTAX;
+		}
 
 		/* treat zero and errors as a no-op */
 		if( incr == 0 ) {
 			return LDAP_SUCCESS;
 		}
 
-		for( i=0; a->a_nvals[i].bv_val != NULL; i++ ) {
+		for( i = 0; !BER_BVISNULL( &a->a_nvals[i] ); i++ ) {
 			char *tmp;
-			long value = atol( a->a_nvals[i].bv_val );
-			size_t strln = snprintf( str, sizeof(str), "%ld", value+incr );
+			long value;
+			size_t strln;
+			if ( lutil_atol( &value, a->a_nvals[i].bv_val ) != 0 ) {
+				*text = "modify/increment: invalid syntax of original value";
+				return LDAP_INVALID_SYNTAX;
+			}
+			strln = snprintf( str, sizeof(str), "%ld", value+incr );
 
 			tmp = SLAP_REALLOC( a->a_nvals[i].bv_val, strln+1 );
 			if( tmp == NULL ) {
 				*text = "modify/increment: reallocation error";
-				return LDAP_OTHER;;
+				return LDAP_OTHER;
 			}
 			a->a_nvals[i].bv_val = tmp;
 			a->a_nvals[i].bv_len = strln;
@@ -447,14 +455,16 @@ slap_mod_free(
 
 void
 slap_mods_free(
-    Modifications	*ml )
+    Modifications	*ml,
+    int			freevals )
 {
 	Modifications *next;
 
 	for ( ; ml != NULL; ml = next ) {
 		next = ml->sml_next;
 
-		slap_mod_free( &ml->sml_mod, 0 );
+		if ( freevals )
+			slap_mod_free( &ml->sml_mod, 0 );
 		free( ml );
 	}
 }
