@@ -1,8 +1,18 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/module.c,v 1.21.2.1 2003/09/18 15:43:34 kurt Exp $ */
-/*
- * Copyright 1998-2003 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+/* $OpenLDAP: pkg/ldap/servers/slapd/module.c,v 1.29.2.2 2007/08/31 23:13:59 quanah Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 1998-2007 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
+
 #include "portable.h"
 #include <stdio.h>
 #include "slap.h"
@@ -34,11 +44,12 @@ struct module_regtable_t {
 typedef struct module_loaded_t {
 	struct module_loaded_t *next;
 	lt_dlhandle lib;
+	char name[1];
 } module_loaded_t;
 
 module_loaded_t *module_list = NULL;
 
-static int module_unload (module_loaded_t *module);
+static int module_int_unload (module_loaded_t *module);
 
 #ifdef HAVE_EBCDIC
 static char ebuf[BUFSIZ];
@@ -53,23 +64,19 @@ int module_init (void)
 		__etoa( ebuf );
 		error = ebuf;
 #endif
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, 
-			"module_init: lt_dlinit failed: %s\n", error, 0, 0 );
-#else
 		Debug(LDAP_DEBUG_ANY, "lt_dlinit failed: %s\n", error, 0, 0);
-#endif
 
 		return -1;
 	}
-	return 0;
+
+	return module_path( LDAP_MODULEDIR );
 }
 
 int module_kill (void)
 {
 	/* unload all modules before shutdown */
 	while (module_list != NULL) {
-		module_unload(module_list);
+		module_int_unload(module_list);
 	}
 
 	if (lt_dlexit()) {
@@ -79,18 +86,37 @@ int module_kill (void)
 		__etoa( ebuf );
 		error = ebuf;
 #endif
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "module_kill: lt_dlexit failed: %s\n", 
-			error, 0, 0 );
-#else
 		Debug(LDAP_DEBUG_ANY, "lt_dlexit failed: %s\n", error, 0, 0);
-#endif
 
 		return -1;
 	}
 	return 0;
 }
 
+void * module_handle( const char *file_name )
+{
+	module_loaded_t *module;
+
+	for ( module = module_list; module; module= module->next ) {
+		if ( !strcmp( module->name, file_name )) {
+			return module;
+		}
+	}
+	return NULL;
+}
+
+int module_unload( const char *file_name )
+{
+	module_loaded_t *module;
+
+	module = module_handle( file_name );
+	if ( module ) {
+		module_int_unload( module );
+		return 0;
+	}
+	return -1;	/* not found */
+}
+
 int module_load(const char* file_name, int argc, char *argv[])
 {
 	module_loaded_t *module = NULL;
@@ -103,18 +129,15 @@ int module_load(const char* file_name, int argc, char *argv[])
 #define	file	file_name
 #endif
 
-	module = (module_loaded_t *)ch_calloc(1, sizeof(module_loaded_t));
+	module = (module_loaded_t *)ch_calloc(1, sizeof(module_loaded_t) +
+		strlen(file_name));
 	if (module == NULL) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, 
-			"module_load:  (%s) out of memory.\n", file_name, 0, 0 );
-#else
 		Debug(LDAP_DEBUG_ANY, "module_load failed: (%s) out of memory\n", file_name,
 			0, 0);
-#endif
 
 		return -1;
 	}
+	strcpy( module->name, file_name );
 
 #ifdef HAVE_EBCDIC
 	strcpy( file, file_name );
@@ -132,24 +155,14 @@ int module_load(const char* file_name, int argc, char *argv[])
 		__etoa( ebuf );
 		error = ebuf;
 #endif
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, 
-			"module_load: lt_dlopenext failed: (%s) %s.\n", 
-			file_name, error, 0 );
-#else
 		Debug(LDAP_DEBUG_ANY, "lt_dlopenext failed: (%s) %s\n", file_name,
 			error, 0);
-#endif
 
 		ch_free(module);
 		return -1;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, "module_load: loaded module %s\n", file_name, 0, 0 );
-#else
 	Debug(LDAP_DEBUG_CONFIG, "loaded module %s\n", file_name, 0, 0);
-#endif
 
    
 #ifdef HAVE_EBCDIC
@@ -159,14 +172,8 @@ int module_load(const char* file_name, int argc, char *argv[])
 #ifdef HAVE_EBCDIC
 #pragma convlit(resume)
 #endif
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, ERR, 
-			"module_load: module %s : no init_module() function found\n",
-		    file_name, 0, 0 );
-#else
 		Debug(LDAP_DEBUG_CONFIG, "module %s: no init_module() function found\n",
 			file_name, 0, 0);
-#endif
 
 		lt_dlclose(module->lib);
 		ch_free(module);
@@ -190,13 +197,8 @@ int module_load(const char* file_name, int argc, char *argv[])
 	 */
 	rc = initialize(argc, argv);
 	if (rc == -1) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, ERR, 
-			"module_load:  module %s init_module() failed\n", file_name, 0, 0);
-#else
 		Debug(LDAP_DEBUG_CONFIG, "module %s: init_module() failed\n",
 			file_name, 0, 0);
-#endif
 
 		lt_dlclose(module->lib);
 		ch_free(module);
@@ -206,45 +208,27 @@ int module_load(const char* file_name, int argc, char *argv[])
 	if (rc >= (int)(sizeof(module_regtable) / sizeof(struct module_regtable_t))
 		|| module_regtable[rc].proc == NULL)
 	{
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, ERR, 
-			"module_load: module %s: unknown registration type (%d).\n", 
-			file_name, rc, 0);
-#else
 		Debug(LDAP_DEBUG_CONFIG, "module %s: unknown registration type (%d)\n",
 			file_name, rc, 0);
-#endif
 
-		module_unload(module);
+		module_int_unload(module);
 		return -1;
 	}
 
 	rc = (module_regtable[rc].proc)(module, file_name);
 	if (rc != 0) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, ERR, 
-			"module_load: module %s:%s could not be registered.\n",
-			file_name, module_regtable[rc].type, 0 );
-#else
 		Debug(LDAP_DEBUG_CONFIG, "module %s: %s module could not be registered\n",
 			file_name, module_regtable[rc].type, 0);
-#endif
 
-		module_unload(module);
+		module_int_unload(module);
 		return rc;
 	}
 
 	module->next = module_list;
 	module_list = module;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, 
-		"module_load: module %s:%s registered\n", file_name,
-		module_regtable[rc].type, 0 );
-#else
 	Debug(LDAP_DEBUG_CONFIG, "module %s: %s module registered\n",
 		file_name, module_regtable[rc].type, 0);
-#endif
 
 	return 0;
 }
@@ -271,7 +255,7 @@ void *module_resolve (const void *module, const char *name)
 	return(lt_dlsym(((module_loaded_t *)module)->lib, name));
 }
 
-static int module_unload (module_loaded_t *module)
+static int module_int_unload (module_loaded_t *module)
 {
 	module_loaded_t *mod;
 	MODULE_TERM_FN terminate;
@@ -342,6 +326,8 @@ load_extop_module (
 		return(-1);
 	}
 
+	/* FIXME: this is broken, and no longer needed, 
+	 * as a module can call load_extop() itself... */
 	rc = load_extop( &oid, ext_main );
 	return rc;
 }
