@@ -1,8 +1,8 @@
 /* schema.c - routines to manage schema definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/schema.c,v 1.90.2.1 2003/12/01 00:58:36 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/schema.c,v 1.105.2.3 2007/08/31 23:14:00 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -23,7 +23,6 @@
 #include <ac/socket.h>
 
 #include "slap.h"
-#include "ldap_pvt.h"
 #include "lutil.h"
 
 
@@ -43,16 +42,11 @@ schema_info( Entry **entry, const char **text )
 	struct berval	vals[5];
 	struct berval	nvals[5];
 
-	e = (Entry *) SLAP_CALLOC( 1, sizeof(Entry) );
+	e = entry_alloc();
 	if( e == NULL ) {
 		/* Out of memory, do something about it */
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"schema_info: SLAP_CALLOC failed - out of memory.\n", 0, 0,0 );
-#else
 		Debug( LDAP_DEBUG_ANY, 
-			"schema_info: SLAP_CALLOC failed - out of memory.\n", 0, 0, 0 );
-#endif
+			"schema_info: entry_alloc failed - out of memory.\n", 0, 0, 0 );
 		*text = "out of memory";
 		return LDAP_OTHER;
 	}
@@ -61,31 +55,24 @@ schema_info( Entry **entry, const char **text )
 	/* backend-specific schema info should be created by the
 	 * backend itself
 	 */
-	ber_dupbv( &e->e_name, &global_schemadn );
-	ber_dupbv( &e->e_nname, &global_schemandn );
+	ber_dupbv( &e->e_name, &frontendDB->be_schemadn );
+	ber_dupbv( &e->e_nname, &frontendDB->be_schemandn );
 	e->e_private = NULL;
 
-	vals[0].bv_val = "subentry";
-	vals[0].bv_len = sizeof("subentry")-1;
-	if( attr_merge_one( e, ad_structuralObjectClass, vals, vals ) )
-	{
+	BER_BVSTR( &vals[0], "subentry" );
+	if( attr_merge_one( e, ad_structuralObjectClass, vals, NULL ) ) {
 		/* Out of memory, do something about it */
 		entry_free( e );
 		*text = "out of memory";
 		return LDAP_OTHER;
 	}
 
-	vals[0].bv_val = "top";
-	vals[0].bv_len = sizeof("top")-1;
-	vals[1].bv_val = "subentry";
-	vals[1].bv_len = sizeof("subentry")-1;
-	vals[2].bv_val = "subschema";
-	vals[2].bv_len = sizeof("subschema")-1;
-	vals[3].bv_val = "extensibleObject";
-	vals[3].bv_len = sizeof("extensibleObject")-1;
-	vals[4].bv_val = NULL;
-	if( attr_merge( e, ad_objectClass, vals, vals ) )
-	{
+	BER_BVSTR( &vals[0], "top" );
+	BER_BVSTR( &vals[1], "subentry" );
+	BER_BVSTR( &vals[2], "subschema" );
+	BER_BVSTR( &vals[3], "extensibleObject" );
+	BER_BVZERO( &vals[4] );
+	if ( attr_merge( e, ad_objectClass, vals, NULL ) ) {
 		/* Out of memory, do something about it */
 		entry_free( e );
 		*text = "out of memory";
@@ -95,8 +82,8 @@ schema_info( Entry **entry, const char **text )
 	{
 		int rc;
 		AttributeDescription *desc = NULL;
-		struct berval rdn = global_schemadn;
-		vals[0].bv_val = strchr( rdn.bv_val, '=' );
+		struct berval rdn = frontendDB->be_schemadn;
+		vals[0].bv_val = ber_bvchr( &rdn, '=' );
 
 		if( vals[0].bv_val == NULL ) {
 			*text = "improperly configured subschema subentry";
@@ -115,14 +102,13 @@ schema_info( Entry **entry, const char **text )
 			return LDAP_OTHER;
 		}
 
-		nvals[0].bv_val = strchr( global_schemandn.bv_val, '=' );
-		assert( nvals[0].bv_val );
+		nvals[0].bv_val = ber_bvchr( &frontendDB->be_schemandn, '=' );
+		assert( nvals[0].bv_val != NULL );
 		nvals[0].bv_val++;
-		nvals[0].bv_len = global_schemandn.bv_len -
-			(nvals[0].bv_val - global_schemandn.bv_val);
+		nvals[0].bv_len = frontendDB->be_schemandn.bv_len -
+			(nvals[0].bv_val - frontendDB->be_schemandn.bv_val);
 
-		if( attr_merge_one( e, desc, vals, nvals ) )
-		{
+		if ( attr_merge_one( e, desc, vals, nvals ) ) {
 			/* Out of memory, do something about it */
 			entry_free( e );
 			*text = "out of memory";
@@ -131,37 +117,31 @@ schema_info( Entry **entry, const char **text )
 	}
 
 	{
-		struct		tm *ltm;
 		char		timebuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
 
 		/*
-		 * According to RFC 2251:
+		 * According to RFC 4512:
 
-   Servers SHOULD provide the attributes createTimestamp and
-   modifyTimestamp in subschema entries, in order to allow clients to
-   maintain their caches of schema information.
+   Servers SHOULD maintain the 'creatorsName', 'createTimestamp',       
+   'modifiersName', and 'modifyTimestamp' attributes for all entries of 
+   the DIT. 
 
 		 * to be conservative, we declare schema created 
 		 * AND modified at server startup time ...
 		 */
 
-		ldap_pvt_thread_mutex_lock( &gmtime_mutex );
-		ltm = gmtime( &starttime );
-		lutil_gentime( timebuf, sizeof(timebuf), ltm );
-		ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
-
 		vals[0].bv_val = timebuf;
-		vals[0].bv_len = strlen( timebuf );
+		vals[0].bv_len = sizeof( timebuf );
 
-		if( attr_merge_one( e, ad_createTimestamp, vals, vals ) )
-		{
+		slap_timestamp( &starttime, vals );
+
+		if( attr_merge_one( e, ad_createTimestamp, vals, NULL ) ) {
 			/* Out of memory, do something about it */
 			entry_free( e );
 			*text = "out of memory";
 			return LDAP_OTHER;
 		}
-		if( attr_merge_one( e, ad_modifyTimestamp, vals, vals ) )
-		{
+		if( attr_merge_one( e, ad_modifyTimestamp, vals, NULL ) ) {
 			/* Out of memory, do something about it */
 			entry_free( e );
 			*text = "out of memory";
