@@ -1,8 +1,8 @@
 /* slapcommon.c - common routine for the slap tools */
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapcommon.c,v 1.23.2.7 2005/05/06 16:42:51 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/slapcommon.c,v 1.73.2.4 2007/08/31 23:14:00 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 2003 IBM Corporation.
  * All rights reserved.
@@ -36,6 +36,7 @@
 
 #include "slapcommon.h"
 #include "lutil.h"
+#include "ldif.h"
 
 tool_vars tool_globals;
 
@@ -44,23 +45,33 @@ static char *leakfilename;
 static FILE *leakfile;
 #endif
 
+static LDIFFP dummy;
+
+#if defined(LDAP_SYSLOG) && defined(LDAP_DEBUG)
+int start_syslog;
+static char **syslog_unknowns;
+#ifdef LOG_LOCAL4
+static int syslogUser = SLAP_DEFAULT_SYSLOG_USER;
+#endif /* LOG_LOCAL4 */
+#endif /* LDAP_DEBUG && LDAP_SYSLOG */
+
 static void
 usage( int tool, const char *progname )
 {
 	char *options = NULL;
 	fprintf( stderr,
-		"usage: %s [-v] [-d debuglevel] [-f configfile] [-F configdir]",
+		"usage: %s [-v] [-d debuglevel] [-f configfile] [-F configdir] [-o <name>[=<value>]]",
 		progname );
 
 	switch( tool ) {
 	case SLAPACL:
 		options = "\n\t[-U authcID | -D authcDN] [-X authzID | -o authzDN=<DN>]"
-			"\n\t-b DN -o <var>[=<val>] [-u] [attr[/access][:value]] [...]\n";
+			"\n\t-b DN [-u] [attr[/access][:value]] [...]\n";
 		break;
 
 	case SLAPADD:
-		options = " [-c]\n\t[-n databasenumber | -b suffix]\n"
-			"\t[-l ldiffile] [-q] [-u] [-w]\n";
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix]\n"
+			"\t[-l ldiffile] [-j linenumber] [-q] [-u] [-s] [-w]\n";
 		break;
 
 	case SLAPAUTH:
@@ -68,7 +79,7 @@ usage( int tool, const char *progname )
 		break;
 
 	case SLAPCAT:
-		options = " [-c]\n\t[-n databasenumber | -b suffix]"
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix]"
 			" [-l ldiffile] [-a filter]\n";
 		break;
 
@@ -77,7 +88,7 @@ usage( int tool, const char *progname )
 		break;
 
 	case SLAPINDEX:
-		options = " [-c]\n\t[-n databasenumber | -b suffix] [-q]\n";
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix] [attr ...] [-q]\n";
 		break;
 
 	case SLAPTEST:
@@ -92,19 +103,17 @@ usage( int tool, const char *progname )
 }
 
 static int
-parse_slapacl( void )
+parse_slapopt( void )
 {
-	size_t	len;
+	size_t	len = 0;
 	char	*p;
 
 	p = strchr( optarg, '=' );
-	if ( p == NULL ) {
-		return -1;
+	if ( p != NULL ) {
+		len = p - optarg;
+		p++;
 	}
 
-	len = p - optarg;
-	p++;
-
 	if ( strncasecmp( optarg, "sockurl", len ) == 0 ) {
 		if ( !BER_BVISNULL( &listener_url ) ) {
 			ber_memfree( listener_url.bv_val );
@@ -130,20 +139,54 @@ parse_slapacl( void )
 		ber_str2bv( p, 0, 1, &sock_name );
 
 	} else if ( strncasecmp( optarg, "ssf", len ) == 0 ) {
-		ssf = atoi( p );
+		if ( lutil_atou( &ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
 
 	} else if ( strncasecmp( optarg, "transport_ssf", len ) == 0 ) {
-		transport_ssf = atoi( p );
+		if ( lutil_atou( &transport_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse transport_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
 
 	} else if ( strncasecmp( optarg, "tls_ssf", len ) == 0 ) {
-		tls_ssf = atoi( p );
+		if ( lutil_atou( &tls_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse tls_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
 
 	} else if ( strncasecmp( optarg, "sasl_ssf", len ) == 0 ) {
-		sasl_ssf = atoi( p );
+		if ( lutil_atou( &sasl_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse sasl_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
 
 	} else if ( strncasecmp( optarg, "authzDN", len ) == 0 ) {
 		ber_str2bv( p, 0, 1, &authzDN );
 
+#if defined(LDAP_SYSLOG) && defined(LDAP_DEBUG)
+	} else if ( strncasecmp( optarg, "syslog", len ) == 0 ) {
+		if ( parse_debug_level( p, &ldap_syslog, &syslog_unknowns ) ) {
+			return -1;
+		}
+		start_syslog = 1;
+
+	} else if ( strncasecmp( optarg, "syslog-level", len ) == 0 ) {
+		if ( parse_syslog_level( p, &ldap_syslog_level ) ) {
+			return -1;
+		}
+		start_syslog = 1;
+
+#ifdef LOG_LOCAL4
+	} else if ( strncasecmp( optarg, "syslog-user", len ) == 0 ) {
+		if ( parse_syslog_user( p, &syslogUser ) ) {
+			return -1;
+		}
+		start_syslog = 1;
+#endif /* LOG_LOCAL4 */
+#endif /* LDAP_DEBUG && LDAP_SYSLOG */
+
 	} else {
 		return -1;
 	}
@@ -159,6 +202,8 @@ parse_slapacl( void )
  *	argc, argv	command line arguments
  */
 
+static int need_shutdown;
+
 void
 slap_tool_init(
 	const char* progname,
@@ -172,9 +217,18 @@ slap_tool_init(
 	char *filterstr = NULL;
 	char *subtree = NULL;
 	char *ldiffile	= NULL;
+	char **debug_unknowns = NULL;
 	int rc, i, dbnum;
 	int mode = SLAP_TOOL_MODE;
 	int truncatemode = 0;
+	int use_glue = 1;
+
+#ifdef LDAP_DEBUG
+	/* tools default to "none", so that at least LDAP_DEBUG_ANY 
+	 * messages show up; use -d 0 to reset */
+	slap_debug = LDAP_DEBUG_NONE;
+#endif
+	ldap_syslog = 0;
 
 #ifdef CSRIMALLOC
 	leakfilename = malloc( strlen( progname ) + STRLENOF( ".leak" ) + 1 );
@@ -187,31 +241,31 @@ slap_tool_init(
 
 	switch( tool ) {
 	case SLAPADD:
-		options = "b:cd:f:F:l:n:qtuvw";
+		options = "b:cd:f:F:gj:l:n:o:qsS:tuvw";
 		break;
 
 	case SLAPCAT:
-		options = "a:b:cd:f:F:l:n:s:v";
+		options = "a:b:cd:f:F:gl:n:o:s:v";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPDN:
-		options = "d:f:F:NPv";
+		options = "d:f:F:No:Pv";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPTEST:
-		options = "d:f:F:uv";
+		options = "d:f:F:o:Quv";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPAUTH:
-		options = "d:f:F:M:R:U:vX:";
+		options = "d:f:F:M:o:R:U:vX:";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPINDEX:
-		options = "b:cd:f:F:n:qv";
+		options = "b:cd:f:F:gn:o:qv";
 		mode |= SLAP_TOOL_READMAIN;
 		break;
 
@@ -229,7 +283,7 @@ slap_tool_init(
 	while ( (i = getopt( argc, argv, options )) != EOF ) {
 		switch ( i ) {
 		case 'a':
-			filterstr = strdup( optarg );
+			filterstr = ch_strdup( optarg );
 			break;
 
 		case 'b':
@@ -240,24 +294,51 @@ slap_tool_init(
 			continuemode++;
 			break;
 
-		case 'd':	/* turn on debugging */
-			ldap_debug += atoi( optarg );
-			break;
+		case 'd': {	/* turn on debugging */
+			int	level = 0;
+
+			if ( parse_debug_level( optarg, &level, &debug_unknowns ) ) {
+				usage( tool, progname );
+			}
+#ifdef LDAP_DEBUG
+			if ( level == 0 ) {
+				/* allow to reset log level */
+				slap_debug = 0;
+
+			} else {
+				slap_debug |= level;
+			}
+#else
+			if ( level != 0 )
+				fputs( "must compile with LDAP_DEBUG for debugging\n",
+				       stderr );
+#endif
+			} break;
 
 		case 'D':
 			ber_str2bv( optarg, 0, 1, &authcDN );
 			break;
 
 		case 'f':	/* specify a conf file */
-			conffile = strdup( optarg );
+			conffile = ch_strdup( optarg );
 			break;
 
 		case 'F':	/* specify a conf dir */
-			confdir = strdup( optarg );
+			confdir = ch_strdup( optarg );
+			break;
+
+		case 'g':	/* disable subordinate glue */
+			use_glue = 0;
+			break;
+
+		case 'j':	/* jump to linenumber */
+			if ( lutil_atoi( &jumpline, optarg ) ) {
+				usage( tool, progname );
+			}
 			break;
 
 		case 'l':	/* LDIF file */
-			ldiffile = strdup( optarg );
+			ldiffile = ch_strdup( optarg );
 			break;
 
 		case 'M':
@@ -272,11 +353,13 @@ slap_tool_init(
 			break;
 
 		case 'n':	/* which config file db to index */
-			dbnum = atoi( optarg );
+			if ( lutil_atoi( &dbnum, optarg ) ) {
+				usage( tool, progname );
+			}
 			break;
 
 		case 'o':
-			if ( parse_slapacl() ) {
+			if ( parse_slapopt() ) {
 				usage( tool, progname );
 			}
 			break;
@@ -288,6 +371,11 @@ slap_tool_init(
 			dn_mode = SLAP_TOOL_LDAPDN_PRETTY;
 			break;
 
+		case 'Q':
+			quiet++;
+			slap_debug = 0;
+			break;
+
 		case 'q':	/* turn on quick */
 			mode |= SLAP_TOOL_QUICK;
 			break;
@@ -296,8 +384,19 @@ slap_tool_init(
 			realm = optarg;
 			break;
 
+		case 'S':
+			if ( lutil_atoul( &csnsid, optarg )
+				|| csnsid > SLAP_SYNC_SID_MAX )
+			{
+				usage( tool, progname );
+			}
+			break;
+
 		case 's':	/* dump subtree */
-			subtree = strdup( optarg );
+			if ( tool == SLAPADD )
+				mode |= SLAP_TOOL_NO_SCHEMA_CHECK;
+			else if ( tool == SLAPCAT )
+				subtree = ch_strdup( optarg );
 			break;
 
 		case 't':	/* turn on truncate */
@@ -331,16 +430,43 @@ slap_tool_init(
 		}
 	}
 
+#if defined(LDAP_SYSLOG) && defined(LDAP_DEBUG)
+	if ( start_syslog ) {
+		char *logName;
+#ifdef HAVE_EBCDIC
+		logName = ch_strdup( progname );
+		__atoe( logName );
+#else
+		logName = (char *)progname;
+#endif
+
+#ifdef LOG_LOCAL4
+		openlog( logName, OPENLOG_OPTIONS, syslogUser );
+#elif defined LOG_DEBUG
+		openlog( logName, OPENLOG_OPTIONS );
+#endif
+#ifdef HAVE_EBCDIC
+		free( logName );
+#endif
+	}
+#endif /* LDAP_DEBUG && LDAP_SYSLOG */
+
 	switch ( tool ) {
 	case SLAPADD:
 	case SLAPCAT:
-	case SLAPINDEX:
 		if ( ( argc != optind ) || (dbnum >= 0 && base.bv_val != NULL ) ) {
 			usage( tool, progname );
 		}
 
 		break;
 
+	case SLAPINDEX:
+		if ( dbnum >= 0 && base.bv_val != NULL ) {
+			usage( tool, progname );
+		}
+
+		break;
+
 	case SLAPDN:
 		if ( argc == optind ) {
 			usage( tool, progname );
@@ -373,12 +499,11 @@ slap_tool_init(
 		break;
 	}
 
-	ldap_syslog = 0;
-
 	if ( ldiffile == NULL ) {
-		ldiffp = tool == SLAPCAT ? stdout : stdin;
+		dummy.fp = tool == SLAPCAT ? stdout : stdin;
+		ldiffp = &dummy;
 
-	} else if ((ldiffp = fopen( ldiffile, tool == SLAPCAT ? "w" : "r" ))
+	} else if ((ldiffp = ldif_open( ldiffile, tool == SLAPCAT ? "w" : "r" ))
 		== NULL )
 	{
 		perror( ldiffile );
@@ -389,44 +514,38 @@ slap_tool_init(
 	 * initialize stuff and figure out which backend we're dealing with
 	 */
 
-#ifdef SLAPD_MODULES
-	if ( module_init() != 0 ) {
-		fprintf( stderr, "%s: module_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-#endif
-		
-	rc = slap_schema_init();
-
-	if ( rc != 0 ) {
-		fprintf( stderr, "%s: slap_schema_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-
 	rc = slap_init( mode, progname );
-
 	if ( rc != 0 ) {
 		fprintf( stderr, "%s: slap_init failed!\n", progname );
 		exit( EXIT_FAILURE );
 	}
 
-	if ( frontend_init() ) {
-		fprintf( stderr, "%s: frontend_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-
-	if ( overlay_init() ) {
-		fprintf( stderr, "%s: overlay_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-
 	rc = read_config( conffile, confdir );
 
 	if ( rc != 0 ) {
-		fprintf( stderr, "%s: bad configuration file!\n", progname );
+		fprintf( stderr, "%s: bad configuration %s!\n",
+			progname, confdir ? "directory" : "file" );
 		exit( EXIT_FAILURE );
 	}
 
+	if ( debug_unknowns ) {
+		rc = parse_debug_unknowns( debug_unknowns, &slap_debug );
+		ldap_charray_free( debug_unknowns );
+		debug_unknowns = NULL;
+		if ( rc )
+			exit( EXIT_FAILURE );
+	}
+
+#if defined(LDAP_SYSLOG) && defined(LDAP_DEBUG)
+	if ( syslog_unknowns ) {
+		rc = parse_debug_unknowns( syslog_unknowns, &ldap_syslog );
+		ldap_charray_free( syslog_unknowns );
+		syslog_unknowns = NULL;
+		if ( rc )
+			exit( EXIT_FAILURE );
+	}
+#endif
+
 	at_oc_cache = 1;
 
 	switch ( tool ) {
@@ -444,6 +563,16 @@ slap_tool_init(
 		break;
 	}
 
+	if ( use_glue ) {
+		rc = glue_sub_attach();
+
+		if ( rc != 0 ) {
+			fprintf( stderr,
+				"%s: subordinate configuration error\n", progname );
+			exit( EXIT_FAILURE );
+		}
+	}
+
 	rc = slap_schema_check();
 
 	if ( rc != 0 ) {
@@ -497,7 +626,7 @@ slap_tool_init(
 			exit( EXIT_FAILURE );
 		}
 
-		be = select_backend( &nbase, 0, 0 );
+		be = select_backend( &nbase, 0 );
 		ber_memfree( nbase.bv_val );
 
 		switch ( tool ) {
@@ -521,12 +650,17 @@ slap_tool_init(
 		}
 
 	} else if ( dbnum == -1 ) {
+		/* no suffix and no dbnum specified, just default to
+		 * the first available database
+		 */
 		if ( nbackends <= 0 ) {
 			fprintf( stderr, "No available databases\n" );
 			exit( EXIT_FAILURE );
 		}
 		LDAP_STAILQ_FOREACH( be, &backendDB, be_next ) {
 			dbnum++;
+
+			/* db #0 is cn=config, don't select it as a default */
 			if ( dbnum < 1 ) continue;
 		
 		 	if ( SLAP_MONITOR(be))
@@ -548,19 +682,19 @@ slap_tool_init(
 			exit( EXIT_FAILURE );
 		}
 		
-		if ( nosubordinates == 0 && dbnum > 0 ) {
+		if ( nosubordinates == 0 && dbnum > 1 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"The first database does not allow %s;"
 				" using the first available one (%d)\n",
-				progname, dbnum + 1, 0 );
+				progname, dbnum, 0 );
 		}
 
 	} else if ( dbnum < 0 || dbnum > (nbackends-1) ) {
 		fprintf( stderr,
 			"Database number selected via -n is out of range\n"
-			"Must be in the range 1 to %d"
-			" (number of databases in the config file)\n",
-			nbackends );
+			"Must be in the range 0 to %d"
+			" (number of configured databases)\n",
+			nbackends-1 );
 		exit( EXIT_FAILURE );
 
 	} else {
@@ -576,27 +710,42 @@ startup:;
 	mal_leaktrace(1);
 #endif
 
-	if ( !dryrun && be && slap_startup( be ) ) {
-		switch ( tool ) {
-		case SLAPTEST:
-			fprintf( stderr, "slap_startup failed "
-					"(test would succeed using "
-					"the -u switch)\n" );
-			break;
+	if ( conffile != NULL ) {
+		ch_free( conffile );
+	}
 
-		default:
-			fprintf( stderr, "slap_startup failed\n" );
-			break;
+	if ( ldiffile != NULL ) {
+		ch_free( ldiffile );
+	}
+
+	/* slapdn doesn't specify a backend to startup */
+	if ( !dryrun && tool != SLAPDN ) {
+		need_shutdown = 1;
+
+		if ( slap_startup( be ) ) {
+			switch ( tool ) {
+			case SLAPTEST:
+				fprintf( stderr, "slap_startup failed "
+						"(test would succeed using "
+						"the -u switch)\n" );
+				break;
+
+			default:
+				fprintf( stderr, "slap_startup failed\n" );
+				break;
+			}
+
+			exit( EXIT_FAILURE );
 		}
-		
-		exit( EXIT_FAILURE );
 	}
 }
 
 void slap_tool_destroy( void )
 {
 	if ( !dryrun ) {
-		slap_shutdown( be );
+		if ( need_shutdown ) {
+			slap_shutdown( be );
+		}
 		slap_destroy();
 	}
 #ifdef SLAPD_MODULES
@@ -619,4 +768,8 @@ void slap_tool_destroy( void )
 	if ( !BER_BVISNULL( &authcDN ) ) {
 		ch_free( authcDN.bv_val );
 	}
+
+	if ( ldiffp && ldiffp != &dummy ) {
+		ldif_close( ldiffp );
+	}
 }
