@@ -1,8 +1,8 @@
 /* syntax.c - routines to manage syntax definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/syntax.c,v 1.36.2.1 2003/12/01 00:58:37 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/syntax.c,v 1.43.2.2 2007/08/31 23:14:00 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -23,7 +23,6 @@
 #include <ac/socket.h>
 
 #include "slap.h"
-#include "ldap_pvt.h"
 
 struct sindexrec {
 	char		*sir_name;
@@ -31,7 +30,7 @@ struct sindexrec {
 };
 
 static Avlnode	*syn_index = NULL;
-static LDAP_SLIST_HEAD(SyntaxList, slap_syntax) syn_list
+static LDAP_SLIST_HEAD(SyntaxList, Syntax) syn_list
 	= LDAP_SLIST_HEAD_INITIALIZER(&syn_list);
 
 static int
@@ -77,16 +76,48 @@ syn_find_desc( const char *syndesc, int *len )
 	return( NULL );
 }
 
+int
+syn_is_sup( Syntax *syn, Syntax *sup )
+{
+	int	i;
+
+	assert( syn != NULL );
+	assert( sup != NULL );
+
+	if ( syn == sup ) {
+		return 1;
+	}
+
+	if ( syn->ssyn_sups == NULL ) {
+		return 0;
+	}
+
+	for ( i = 0; syn->ssyn_sups[i]; i++ ) {
+		if ( syn->ssyn_sups[i] == sup ) {
+			return 1;
+		}
+
+		if ( syn_is_sup( syn->ssyn_sups[i], sup ) ) {
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 void
 syn_destroy( void )
 {
-	Syntax *s;
+	Syntax	*s;
 
-	avl_free(syn_index, ldap_memfree);
-	while( !LDAP_SLIST_EMPTY(&syn_list) ) {
-		s = LDAP_SLIST_FIRST(&syn_list);
-		LDAP_SLIST_REMOVE_HEAD(&syn_list, ssyn_next);
-		ldap_syntax_free((LDAPSyntax *)s);
+	avl_free( syn_index, ldap_memfree );
+	while( !LDAP_SLIST_EMPTY( &syn_list ) ) {
+		s = LDAP_SLIST_FIRST( &syn_list );
+		LDAP_SLIST_REMOVE_HEAD( &syn_list, ssyn_next );
+		if ( s->ssyn_sups ) {
+			SLAP_FREE( s->ssyn_sups );
+		}
+		ldap_syntax_free( (LDAPSyntax *)s );
 	}
 }
 
@@ -105,12 +136,7 @@ syn_insert(
 		sir = (struct sindexrec *)
 			SLAP_CALLOC( 1, sizeof(struct sindexrec) );
 		if( sir == NULL ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, ERR, 
-				"syn_insert: SLAP_CALLOC Error\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
-#endif
 			return LDAP_OTHER;
 		}
 		sir->sir_name = ssyn->ssyn_oid;
@@ -135,17 +161,12 @@ syn_add(
 )
 {
 	Syntax		*ssyn;
-	int		code;
+	int		code = 0;
 
 	ssyn = (Syntax *) SLAP_CALLOC( 1, sizeof(Syntax) );
-	if( ssyn == NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"syn_add: SLAP_CALLOC Error\n", 0, 0, 0 );
-#else
+	if ( ssyn == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
-#endif
-		return LDAP_OTHER;
+		return SLAP_SCHERR_OUTOFMEM;
 	}
 
 	AC_MEMCPY( &ssyn->ssyn_syn, syn, sizeof(LDAPSyntax) );
@@ -162,12 +183,48 @@ syn_add(
 	ssyn->ssyn_validate = def->sd_validate;
 	ssyn->ssyn_pretty = def->sd_pretty;
 
+	ssyn->ssyn_sups = NULL;
+
 #ifdef SLAPD_BINARY_CONVERSION
 	ssyn->ssyn_ber2str = def->sd_ber2str;
 	ssyn->ssyn_str2ber = def->sd_str2ber;
 #endif
 
-	code = syn_insert(ssyn, err);
+	if ( def->sd_sups != NULL ) {
+		int	cnt;
+
+		for ( cnt = 0; def->sd_sups[cnt] != NULL; cnt++ )
+			;
+		
+		ssyn->ssyn_sups = (Syntax **)SLAP_CALLOC( cnt + 1,
+			sizeof( Syntax * ) );
+		if ( ssyn->ssyn_sups == NULL ) {
+			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
+			code = SLAP_SCHERR_OUTOFMEM;
+
+		} else {
+			for ( cnt = 0; def->sd_sups[cnt] != NULL; cnt++ ) {
+				ssyn->ssyn_sups[cnt] = syn_find( def->sd_sups[cnt] );
+				if ( ssyn->ssyn_sups[cnt] == NULL ) {
+					*err = def->sd_sups[cnt];
+					code = SLAP_SCHERR_SYN_SUP_NOT_FOUND;
+				}
+			}
+		}
+	}
+
+	if ( code == 0 ) {
+		code = syn_insert( ssyn, err );
+
+	}
+
+	if ( code != 0 && ssyn != NULL ) {
+		if ( ssyn->ssyn_sups != NULL ) {
+			SLAP_FREE( ssyn->ssyn_sups );
+		}
+		SLAP_FREE( ssyn );
+	}
+
 	return code;
 }
 
@@ -181,35 +238,24 @@ register_syntax(
 
 	syn = ldap_str2syntax( def->sd_desc, &code, &err, LDAP_SCHEMA_ALLOW_ALL);
 	if ( !syn ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( CONFIG, ERR, 
-			"register_syntax: Error - %s before %s in %s.\n",
-			ldap_scherr2str(code), err, def->sd_desc );
-#else
 		Debug( LDAP_DEBUG_ANY, "Error in register_syntax: %s before %s in %s\n",
 		    ldap_scherr2str(code), err, def->sd_desc );
-#endif
 
 		return( -1 );
 	}
 
 	code = syn_add( syn, def, &err );
 
-	ldap_memfree( syn );
-
 	if ( code ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( CONFIG, ERR, 
-			"register_syntax: Error - %s %s in %s\n", 
-			scherr2str(code), err, def->sd_desc );
-#else
 		Debug( LDAP_DEBUG_ANY, "Error in register_syntax: %s %s in %s\n",
 		    scherr2str(code), err, def->sd_desc );
-#endif
+		ldap_syntax_free( syn );
 
 		return( -1 );
 	}
 
+	ldap_memfree( syn );
+
 	return( 0 );
 }
 
@@ -235,14 +281,8 @@ syn_schema_info( Entry *e )
 			return -1;
 		}
 #if 0
-#ifdef NEW_LOGGING
-		LDAP_LOG( config, ENTRY,
-			   "syn_schema_info: Merging syn [%ld] %s\n",
-			   (long)val.bv_len, val.bv_val, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "Merging syn [%ld] %s\n",
 	       (long) val.bv_len, val.bv_val, 0 );
-#endif
 #endif
 
 		nval.bv_val = syn->ssyn_oid;
